#!/usr/bin/env perl
# /home/foth/dawai/bin/evalmethod.pl

# Takes two patterns and compares solution transcripts found in files
# with those names. Alternatively, takes two patterns and performs a
# direct comparison between corresponding files.

# Say the test pattern is `pruning' and the reference pattern is
# `search'. Then each file in the cwd whose name matches /pruning/ is
# analysed and the performance logged in it is compared to the
# corresponding file with `search' in place of `pruning'. The
# reference pattern may have pathnames in it, so you can easily
# compare files stored in different directories.

# The files are grepped for the `took 7654 ms' bit to find the
# speed of the method, and for the output of parseVerify()
# to find the quality.

use warnings;
use strict;

use DB_File;
my %db;
my $db_name = 'RESULTS';
tie %db, 'DB_File', $db_name or die "Can't write `$db_name': $!\n";


# get options
use Getopt::Std;
our ($opt_b, $opt_d, $opt_t, $opt_F, $opt_f, $opt_h, $opt_l, $opt_m, $opt_n,
     $opt_q, $opt_s, $opt_c, $opt_r, $opt_x, $opt_R);
getopts("bc:deF:f:hl:m:nqr:stx:R:");

our $usage = "
Usage: $0 <experiment name> [<experiment name>]

   -R NUM     tabulate sentences in classes by Range
   -b         measure Build time, not solution time
   -c NUM     tabulate sentences in Classes by length
   -d         Disregard failed parse runs
   -f STRING  use only sentenes named in File STRING
   -F STRING  ignore sentences named in File STRING
   -h         show this help
   -l STRING  use only Level STRING
   -m STRING  Measure structural/labelled/lexical/strict accuracy
   -n         print No leading newline
   -r STRING  ignore sentences with number not in Range STRING
   -q         do not print progress information
   -s         be Silent about sentences, give only overall figures
   -t         compute Tree accuracy
   -x STRING  ignore sentences whose length is not in range STRING

To compare against an experiment with logfiles in another directory,
specify an experiment name like `../baseline:fullsearch'.

";

die $usage if $opt_h;
die $usage unless (@ARGV == 1 or @ARGV == 2);

# default to `SYN'
our $level = defined $opt_l ? $opt_l : 'SYN';

# default to measuring structural accuracy
our $measure = defined $opt_m ? $opt_m : 'structural';

our $range = defined $opt_r ? $opt_r : '';
our %ids=();
our $min = 0;
our $max = 999999;

our $minsize = 1;
our $maxsize = 999;

my $nr_runs = 0;
my $nr_failed = 0;

if($range) {
  if($range =~ /^(\d+)-(\d+)/) {
    ($min,$max) = ($1, $2);
  }
}

if($opt_x) {
  if($opt_x =~ /^(\d+)-(\d+)/) {
    ($minsize,$maxsize) = ($1, $2);
  }
  elsif($opt_x =~ /^(\d+)/) {
    ($minsize,$maxsize) = (1, $1);
  }
}

if (defined $opt_f || defined $opt_F) {
  my $file = $opt_f || $opt_F;
  open(FILE, "$file") or die "Can't read `$file': $!\n";
  while (<FILE>) {
    chomp;
    $ids{$_}++;
  }
  close(FILE);
}

our $compare = (@ARGV == 2);

my $test = shift;
my $reference = $compare ? shift : undef;

# unpack reference spec, e.g. "../January:search"
our $refdir = '.';
if(defined $reference && $reference =~ /^(.+):(.+)$/) {
  ($refdir, $reference) = ($1,$2);
}

my $qtest = quotemeta $test;
my $qreference;
if($compare) {
 $qreference = quotemeta $reference;
}

our $max_length = 0;

our %data;
our %refdata;
our %total;


# Find reference file for a test file
sub find_reference {

  my $testfile = shift;
  my $reffile = "$refdir/$testfile";
  $reffile =~ s/\b$qtest\./$qreference./;

  # if the reference file is the same as the test file, that is
  # probably a programming error, UNLESS the user explicitly specified
  # identical test and reference methods.
  if ($reffile eq "$refdir/$testfile" &&
     $test ne $reference) {
    die "Substitution s/\\b$qtest/$qreference/ failed!\n";
  }

  return $reffile;
}


# compare strings intelligently
sub byVName {

  # don't change the names themselves
  my ($x,$y) = ($a,$b);

  while($x ne "" && $y ne "") {

    # skip identical parts
    while(substr($x,0,1) eq substr($y,0,1) && $x !~ /^[0-9]/) {
      substr($x,0,1) = "";
      substr($y,0,1) = "";
    }

    # compare alphabetical parts normally
    if($x !~ /^\d/ || $y !~ /^\d/) {
      return $x cmp $y;
    }

    # compare numeric parts numerically
    $x=~s/^(\d+)//;
    my $nx = $1;
    $y=~s/^(\d+)//;
    my $ny = $1;
    my $result = $nx <=> $ny;
    next unless $result;
    return $result;
  }

}

# Find all relevant log files:
#
#  - when evaluating a method, all logs with a fitting name
#
#  - when comparing two methods, all logs with a fitting name for
#    which a reference log exists)
sub find_logs {

  my @result;
  opendir(DOT,'.') or die "Can't read . : $!\n";

  # find test files
  for my $testfile (readdir(DOT)) {
    chomp $testfile;
    next if $testfile !~ /\b$qtest\.log\b/;

    # obey -F flag
    if ($opt_F) {
      $testfile =~ /^([^.]+)\./;
      my $utterance = $1;
      next if exists $ids{$utterance};
    }

    # obey -f flag
    if ($opt_f) {
      $testfile =~ /^([^.]+)\./;
      my $utterance = $1;
      next unless exists $ids{$utterance};
    }

    # obey -r flag
    if ($range && $testfile =~ /-s(\d+)/) {
      my $n = $1;
      if ($n < $min || $n > $max) {
        next;
      }
    }

    # maybe find reference file
    if ($compare) {
      my $reffile = find_reference $testfile;
      next unless -f $reffile;
    }
    push @result, $testfile;
  }

  die "No data for comparison.\n" unless @result;

  return sort byVName @result;
}

# Extract data about a test run from a log
sub read_log {

  my $file = shift;

  # normalize file name
  $file =~ s|^./||;

  # if the data was already extracted once, read it from the cache
  my $key = "$file:$measure";
  # TODO: should include the file modification time, too

  if(exists $db{$key}) {
    my($time, $size, $correct, $expired) = split( /:/, $db{$key});

    if ($size > $maxsize or $size < $minsize) {
      return undef;
    }

    $max_length = $size if $size > $max_length;

    return {
            filename      => $file,
            no            => $nr_runs+1,
            time          => $time,
            size          => $size,
            correct       => $correct,
            expired       => $expired,
           };
  }
  my %result = (
     filename      => $file,
     no            => $nr_runs+1,  # because nr_files is not yet incremented
     time          => undef,
     size          => undef,
     correct       => undef,
     expired       => 0,
    );

  my $nrlevels = 0;
  my $is_complete = 0;
  my $empty_domain = '';

  open(LOG, $file) or die "Can't open $file: $!";
  while (<LOG>) {

    # extract expiration flag
    if (/timer expired!/) {
      $result{expired} = 1;
    }

    # detect lexicon lookup failure
    elsif(m!can't find a value for (.+)/SYN!) {
      $empty_domain = $1;
    }

    # detect failure to build net
    elsif(/ERROR: Don't know what problem to solve./) {
      if($empty_domain) {
        warn "empty domain for `$empty_domain' in $file\n";
      } else {
        warn "Unknown error in $file\n";
      }
      return undef;
    }

    # extract info that this is a log file at all
    elsif (/^CDG parser/) {
      $result{is_log} = 1;
    }

    # extract number of levels
    elsif (m|loaded: \d+/(\d+)/\d+/\d+/\d+/\d+/\d+|) {
      if ($1 > 0) {
        $nrlevels += $1;
      }
    }

    # extract problem size
    elsif (!$result{size} && /#nodes (\d+), #edges/) {
      my $size = $1;
      $size /= $nrlevels;

      if ($size > $maxsize or $size < $minsize) {
        return undef;
      }

      $result{size} = $size;
      $max_length = $size if $size > $max_length;

    }

    # extract runtime
    elsif ($opt_b && /newnet took ([0-9.]+)ms\b/) {
      $result{'time'} = $1 / 1000;
    } elsif (!$opt_b && /([0-9.]+)s spent [a-z]+ing\b/) {

      # This is correct for frobbing and other methods that use the
      # transform module, since the `spent frobbing' method prints
      # timerElapsed(P->profileTime), and that is cumulative over
      # multiple frobbing runs. If you mix frobbing with a solution
      # method that reuses the same ConstraintNet but not the same
      # Problem, you would need +=.
      $result{'time'} = $1;
    } elsif (!$opt_b && /total         : *(\d+) ms/) {
      $result{'time'} += $1 / 1000;
    }

    # extract accuracy
    elsif (/$level:\s+([0-9.]+)\s+([0-9.]+)\s+([0-9.]+)\s+([0-9.]+)/) {
      my($structural,$labelled,$lexical,$strict) = ($1,$2,$3,$4);

      if ($measure eq 'structural') {
        $result{correct} = sprintf "%.0f", ($structural / 100.0 * $result{size});
      }

      elsif ($measure eq 'labelled') {
        $result{correct} = sprintf "%.0f", ($labelled / 100.0 * $result{size});
      }

      elsif ($measure eq 'lexical') {
        $result{correct} = sprintf "%.0f", ($lexical / 100.0 * $result{size});
      }

      elsif ($measure eq 'strict') {
        $result{correct} = sprintf "%.0f", ($strict / 100.0 * $result{size});
      }
      else {
        die "Don't know measure `$measure'!\n";
      }

    }

    # extract completion flag
    elsif (/^ *quit/) {
      $is_complete = 1;
    }

  }
  close(LOG);

  # perhaps this was not a CDG log file at all?
  if (! $result{is_log}) {
    warn "$file is not a log!\n";
    return undef;
  }

  # perhaps this log is not yet complete?
  if (! $is_complete) {
    warn "$file is incomplete!\n";
    return undef;
  }

  # perhaps parsing failed?
  if (!defined($result{correct})) {
    if ($opt_d) {
      $nr_failed++;
      return undef;
    } else {
      $result{correct} = 0;
      $result{'time'} = 0;
    }
  }

  # no run time found?
  if (!defined $result{'time'}) {

    # if the file is fairly old, assume something went wrong
    if (! -z $file and -M $file > 1/(24*60)) {
      warn "WARNING: no time parsed for $file, disregarding...\n";
    }
    return undef;
  }

  # no size found?
  if (!defined $result{size}) {
    warn "WARNING: size not known for $file, disregarding...\n";
    return undef;
  }

  $db{$key} = "$result{time}:$result{size}:$result{correct}:$result{expired}";

  return \%result;
}


# Print result of one run or comparison
sub local_stats {

  my $data = shift;
  my $refdata = shift;

  $total{'time'} += $data->{'time'};
  $total{correct} += $data->{correct};
  $total{edges} += $data->{size};
  if($refdata) {
    $total{reftime} += $refdata->{'time'};
    $total{refcorrect} += $refdata->{correct};
  }

  return if $opt_s;

  if ($refdata) {
    printf("%4d  %30s: time %6.3f (%6d%sms), quality %s (%d/%d/%d)\n",
           $nr_runs,
           $data->{filename},
           $refdata->{'time'} > 0 ? $data->{'time'} / $refdata->{'time'} : 0,
           1000*$refdata->{'time'},
           $data->{expired} ? '*' : ' ',
           $refdata->{correct} ? sprintf ("%6.3f", $data->{correct} / $refdata->{correct}) : "Inf. ",
           $data->{correct},
           $refdata->{correct},
           $data->{size}
          );

  } else {
    printf("%4d %35s: quality %5.1f%% (%d/%d, %d ms%s)\n",
           $nr_runs,
           $data->{filename},
           100 * $data->{correct} / $data->{size},
           $data->{correct},
           $data->{size},
           1000*$data->{'time'},
           $data->{expired} ? '*' : ''
          );
  }

}

# extract performance information from log files
sub process_logs {

 FILE:
  for my $testfile (@_) {

    my $data = read_log($testfile);
    my $refdata;
    next FILE unless defined $data;

    if (0 == $nr_runs % 100) {
      print STDERR "$nr_runs\r" unless $opt_q;
    }

    if ($compare) {
      my $reffile = find_reference $testfile;
      $refdata = read_log ($reffile);
      next FILE unless defined $refdata;
    }

    $nr_runs++;

    # print one line of information and update in %total
    local_stats($data, $refdata);

    # retain detailed data for later use by $opt_c or the like
    $data{$testfile} = $data;
    $refdata{$testfile} = $refdata;
  }

  die "No data for comparison.\n" unless $nr_runs;

}

# Print results by sentence length
sub size_stats {

  my ($step_size, $cutoff) = ($opt_c =~ /(\d+):(\d+)/);
  if(!$cutoff) {
    $step_size = $opt_c;
    $cutoff = 1000;
  }

  my $i = 0;
  print "\n\nResults by sentence length:\n===========================\n\n";

  while ($i <= $max_length) {
    $i += $step_size;

    my $min = $i - $step_size + 1;
    my $max = $i;
    if($i > $cutoff) {
      $max = 1000;
    }

    my($classtime,
       $classcorrect,
       $totalsize,
       $classreftime,
       $classrefcorrect,
       $no);

    for (keys %data) {
      my $d = $data{$_};
      my $s = $d->{size};

      next if $s < $min;
      next if $s > $max;

      $no++;

      $classtime += $d->{time};
      $classcorrect += $d->{correct};
      $totalsize += $d->{size};
      if ($compare) {
        my $r = $refdata{$_};
        $classreftime += $r->{'time'};
        $classrefcorrect += $r->{correct};
      }
    }

    if ($no) {
      if($max - $min > $step_size+1) {
        printf("           >%3d:", $min-1);
      } else {
        printf("       %3d -%3d:", $min, $max);
      }
      if ($compare) {
        printf(" %4d cases, time %8.2f, quality %8.3f\n",
               $no,
               $classtime / $classreftime,
               $classcorrect / $classrefcorrect);
      } else {
        printf(" %4d cases, time %8.2f, quality %6.2f\n",
               $no,
               $classtime / $no,
               100 * $classcorrect / $totalsize);
      }
    }
    last if $max - $min > $step_size+1;
  }
}

# Print results on parts of the corpus
sub range_stats {
  my $step = shift;
  my $i = 0;
  my $found;
  my $no;
  print "\n\nResults by range:\n=================\n\n";

  do {
    $found = 0;

    my($classsize,
       $classtime,
       $classcorrect,
       $classreftime,
       $classrefcorrect);

    for (keys %data) {
      my $d = $data{$_};
      my $nr = $d->{no};

      next unless ($nr > $i && $nr <= $i+$step);

      $no++;
      $found++;
      $classtime += $d->{time};
      $classsize += $d->{size};
      $classcorrect += $d->{correct};
      if ($compare) {
        my $r = $refdata{$_};
        $classreftime += $r->{time};
        $classrefcorrect += $r->{correct};
      }
    }

    if ($found) {
      if ($compare) {
        printf("               %3d: %4d cases, time %8.2f, quality %8.3f (%d/%d)\n",
               $i, $found,
               $classtime / $classreftime,
               $classcorrect / $classrefcorrect,
               $classcorrect,
               $classrefcorrect);
      } else {
        printf("                %4d: %4d cases, time %8.2f, quality %6.3f (%.1f words)\n",
               $i,
               $found,
               $classtime / $found,
               100 * $classcorrect / $classsize,
               $classsize / $found );
      }
    }
    $i += $step;
  } while $found;
}

# Print overall results
sub global_stats {

  print "="x80, "\n" unless $opt_s;

  if ($opt_c) {
    size_stats();
  }

  if ($opt_R) {
    range_stats($opt_R);
  }

  # give overall results
  if ($compare) {
    printf "\n" unless $opt_n;
    printf("total: %5d cases, time %8.3f, quality %4.3f (%d/%d/%d)\n",
           $nr_runs,
           $total{'time'} / $total{reftime},
           $total{correct} / $total{refcorrect},
           $total{correct},
           $total{refcorrect},
           $total{edges});
  } else {
    printf "\n" unless $opt_n;
    printf("total: %5d cases, time %8.2f, quality %4.1f%% (%d/%d) (%.1f words)\n",
           $nr_runs,
           $total{time} / $nr_runs,
           100 * $total{correct} / $total{edges},
           $total{correct},
           $total{edges},
           $total{edges} / $nr_runs);

    if ($nr_failed) {
      print "$nr_failed parsing runs failed.\n";
    }

  }

  # give tree error rate
  if($opt_t) {
    my $perfect = 0;
    for my $d (values %data) {
      if($$d{size} == $$d{correct}) {
        $perfect++;
      }
    }
    printf "\n Tree accuracy: %4.1f%% ($perfect/$nr_runs)\n", 100 * $perfect/$nr_runs;
  }

  print "\n";

}

### MAIN ####

# decide which files to process
my @logs = find_logs();

# do it
process_logs(@logs);

# print report
global_stats();


untie %db;
