#!/bin/bash 
# prepare-experiment.sh
#

# parse options
while getopts "c:g:r:" arg; do
  case $arg in

    # saying "-c agree1,agree2,agree3" schedules an experiment
    # which will test different weights for these constraints
    c)
    CONSTRAINTS=$OPTARG
    ;;
    g)
    GRAMMARFILE=$OPTARG
    ;;
    r)
    RANGE=$OPTARG
    ;;
  esac
done

shift $(($OPTIND - 1))

# check invocation
if [ "$#" != 1 ]; then
  echo Usage: $0 \<directory name\>
  exit 1
fi


# command-line parameter gives the name of the experimentation dir
name=$1

if [ -a "$name" ]; then
  echo File exists!
  exit 1
else 
  mkdir "$name"
fi
cd "$name"

# File that the entire grammar resides in. It will be copied to the
# experimentation directory so development on the grammar can proceed
# during the experiment.
if [ -z "$GRAMMARFILE" ]; then
  GRAMMARFILE=/data/corpus/ccc/heiseticker/heiseticker.cdg
fi
(cd $(dirname $GRAMMARFILE) && make local)

# The parser itself is also copied, so recompilation will not break
# the experiment. 
CDGP=$CDGHOME/cdgp/cdgp_static

cp ~/.cdgrc .
mkdir scripts

# The file README is supposed to contain a brief description of what
# it is you are testing.
if [ ! -z "$CONSTRAINTS" ]; then
  echo "Test different weights for constraint $CONSTRAINTS" > README
else
  $EDITOR README
fi

GRAMMAR=$(basename $GRAMMARFILE)
GRAMMAR=${GRAMMAR%.cdg}

# experiments.pl contains an array in perl describing what you wish done.
if [ ! -z "$CONSTRAINTS" ]; then
echo "@experiments = (
  { id => \"zero\", grammar => \"$GRAMMAR\", script => \"trinamic0\" },
  { id => \"hundredth\", grammar => \"$GRAMMAR\", script => \"trinamic001\" },
  { id => \"twentieth\", grammar => \"$GRAMMAR\", script => \"trinamic005\" },
  { id => \"one\", grammar => \"$GRAMMAR\", script => \"trinamic01\" },
  { id => \"two\", grammar => \"$GRAMMAR\", script => \"trinamic02\" },
  { id => \"three\", grammar => \"$GRAMMAR\", script => \"trinamic03\" },
  { id => \"four\", grammar => \"$GRAMMAR\", script => \"trinamic04\" },
  { id => \"five\", grammar => \"$GRAMMAR\", script => \"trinamic05\" },
  { id => \"six\", grammar => \"$GRAMMAR\", script => \"trinamic06\" },
  { id => \"seven\", grammar => \"$GRAMMAR\", script => \"trinamic07\" },
  { id => \"eight\", grammar => \"$GRAMMAR\", script => \"trinamic08\" },
  { id => \"nine\", grammar => \"$GRAMMAR\", script => \"trinamic09\" },
  { id => \"off\", grammar => \"$GRAMMAR\", script => \"trinamic1\" }
 );" > experiments.pl
   
else
echo "@experiments = 
 (
   { id => 'trinamic', grammar => '$GRAMMAR', script => 'trinamic' }
 );" > experiments.pl
$EDITOR experiments.pl
fi

# copy grammar and parser to /usr/local/home for better speed.
cp $GRAMMARFILE .
mkdir -p scripts
cp ~/cdg/utils/scripts/trinamic scripts

(cd ~/cdg/cdgp && make cdgp_static)
cp $CDGP ./cdgp

if [ ! -z "$CONSTRAINTS" ]; then

  # split arguments at commas.
  # no, bash cannot do it any better than this.
  echo $CONSTRAINTS | perl -n -e 'print"$1\n"while s/^([^,]+),?//' > /tmp/tmp.$$
  constraints=`cat /tmp/tmp.$$ | sed "s/ /SPACE/g"`
  rm /tmp/tmp.$$

  # automatically create script files which set different weights for
  # the specified constraints.
  cd scripts
  function create_script {
    local weight=$1                       
    local name=$2

    rm -f trinamic$name
    for x in $constraints; do
      echo "weight '$x' $weight" | sed "s/SPACE/ /g"  >> trinamic$name
    done
    cat trinamic >> trinamic$name

  }        

  create_script 0.0 0
  create_script 0.01 001
  create_script 0.05 005
  create_script 0.1 01
  create_script 0.2 02
  create_script 0.3 03
  create_script 0.4 04
  create_script 0.5 05
  create_script 0.6 06
  create_script 0.7 07
  create_script 0.8 08
  create_script 0.9 09
  create_script 1.0 1

fi

if [ ! -z "$RANGE" ]; then
  echo $RANGE > RANGE
fi
