#! /usr/bin/perl -w

use strict;
use locale;

#	This script compares the CDG annotation file(s) resulting from a parse of sentences from the van Kampen corpus with a given gold standard 
# 	and computes precision and recall based on the number of correctly assigned edge labels. A central assumption for this evaluation is that 
#	sentences in the corpus are unique such that there is only one annotation per sentence. The script will abort in case of multiple occurrences
#	of the same sentence.
#	The van Kampen corpus contains a range of sentences with global genitive-dative ambiguity for which a unique annotation in the gold standard
#	is not possible as - in principle - both genitive and dative reading are grammatical (though not necessarily equally acceptable). For sentences
#	which are known to be globally ambiguous this script therefore considers either annotation as correct.
#
#	The calculation of recall is based on the assumption that the test file aims to annotation all sentences in the Gold Standard. 
#	A relative measure based on the sentences to evaluate might be more useful here for test files not aspiring to full coverage of 
#	the gold standard annotation file.
#	For efficiency, the gold standard annotation is hashed. The hash is multidimensional: 
#	standard{ sentence }{ dependant's word number }{ level }{ attachment label } = regent's word number in the sentence.
#
# 	Invocation parameters:
#	======================
#	1. File containing the gold standard.
#   2. File(s) containing the CDG2 annotations to be compared against the gold standard.


# This flag determines if label assignments deviating from the gold standard are to printed (1) or not (0).
my $diagnostics = 1;	

# This flag determines if the evaluation is to ignore edges originating from punctuation marks (1) or not (0).
my $ignore_punctuation_marks = 1;

my $punctuation_mark = qr /[',.:;!?]/;

my $word_characters = qr /[\w']|$punctuation_mark/o;

# Note that the $word_characters regex string determines which characters make permissible word characters in 
# a sentence. It does NOT define whether punctuation marks are to be included in the evaluation since ignoring
# punctuation marks in the regex would lead to mismatches in word counts between the script and the 
# input annotation file. The decision of whether to skip a punctuation mark is therefore included in the extraction 
# step prior to incrementing the evaluation counters.

# Define which levels and labels in the CDG annotation file are to be considered in the evaluation. 
# '0' means to consider only the levels specified below, '1' means to consider all levels.
	my $consider_all_levels = 0;

# If this variable is set to '1' the comparison will be done over all levels, otherwise only over the levels specified subsequently.
	my %levels_considered = (	"SYN"	=>	1,
# 								"REF"	=>	1,
# 								"AGNT"	=>	1,
# 								"THME"	=>	1, 
# 								"INST"	=>	1,
							);
					

my $label_characters = qr/[\w']/;	# This regex string determines whether empty labels are also included in the evaluation.

my $consider_all_labels = 1;
# If this variable is set to '1' the comparison will be done over all labels, otherwise only over the labels specified in the hash below.

my %labels_considered = (	"SUBJ"			=>	1,
							"OBJA"			=>	1,
							"OBJD"			=>	1,
							"ETH"			=>	1, 
							"AGENT"			=>	1,
							"THEME"			=>	1,
							"INSTRUMENT"	=>	1,
							"COMITATIVE" 	=>	1,
							"OWNER"			=>	1,
							"RECIPIENT"		=>	1,
						);

# Create a hash to mark all sentences in the van Kampen corpus which have been identified as globally genitive-dative ambiguous.
my %ambiguous_sentence;

# These ambiguous sentence numbers hold for the van Kampen corpus with 380 sentences.
# $ambiguous_sentence{ $_ } = 1 for ( 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 14, 15, 16, 88, 89, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 114, 115, 144, 145, 147, 148, 149, 150, 151, 152, 153, 155, 156, 172, 173, 174, 175, 176, 177, 178, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 191, 192, 193, 194, 195, 196, 197, 226, 227, 228, 229, 231, 232, 233, 234, 236, 237, 242, 248, 255, 256, 257, 258, 259, 260, 261, 262, 263, 271, 278, 279, 285, 291, 293, 296, 299, 301, 308, 312, 313, 314, 318, 319, 323, 343, 344, 345, 346, 347, 349, 350, 351, 352, 353, 354, 355, 356, 357, 358, 360, 378, 379, 380 );

# These ambiguous sentence numbers hold for the truly globally ambiguous sentences in teh van Kampen corpus with 337 sentences.
$ambiguous_sentence{ $_ } = 1 for qw( 3 6 7 8 10 11 15 89 91 92 93 94 95 96 98 100 102 103 104 105 106 107 109 111 114 145 147 148 149 151 152 155 156 175 177 183 187 194 226 228 229 232 233 234 236 240 241 242 243 244 245 246 247 252 253 258 263 264 266 268 270 274 276 277 281 283 300 301 302 303 306 307 308 309 310 311 312 313 314 315 317 );

my @words;
my $word;
my $sentence;
my %standard;
my %attaches_to;
my %gold_attaches_to; 
my $wordnumber;

my %test_corpus_levels_extracted;		# This array holds all CDG levels of analysis extracted from the test file.
my @test_sentence_correct_labels;  		# This array contains all labels for which a gold standard annotation was found
my @test_sentence_incorrect_labels;   	# This array contains all labels for which no gold standard annotation was found
my @test_corpus_correct_regents;
my @test_sentence_incorrect_regents;
my %standard_sentence_total_label_count;	# This hash contains the number of labels assigned to in the gold standard for a given sentence.
my $standard_corpus_total_label_count;	# This scalar contains the number of labels assigned in the gold standard.
my %standard_corpus_levels_extracted;	# This keys in this hash are the labels assigned in the gold standard.
my $standard_sentence_count;
my %standard_sentences;

my @sentences_in_gold_standard;  	# This array contains all sentences for which a gold standard annotation was found
my @sentences_no_gold_standard;   	# This array contains all sentences for which no gold standard annotation was found

my $test_corpus_total_label_count = 0;
my $test_corpus_correct_label_count = 0;
my $test_corpus_incorrect_label_count = 0;
my $test_corpus_undecided_label_count = 0;
my $test_corpus_correct_regent_count = 0;
my $test_corpus_incorrect_regent_count = 0;
my $test_corpus_undecided_regent_count = 0;
my $test_sentence_total_label_count;
my $test_sentence_start_line;
my %test_sentences;
my $test_sentence_count;
my $sentence_mismatch_count;

my $number;
my $level;
my $regent;
my $element;
my $print = 0;

# my $syn;
# my $ref;

### Hash the gold standard's attachments to be considered in the comparison.

my $gold_standard = shift @ARGV;
open GOLD, "<", $gold_standard;
open GOLD_SENTENCES, ">", "./Sentences.in.Gold.Standard";

# open DEBUG, ">", "/home/mccrae/Desktop/Debug";

while ( <GOLD> )
{	
	# Skip empty lines or lines which are uninteresting from an annotation point of view .
	if ( /^$/o || /wordgraph\d/o )	# Skip further processing of this line for uninteresting annotation header lines.
	{
		$standard_sentence_count ++;
		next;
	}
	chomp;
		
	if ( /^\s+\d+ (\d+) ($word_characters+)/o ) 
	{	
		$wordnumber = $1;
		$word = $2;
		$word="\u$word" if ( $wordnumber == 1 );		# Capitalise the first word in the sentence.
		push @words, $word; 
		next;
	}
	elsif ( /([A-Z]{3,4}) -> ($label_characters+) -> (\d+)/o ) 	# $1 captures the level, $2 captures the label and $3 captures the regent's word number. 
	{
		next unless ( ( $consider_all_levels || $levels_considered{ $1 } ) && ( $consider_all_labels || $labels_considered{ $2 } ) );  # Only hash labels to be considered.
		
		# print GOLD_SENTENCES "Punctuation Mark in Gold Standard: $word \n" if (  $word =~ /$punctuation_mark/ );
		
		next if ( $ignore_punctuation_marks && $word =~ /$punctuation_mark/ );
		
		$attaches_to{ $wordnumber }{ $1 }{ $3 } = $2;
		# print "Gold Standard, Sentence $standard_sentence_count: Extracted word $wordnumber $words[ $wordnumber -1 ] on level $1 attaching with label $2 to word $3.\n";;
		$standard_corpus_total_label_count ++;
		$standard_corpus_levels_extracted{ $1 } ++;

		next;
	} 
	elsif ( /^;$/o ) 
	{
		$sentence = join " ", @words;
		push @words, "ROOT";		# This adds the word "ROOT" as last word number in the sentence.
		
		die "Standard Sentence $standard_sentence_count \n$sentence\nin line $. is a multiple occurrence of sentence $standard_sentences{ $sentence } in the Gold Standard." if ( $standard_sentences{ $sentence } ); 
		$standard_sentences{ $sentence } = $standard_sentence_count;
							
		for $number ( 1 .. scalar @words - 1 )				# Iterate over all words in the sentence.
		{			
			for $level ( keys %{ $attaches_to{ $number } } )	# Iterate over all levels for a given word in the sentence.
			{
				for $regent ( keys %{ $attaches_to{ $number }{ $level } } )	# For each word and level iterate over all its regents (should be just one).
				{
					my $label = $attaches_to{ $number }{ $level }{ $regent };
					$standard{ $sentence }{ $number }{ $level }{ $regent } = $label;
					$standard_sentence_total_label_count{ $sentence } ++;
					# print "Gold Standard: Hashed word $number $words[ $number -1 ] on level $level attaching with label $label to word $regent $words[ $regent -1 ].\n";;
					# print "In the gold standard sentence \n", $sentence, "\nword $number $words[ $number -1 ] on level $level attaches to word number $regent $words[ $regent -1 ] via label $label.\n\n"; 
				}
			}
		}
		
#		printf "%-42s%s", "Sentence $standard_sentence_count  after hashing the gold standard:", "$sentence.\n";
#  		printf "%-42s%s", "Words in gold standard sentence:", "@words\n";
# 		print "Number of gold standard labels in this sentence: $standard_sentence_total_label_count{ $sentence }.\n\n";  	

		@words = ();
 		print GOLD_SENTENCES $sentence, "\n";
		$sentence = "";        
		%attaches_to = ();
	}		
}

my %standard_by_number = reverse %standard_sentences;


# print "Number of labels in gold standard: $standard_corpus_total_label_count.\n";  	

##

##	Now go through the test file(s) checking line by line whether their annotations are contained in 
##	the gold standard and if so, whether they match.

my $regent_counter;

while ( <> )
{
	# next if /^$/;	# Skipping empty lines will prevent the detection of eof in case multiple files are evaluated against the gold standard.
	
	chomp;
		
	if ( /wordgraph/o )
	{
		$test_sentence_count ++;
		$test_sentence_start_line = $.;
		next;		# Skip further processing of this line for uninteresting annotation header lines.
	}		
	elsif ( /^\s+\d+ (\d+) ($word_characters+)/io ) 	# $1 captures the word number, $2 captures the word string.
	{	
		$wordnumber = $1;
		$word = $2;
		$word="\u$word" if ( $wordnumber == 1 );		# Capitalise the first word in the sentence.
		push @words, $word;
		next;
	}
	elsif ( /([A-Z]{3,4}) -> ($label_characters+) -> (\d+)/o ) 	# $1 captures the level, $2 captures the label and $3 captures the regent's word number. 
	{
		next unless ( ( $consider_all_levels || $levels_considered{ $1 } ) && ( $consider_all_labels || $labels_considered{ $2 } ) );  # Only labels to be considered are added to the hash.
		
		# print GOLD_SENTENCES "Punctuation Mark in Test File: $word \n" if (  $word =~ /$punctuation_mark/ );
		
		next if ( $ignore_punctuation_marks && $word =~ /$punctuation_mark/ );
		
		$test_corpus_levels_extracted{ $1 } ++;
		$attaches_to{ $wordnumber }{ $1 }{ $3 } = $2;
		$test_sentence_total_label_count ++;
		$test_corpus_total_label_count ++;
		# print "Test: Extracted word $wordnumber $words[ $wordnumber -1 ] on level $1 attaching with label $2 to word $3.\n";;
		next;  		
	} 
	elsif ( /^;$/o ) 
	{
		# print "\n\nThese are the words in the sentence: @words.\n\n";
			
		$sentence = join " ", @words;
		push @words, "ROOT";	# This adds the word "ROOT" as last word number in the sentence.
					
		die "Test sentence $test_sentence_count \n$sentence\nin line $. is a multiple occurrence of sentence $test_sentences{ $sentence } in the test file." if ( $test_sentences{ $sentence } ); 
		$test_sentences{ $sentence } = $test_sentence_count;
		
		if ( $standard{ $sentence } )	# Check if the current sentence is contained in the gold standard.
		{
			push @sentences_in_gold_standard, $sentence;
		
			delete $standard_by_number{ $standard_sentences{ $sentence } };


			# Now check all labels in the test sentence against the gold standard.
			
			for $number ( 1 .. @words - 1 )	# Iterate over all words for a given sentence ( -1 because of words[0] = "ROOT").
			{				
				for $level ( keys %{ $attaches_to{ $number } } )	# Iterate over all levels of analysis for a given word in a sentence.
				{	
					# print "Hash attaches_to has the following ", scalar keys %{ $attaches_to{ $number } }, " level:\n", join "\n", keys %{ $attaches_to{ $number } }, "\n";					

					for $regent ( keys %{ $attaches_to{ $number }{ $level } } )
					{
						# print "On level $level hash attaches_to has the following ", scalar keys %{ $attaches_to{ $number }{ $level } }, " regent(s):", join "\n", keys %{ $attaches_to{ $number }{ $level } }, "\n";					
						
						my @gold_regents = keys %{ $standard{ $sentence }{ $number }{ $level } }; 
						my $gold_regent = shift @gold_regents;
						
						my $label = $attaches_to{ $number }{ $level }{ $regent }; 
						
						my $gold_label = $standard{ $sentence }{ $number }{ $level }{ $gold_regent };
						
						if ( defined $gold_label && $regent == $gold_regent && $label eq $gold_label ) # Check for correct structure and correct label.
						{
							push @test_sentence_correct_labels, $label;
							$test_corpus_correct_label_count ++;
							
							push @test_corpus_correct_regents, $regent;
							$test_corpus_correct_regent_count ++;
						}
						elsif ( defined $gold_regent && $regent eq $gold_regent )	# Check for correct structure only.
						{
							push @test_corpus_correct_regents, $regent;
							$test_corpus_correct_regent_count ++;
							
							$test_corpus_incorrect_label_count ++;
							push @test_sentence_incorrect_labels, $label;   	
							
							$print = 1;
						}
						elsif ( $ambiguous_sentence{ $test_sentence_count } && defined $gold_label && ( $label eq "OBJD" | $label eq "RECIPIENT" ) && ( $gold_label eq "GMOD" | $gold_label eq "OWNER" ) && $regent == @words - 2 && $gold_regent == $number -2 )
						{
							# For gen-dat ambiguous sentences accept the systematically (sic!) incorrect label and structure if an OBJD 
							# is found in test instead of GMOD in gold standard. Attachment position varies systematically with the current
							# word position: OBJD always attaches to the second last word in the sentence (the last word being the
							# full-stop) while GMOD always attaches to the noun located words earlier in the sentence.

							# print "---------------\n\nThe encountered sentence $test_sentence_count in line $test_sentence_start_line of $ARGV:\n$sentence\nis genitive-dative ambiguous.\n\n";
							# print "Found OBJD in test file instead of GMOD in gold standard.\n";
							# print "\nTest label towards word $regent: $label \nGold label towards word $gold_regent: $gold_label \n";
							
							push @test_sentence_correct_labels, $label;
							$test_corpus_correct_label_count ++;
							
							push @test_corpus_correct_regents, $regent;
							$test_corpus_correct_regent_count ++;
						}
						elsif ( $ambiguous_sentence{ $test_sentence_count } && defined $gold_label && ( $label eq "GMOD" | $label eq "OWNER" ) && ( $gold_label eq "OBJD" | $gold_label eq "RECIPIENT" ) && $regent == $number -2 && $gold_regent == @words - 2 )
						{
							# For gen-dat ambiguous sentences accept the systematically (sic!) incorrect label and structure if a GMOD
							# is found in test instead of an OBJD in gold standard. Attachment position varies systematically with the current
							# word position: OBJD always attaches to the second last word in the sentence (the last word being the
							# full-stop) while GMOD always attaches to the noun located two words earlier in the sentence.

							# For gen-dat ambiguous sentences admit systematically incorrect label and structure if GMOD is found in test instead of OBJD in gold standard.
							# print "---------------\n\nThe encountered sentence $test_sentence_count in line $test_sentence_start_line of $ARGV:\n$sentence\nis genitive-dative ambiguous.\n\n";
							# print "Found GMOD in test file instead of OBJD in gold standard.\n";
							# print "\nTest label towards word $regent: $label \nGold label towards word $gold_regent: $gold_label \n";
							
							push @test_sentence_correct_labels, $label;
							$test_corpus_correct_label_count ++;
							
							push @test_corpus_correct_regents, $regent;
							$test_corpus_correct_regent_count ++;
						}
						else	# Do this is the current label is not present in the gold standard.
						{
							push @test_sentence_incorrect_regents, $regent;
							$test_corpus_incorrect_regent_count ++;
							
							push @test_sentence_incorrect_labels, $label;
							$test_corpus_incorrect_label_count ++;
							
							$print = 1;
						}
														
						if ( $diagnostics && $print ) # Only label assignments deviating from the gold standard will be output.
						{		
							printf "---------------\n\nSentence $test_sentence_count in line $test_sentence_start_line of $ARGV:\n$sentence\n\n" unless $sentence_mismatch_count;
							$sentence_mismatch_count ++;
								
							printf "%-20s %3d %-15s %-16s %2d %s\n", "Test on $level:", $number, $words[ $number -1 ], "--$label-->", $regent, $words[ $regent - 1 ];							
																			
							if ( defined $gold_label )
							{
								printf "%-20s %3d %-15s %-16s %2d %s\n\n", "Standard on $level:", $number, $words[ $number -1 ], "--$gold_label-->", $gold_regent, $words[ $gold_regent - 1 ];
							}
							else
							{
								printf "%-22s No gold standard annotation on for word $number $words[ $number -1 ] on level $level.\n\n", "Standard:";
							}
							
							$print = 0;
						}
					}
				}
			}
		
			if ( $diagnostics & $print )	# Diagnostic output only makes sense for sentences which have a gold standard annotation.
			{	
				print "---------------\n\nSentence $test_sentence_count in line $test_sentence_start_line of 12 $ARGV:\n$sentence\n\n" unless $sentence_mismatch_count;
				# print "The test sentence label(s) \n", ( join "\n", @test_sentence_correct_labels ), "\nmatched the gold standard.\n\n" if ( @test_sentence_correct_labels ); 
				print "The test sentence label(s) \n", ( join "\n", @test_sentence_incorrect_labels ), "\ndid not match the gold standard.\n\n" if ( @test_sentence_incorrect_labels ); 
				print "Test sentence: $test_sentence_total_label_count assigned labels.\n";
				print "Gold standard: $standard_sentence_total_label_count{ $sentence } labels.\n";
				print "Test sentence: ", scalar @test_sentence_correct_labels, " correct labels with correct structure.\n";
				print "Test sentence: ", scalar @test_corpus_correct_regents, " correct structural assignments.\n\n";

				my $sentence_structural_recall = ( scalar @test_corpus_correct_regents ) / $standard_sentence_total_label_count{ $sentence } * 100;
				my $sentence_structural_precision = ( scalar @test_corpus_correct_regents ) / $test_sentence_total_label_count * 100;
				my $sentence_structural_f1_measure = 0;
				$sentence_structural_f1_measure = 2 * $sentence_structural_recall * $sentence_structural_precision / ( $sentence_structural_recall + $sentence_structural_precision ) if ( $sentence_structural_recall && $sentence_structural_precision );
				
				print "Structual\n\n";
				printf "\t%-10s\t%.2f%s", "Recall:", $sentence_structural_recall, "%.\n"; 
				printf "\t%-10s\t%.2f%s", "Precision:", $sentence_structural_precision, "%.\n"; 
				printf "\t%-10s\t%.2f%s", "F1 Measure:", $sentence_structural_f1_measure , "%.\n\n"; 


				my $sentence_labelled_recall = ( scalar @test_sentence_correct_labels ) / $standard_sentence_total_label_count{ $sentence } * 100;
				my $sentence_labelled_precision = ( scalar @test_sentence_correct_labels ) / $test_sentence_total_label_count * 100;
				my $sentence_labelled_f1_measure = 0;
				$sentence_labelled_f1_measure = 2 * $sentence_labelled_recall * $sentence_labelled_precision / ( $sentence_labelled_recall + $sentence_labelled_precision ) if ( $sentence_labelled_recall && $sentence_labelled_precision );
				
				print "\nLabel-based\n\n";
				printf "\t%-10s\t%.2f%s", "Recall:", $sentence_labelled_recall, "%.\n"; 
				printf "\t%-10s\t%.2f%s", "Precision:", $sentence_labelled_precision, "%.\n"; 
				printf "\t%-10s\t%.2f%s", "F1 Measure:", $sentence_labelled_f1_measure , "%.\n\n"; 
				$print = 0;
			}
		}					
		else
		{
			push @sentences_no_gold_standard, $sentence."\n";
			$test_corpus_undecided_label_count = $test_corpus_undecided_label_count + $test_sentence_total_label_count;
			# print "Undecideable sentence: $sentence \n\n Number of labels: $test_sentence_total_label_count";
		}
		
		# Now clear all temporary data containers for the next sentence.
			
		@words = ();
		$sentence = "";
		%attaches_to = ();           
		@test_sentence_correct_labels = ();
		@test_sentence_incorrect_labels = ();
		@test_corpus_correct_regents = ();
		@test_sentence_incorrect_regents = ();
		$test_sentence_total_label_count = 0;
		$sentence_mismatch_count = 0;
	}   
	
	if ( eof ) # Actions in case several annotation files are to be evaluated against the gold standard.
	{	
		print "---------------\n---------------\n\n";
		printf "%-15s %s\n", "Gold Standard:", $gold_standard;
		printf "%-15s %s\n\n", "Test File:", $ARGV;
				
		print "Regular Expression for Word Extaction:  $word_characters \n";
		print "Regular Expression for Label Extaction: $label_characters \n";
		printf "Punctuation marks ignored: %s\n\n", $ignore_punctuation_marks ? "yes" : "no";
		
		printf "%5d %-37s %s.\n", $standard_corpus_total_label_count, "gold standard labels on levels ", join " ", sort keys %standard_corpus_levels_extracted;
		printf "%5d %-37s %s.\n", $test_corpus_total_label_count, "assigned test corpus labels on levels", join " ", sort keys %test_corpus_levels_extracted;
		printf "%5d edges with correct labels and structure.\n", $test_corpus_correct_label_count;
		printf "%5d edges with correct structure.\n", $test_corpus_correct_regent_count;
		printf "%5d incorrect edges with either incorrect structure or incorrect label.\n", $test_corpus_incorrect_label_count;
		printf "%5d undecided edges.\n\n", $test_corpus_undecided_label_count;
		
		# Omit the full-stop at the end of these sentences for better result pastability into MS Excel.
		# Note that the following measures are calculated over all test sentences against the entire gold standard corpus!
		# Hence recall will go down if the test corpus is only a subset of the gold standard corpus.
		
		my $test_corpus_structural_recall = $test_corpus_correct_regent_count / $standard_corpus_total_label_count * 100;
		my $test_corpus_structural_precision = $test_corpus_correct_regent_count / $test_corpus_total_label_count * 100;
		my $test_corpus_structural_f1_measure = 0;
		$test_corpus_structural_f1_measure = 2 * $test_corpus_structural_recall * $test_corpus_structural_precision / ( $test_corpus_structural_recall + $test_corpus_structural_precision ) if ( $test_corpus_structural_recall && $test_corpus_structural_precision );
		
		print "Structural\n\n";
		printf "\t%-10s\t%.2f %s", "Recall:", $test_corpus_structural_recall, "%\n"; 		
		printf "\t%-10s\t%.2f %s", "Precision:", $test_corpus_structural_precision, "%\n"; 
		printf "\t%-10s\t%.2f %s", "F1 Measure:", $test_corpus_structural_f1_measure, "%\n\n"; 
		
		my $test_corpus_labelled_recall = $test_corpus_correct_label_count / $standard_corpus_total_label_count * 100;
		my $test_corpus_labelled_precision = $test_corpus_correct_label_count / $test_corpus_total_label_count * 100;
		my $test_corpus_labelled_f1_measure = 0;
		$test_corpus_labelled_f1_measure = 2 * $test_corpus_labelled_recall * $test_corpus_labelled_precision / ( $test_corpus_labelled_recall + $test_corpus_labelled_precision ) if ( $test_corpus_labelled_recall && $test_corpus_labelled_precision );
		
		print "\nLabel-based\n\n";
		printf "\t%-10s\t%.2f %s", "Recall:", $test_corpus_labelled_recall, "%\n"; 		
		printf "\t%-10s\t%.2f %s", "Precision:", $test_corpus_labelled_precision, "%\n"; 
		printf "\t%-10s\t%.2f %s", "F1 Measure:", $test_corpus_labelled_f1_measure, "%\n\n"; 
	
		print "====================================\n";
	
		# print "\nThe following ", ( scalar @sentences_in_gold_standard ), " sentence(s) had a gold standard annotation:\n\n", ( join "\n\n", @sentences_in_gold_standard ), "\n\n====================================\n" if ( scalar @sentences_in_gold_standard );	
		if ( scalar @sentences_no_gold_standard )
		{ 
			print "\nThe following  ", ( scalar @sentences_no_gold_standard ), " sentence(s) had no gold standard annotation:\n\n", ( join "\n\n", @sentences_no_gold_standard ), "\n\n====================================\n";	
		}
		else
		{
			print "\nAll sentences in the test file had a gold standard annotation.\n\n";
		}
		print "====================================\n";					
		
		if ( keys %standard_by_number )
		{ 
			print "\nThe following  ", ( scalar keys %standard_by_number ), " sentence(s) in the gold standard were not matched by an annotation in the test file:\n\n", ( join "\n\n", values %standard_by_number ), "\n\n====================================\n";	
		}
		else
		{
			print "\nAll sentences in the gold standard were matched by an annotation in the test file.\n\n";
		}
		print "====================================\n\n";			


		# Now clear the counters specific to each test corpus.
		
		$test_corpus_total_label_count = 0;
		$test_corpus_correct_label_count = 0;
		$test_corpus_incorrect_label_count = 0;
		$test_corpus_correct_regent_count = 0;
		$test_corpus_incorrect_regent_count = 0;
		@sentences_in_gold_standard =();  	
		@sentences_no_gold_standard = ();
		%test_corpus_levels_extracted = ();
		$. = 0;
		$test_sentence_count = 0;
		$test_corpus_correct_regent_count= 0;
		$test_corpus_undecided_label_count = 0;
	}
}