#! /usr/bin/perl -w

use strict;
use locale;

#	This script extracts the thematic relations from a CDG annotation file. 
#	The extraction result is made available in the following hash:
#	has_ROLE{ sentence_number }{ dependant_word_number } = [ ( ROLE, regent_word_number ) ].
#
# 	Invocation parameters:
#	======================
#   1. File(s) containing the CDG2 annotations from which to extract the thematic roles.

my $punctuation_mark = qr /[',.:;!?]/;
my $word_characters = qr /[\w']|$punctuation_mark/o;
my $label_characters = qr /[A-Z]/;
my $feature_characters = qr /[\w_]/;

# Note that these character definitions determine which characters make permissible word
# or label characters in a sentence.

# Define which levels and labels in the CDG annotation file are to be considered in the evaluation. 
# '0' means to consider only the levels specified below, '1' means to consider all levels.
my $consider_all_levels = 1;

# If this variable is set to '1' the comparison will be done over all levels, otherwise only over the levels specified subsequently.
my %levels_considered = (	"SYN"	=>	1,
# 								"REF"	=>	1,
# 								"AGNT"	=>	1,
# 								"THME"	=>	1, 
# 								"INST"	=>	1,
							);
					
my $consider_all_labels = 0;
# If this variable is set to '1' the comparison will be done over all labels, otherwise only over the labels specified in the hash below.

my %labels_considered = (	
							"AGENT"		=>	1,
							"THEME"		=>	1,
							"INSTRUMENT"	=>	1,
							"COMITATIVE" 	=>	1,
							"OWNER"		=>	1,
							"RECIPIENT"		=>	1,
						);

my @words;
my %word_string;
my %sentence;
my %has_ROLE;
my %all_roles;
my $word_number;
my $sentence_number;
my %feature;
my %verb_forms;
my %sem_val;


while ( <> )
{		
	# Skip empty lines or lines which are known to be uninteresting from an annotation point of view.
	
	if ( /^$/o || /wordgraph/o )	
	{
		$sentence_number ++ if /wordgraph/o;	# Lines containing 'wordgraph' introduce a new sentence annotation.
		next;
	}
	
	chomp;
		
	if ( /^\s+\d+ (\d+) ($word_characters+)/o ) # $1 captures the word number in the current sentence, $2 captures the word string. 
	{	
		$word_number = $1;
		$word_string{ $sentence_number }{ $word_number } = $2;
		push @words, $word_string{ $sentence_number }{ $word_number }; 
		next;
	}
	elsif ( /([A-Z]{3,4}) -> ($label_characters+) -> (\d+)/o ) 	# $1 captures the level, $2 captures the label and $3 captures the regent's word number. 
	{
		next unless ( ( $consider_all_levels || $levels_considered{ $1 } ) && ( $consider_all_labels || $labels_considered{ $2 } ) );  # Only hash labels to be considered.
		
		$all_roles{ $sentence_number }{ $2 } ++;
		
		# print "Sentence number $sentence_number: Extracted role: $2 with count $all_roles{ $sentence_number }{ $2 } \n";
		
		$has_ROLE{ $sentence_number }{ $word_number } = [ ( $2, $3 ) ];
		
		next;
	}
	elsif ( /^($feature_characters+) \/ (.+)$/o )	# $1 captures the feature name, $2 captures the feature value.
	{	
		$feature{ $sentence_number }{ $word_number }{ $1 } = $2;
		
		if ( $2 =~ /(VVFIN)|(VVPP)|(VAFIN)|(VAPP)|(VMFIN)|(VMPP)/o )	# If the POS is one of the listed, remember this word as a verb form.
		{
			push @{ $verb_forms{ $sentence_number } }, $word_number;   
			# print "$word_string{ $sentence_number }{ $word_number } is a verb form.\n";
		}
 	}
	elsif ( /^;$/o ) 
	{
		$sentence{ $sentence_number } = join " ", @words;
		# print "$sentence{ $sentence_number }\n";
				
		@words = ();
	}		
}

# &print_all_roles();

&print_all_relations();


sub print_all_roles()
{
	for my $count ( 1 .. $sentence_number )
	{	
		for my $key ( keys %{ $all_roles{ $count } } )
		{
			print "Sentence $count : $key\n";
		}	
		print "\n";
	}
}

sub print_all_relations
{
	# 	The following loop construct prints out all thematic relations extracted from the annotation file
	# 	to illustrate that the data structures work as required.
		
	for my $counter ( 1 .. $sentence_number )	# Iterate over all sentences.
	{	
		print "\n\n", $counter, "\t", $sentence{ $counter }, "\n\n";
		
		# First, find the sem_vals for all verb forms.
	
		for my $verb ( @{ $verb_forms{ $counter } } )
		{	
			$sem_val{ $counter }{ $verb } = &find_sem_val( $counter, $verb );
			
			printf "\t%-15s %s", $word_string{ $counter }{ $verb }, $sem_val{ $counter }{ $verb };
		}
		
		
		# 	Now generate output for all words which are involved in a thematic relation.

		for my $word_number ( sort { $a <=> $b } keys %{ $has_ROLE{ $counter } } )	# Iterate over the numercially sorted word numbers for which a thematic annotation was found.
		{
			my $dependant_word_string = $word_string{ $counter }{ $word_number };
			my $thematic_relation = ( shift @{ $has_ROLE{ $counter }{ $word_number } } )."_for";
			my $regent_word_string = $word_string{ $counter }{ shift @{ $has_ROLE{ $counter }{ $word_number } } };
			printf "\t%-15s is_%-15s %s \n", $dependant_word_string, $thematic_relation, $regent_word_string;
		}
	}
}


sub find_sem_val
{
	# This subroutine determines the semantic valence based on an 'egrep' on the deutsch-lexikon.cdg file.
	# Note that the result may not be unambiguous because the egrep is only base on the lexical features
	# provided in the annotation file. These may not be sufficient to achieve full disambiguation of all
	# homonyms. Clearly, not being able to determine the semantic valence unambiguously in all cases is
	# an issue when trying to assign the correct lexicalisation to the homonym.
	
	# The better approach hence would be to assign lexicalisation based on the thematic relations actually
	# encountered in the sentence annotation as it is done in the PPC's Word.class!
	
	( my $sentence_number, my $word_number ) = @_;
	
	my @statements;
	
	for my $feature ( sort keys %{ $feature{ $sentence_number }{ $word_number } } )
	{
		push @statements, "egrep \"$feature:$feature{ $sentence_number }{ $word_number }{ $feature }\"";
		
		# print "$feature:$feature{ $sentence_number }{ $word_number }{ $feature }\n";
	}
	$statements[ 0 ] = $statements[ 0 ] . " ~/cdg2/grammar/deutsch/deutsch-lexikon.cdg";	# Add the lexicon file to egrep through.
	
	push @statements, " egrep \"$word_string{ $sentence_number }{ $word_number }\"";
	push @statements, " egrep -o \"sem_val:([a-z_]+)\"";
	
	my $command = ( join " | ", @statements );	# Construct the expression as a sequence of egrep "..." | egrep "..." statements.
	$command =~ s/ \| $//o; 
	
	# print $command, "\n";
	
	my $sem_val = `$command`;
	$sem_val =~ s/sem_val://;
	
	return $sem_val;	
}