package cia_components.ppc;

import java.util.*;

import org.semanticweb.owl.model.OWLClass;
import java.util.regex.*;

public class Word
{
	// This class represents a single homonym with all its properties in a slot of the input sentence.
	int position_in_slot;
	int attaches_to;
	Slot parent_slot;
	String ppc_integration_mode;
	
	Set <OWLClass> compatible_conceptualisations;
	
	HashMap <String, Object> word_features;
	// This HashMap contains all the attributes and values for this word - both as extracted from the lexicon and as constructed in PPC classes.
	
	HashMap <String, Float> predictions;
	/* 	
		This HashMap contains all PPC predictions for this word object as key-value pairs. Keys take the format RELATION_description where 
		'RELATION' is the name of the thematic relation and 'description' is the description string of the homonym to which the word object 
		establishes its thematic relation. The hashmap will be filled by the Score_Assigner class.
	*/
	String prediction_string = "";
	String requested_info = System.getenv( "REQUESTED_INFO" );
	float plausibility_score;
	float dummy_score;
	
	// ================================ Start Constructor Definitions
			
	public Word( Slot parent, int position, String input_line )
	{	
		word_features = new HashMap <String, Object> ();
		
		set_Parent_Slot( parent );
		set_Position_in_Slot( position );
		tokenize_Input_Line_From_CDG_Main( input_line );
		
		predictions = new HashMap <String, Float> ();
		
		initialise_Conceptualisations();
	}
	
	
	/** This method adds a given concept to the set of conceptualisations
	 *  which denote all possible readings of the word represented by this
	 *  word object.
	 *  
	 * @param concept is the concept to be added to the set of concepts for 
	 * this word object.
	 */
	
	public void add_Conceptualisation( OWLClass addendum )	
	{
		Set <OWLClass> set = (Set <OWLClass>) word_features.get( "conceptualisations" );
		set.add( addendum );
		word_features.put( "conceptualisations", set );

		// System.err.println( "Concept '" + addendum + "' has been added to the list of conceptualisations of word '" + word_features.get( "description" ) + "': " + set + "." );
	}

	
	/** This method adds a given set of concepts to the set of conceptualisations
	 *  which denote all possible readings of the word represented by this
	 *  word object.
	 *  
	 * @param <code>set</code> is the set of concepts to be added to the set of concepts for 
	 * this word object.
	 */
	
	public void add_Conceptualisations( Set <OWLClass> addenda )	
	{
		Set <OWLClass> set = (Set <OWLClass>) word_features.get( "conceptualisations" );
		set.addAll( addenda );
		word_features.put( "conceptualisations", set );

//		System.err.println( "Concept set '" + set + "' has been added to the list of conceptualisations of word '" + word_features.get( "description" ) + "'." );
	}
	
	
	public Set <OWLClass> get_All_Word_Conceptualisations()
	{
		return (Set <OWLClass>) word_features.get( "conceptualisations" );
	}
	
	
	/** This method sets an initially empty set of conceptualisations as a value
	 * 	in the word object's hashmap 'word_features' under key 'conceptualisations'.
	 */
	
	private void initialise_Conceptualisations()
	{
		Set <OWLClass> set_1 = new HashSet<OWLClass> ();
		word_features.put( "conceptualisations", set_1 );		

		word_features.put( "compatible_conceptualisations", new HashSet<OWLClass> () );		
		// This hashmap maps the word's contextually compatible concepts onto the 
		// set of context concepts with which it is compatible.
		
		compatible_conceptualisations = new HashSet <OWLClass> ();
	}


	// ================================ End Constructor Definitions
	
	
	// ================================ Start Method Definitions
		
	private void tokenize_Input_Line_From_CDG_Main( String line )
	{
		/* 	
			This method tokenizes the input line received from CDG via the parent slot
			and writes the tokens as attribute-value pairs of the HashMap 'word_features'. 
			Each line received is of the format (without the whitespaces):
			word \t word_position_in_input_sentence \t CDG's_unique_description_string \t (attribute \t value)*
		*/
		ppc_integration_mode = System.getenv( "PPC_INTEGRATION_MODE" );
		
		// System.out.println( "PPC Integration Mode: " + ppc_integration_mode );
		
		// Don't bother tokenizing the input line for the ID line received from CDG1 or CDG2.			
		
		if ( parent_slot.get_Slot_Number() == 0 ) 	
		{
			set_Word_String( line );
			parent_slot.set_Slot_String( line );
			// Note that no word objects are created for this slot.
		}
		else if ( ppc_integration_mode.equals( "1" ) | ppc_integration_mode.equals( "2" ) | ppc_integration_mode.equals( "3" ) )
		{
			tokenize_Homonym_Specific_Input( line );
		}
		else if ( ppc_integration_mode.equals( "4" ) )
		{
			tokenize_Slot_Based_Input( line );
		} 
	}	


	private void tokenize_Homonym_Specific_Input( String line )
	{	
//		System.err.println( "Line working on: " + line );
		
		try
		{ 				
			StringTokenizer tokenizer = new StringTokenizer( line );
			// System.err.println( "Tokens counted: " + tokenizer.countTokens() );
			
			String attribute = "";
			String value = "";
			
			String token = tokenizer.nextToken().trim();
			// This token contains the actual word as entered in the input sentence.
			
			set_Word_String( token );
//			System.err.println( "The word_string and first token is: " + get_Word_String() );

			parent_slot.set_Slot_String( token );
			// 	This adds the actual word in the sentence to the parent slot's slot_string.
			
			if ( tokenizer.hasMoreTokens() )
			{		
				token = tokenizer.nextToken().trim();	// This token contains the slot number in the sentence.
				word_features.put( "slot number", token );
				// System.out.println( "Second token (slot number): '" + word_features.get( "slot number" ) );
					
				token = tokenizer.nextToken().trim();	// This token contains the homynm's ID string which is unique inside this slot.
				
				set_Description( token );

				// Now add this unique identifier into the vector of all homonym identifiers in the sentence.
				this.get_Parent_Slot().add_Description( token );
			}
			
			
			/*
				Since an unknown number of additional attribute-value pairs may be received from CDG2, 
				attribute-value pairs are created with an iterator. Note that - unlike for standard 
				predictor input - CDG2 also provides the attribute labels in its input.
			*/	

			while ( tokenizer.hasMoreTokens() )
			{
				attribute = tokenizer.nextToken().trim(); 
//				System.err.println( "attribute: >" + attribute + "<" );
				
				value = tokenizer.nextToken().trim(); 
//				System.err.println( "value: >" + value + "<\n" );
				
				word_features.put( attribute, value );
			}
			
				
//			if ( word_features.keySet().contains( "number" ) )
//			{
//				System.err.println( "Detected number: >" + word_features.get( "number" ) + "< in word '" + this.get_Description() + "'." );
//			}
		}
		catch (Exception e)
		{
			System.err.println( "Word, line 234 - Could not tokenize attribute-value pairs for line \n" + line + "\nreceived from " + System.getenv( "CDG_INVOCATION_COMMAND" ).toUpperCase() + ", because: \n" + e + "\n" );
		}
		
		set_Normalisation();
	}
	

	
	private void set_Description( String token )
	{
		word_features.put( "description", token );
		// System.out.println( "Third token (unique description string): '" + get_Description() + "'" );	
	}


	private void tokenize_Slot_Based_Input( String line )
	{
		// System.err.println( "Line working on: " + line );
		
		try
		{ 				
			StringTokenizer tokenizer = new StringTokenizer( line );
								
			String token = tokenizer.nextToken().trim();
			// This token contains the actual word as entered in the input sentence.
		
			set_Word_String( token );
			// System.err.println( "The word_string and first token is: " + get_Word_String() );

			parent_slot.set_Slot_String( token );
			// 	This adds the actual word in the sentence to the parent slot's slot_string.
								
			// set_Lexical_Base_Form( line );
			// System.out.println( "The word's lexical base form is: " + get_Feature_Value( "base" ) + "\n" );
				
			String attribute = "";
			String value = "";
			
			try
			{
				StringTokenizer info = new StringTokenizer( requested_info );
				
				// System.out.println( "The line contains " + tokenizer.countTokens() + " tokens." );
				// System.out.println( "There are " + info.countTokens() + " attributes to be filled." );
				
				while ( info.hasMoreTokens() )
				{
					attribute = info.nextToken().trim(); 
					// System.out.println( "attribute: '" + attribute + "'" );
						
					value = tokenizer.nextToken().trim(); 
					// System.out.println( "value: '" + value + "'\n" );
					
					word_features.put( attribute, value );
				}
			}
			catch ( Exception e )
			{
				System.err.println( "\nWord Line 262: Could not tokenize the list of additional attributes requested from XCDG2 in PPC integration mode 4 because:\n" + e + "\n" );
			}			
			
		}
		catch (Exception e)
		{
			System.err.println( "Word, line 205 - Could not tokenize attribute-value pairs for line \n" + line + "\nreceived from " + System.getenv( "CDG_INVOCATION_COMMAND" ).toUpperCase() + ", because: \n" + e + "\n" );
		}
	}

	
	/*	private void tokenize_Script_Output_Line( String line )
	{
		/* 	
			This method tokenizes the input line when homonym input is received from a homonym retrieving script. 
			It writes the tokens as attribute-value pairs of the HashMap 'word_features'. Note that it is not possible 
			to deduce from the a Word object's position in the slot whether it is the Word object corresponding to the 
			word deemed to be of the most probable category by CDG's tagger prediction.
			Each line contains a regular lexicon entry for one homonym.
		*/
/*		
		First reduce the line received from the homonym retriever to the feature definition after the ':='.
		Pattern pattern = Pattern.compile( "[ ]?\\'?" + word_string + "\\'?[ ]*:=[ ]*\\[[ ]*" );
		Matcher matcher = pattern.matcher( line );
		line = matcher.replaceAll( "" );
				
		
		Then decompose the line into tokens and fill them into attribute-value pairs of the HashMap 'word_features'.
		StringTokenizer tokenizer = new StringTokenizer( line, " ,;:]\'\t", false );
		
		String attribute = "";
		String value = "";
		
		try
		{ 
			while ( tokenizer.hasMoreTokens() )
			{	
				attribute = tokenizer.nextToken().trim();
				value = tokenizer.nextToken().trim();
				word_features.put( attribute, value );
			}	
		}
		catch ( NullPointerException e )
		{
			System.out.println( "Could not tokenize input line \n" + line + "\nbecause:\n" + e.getMessage() );
		}	
	}	*/
	
	
	
	/** 	
		This method returns the value of the word feature by the name 'attribute'.
	*/
	
	public String get_Feature_Value( String attribute )
	{

		String value = (String) word_features.get( attribute );
		// System.out.println( "Requested feature: " + attribute + "\nValue: " + value );
		return value;
	}
	
		
	public String get_Word_String()
	{
		return (String) word_features.get( "word_string" );
	}

	
	public String get_Grammatical_Number()
	{
		return (String) word_features.get( "number" );
	}

	
	public boolean has_Grammatical_Number()
	{
		return word_features.keySet().contains( "number" );
	}
	
	
	public float get_Plausibility_Score()
	{
		return plausibility_score;
	}
	
	
	public float get_Prediction_Value( String key )
	{
		/*
			This method returns the specific prediction value
			from the hash predictions for a given key.
			The return value will be the actual prediction score 
			assigned in the hash.
		*/
		
		float result = (Float) predictions.get( key );
		return result;
	}
	
		
	public void set_Prediction( String key, Float value )
	{
		predictions.put( key, value );
		// System.out.println( "Setting prediction for " + key + " to " + value );
	}


	public HashMap <String, Float> get_predictions()
	{
		return predictions;
	}


	public int get_Position_in_Slot()
	{
		return position_in_slot;
	}
	
		
	public int get_Slot_Number()
	{
		return parent_slot.get_Slot_Number();
	}
	
	
	public String get_Description()
	{
		return get_Feature_Value( "description" );
	}
	
	
	public void set_Plausibility_Score( float input )
	{
		plausibility_score = input;
	}
	
	
	private void set_Position_in_Slot( int number )
	{
		position_in_slot = number;
	}
	
	
	private void set_Parent_Slot( Slot slot )
	{
		parent_slot = slot;
	}

	
	public Slot get_Parent_Slot()
	{
		return parent_slot;
	}
	
	
	private void set_Word_String( String word )
	{
		word_features.put( "word_string", word );
	}

	
	public void set_Prediction_String( String input )
	{
		prediction_string = input;
	}


	public String get_Prediction_String()
	{
		/*
			This method returns the string containing all predictions 
			for this word object representing a homonym.
		*/
		return prediction_string;
	}
	

	public String get_Lexical_Base_Form()
	{
		return (String) word_features.get( "base" );
	}


	/**
	 * This method computes the syntactic candidate's normalisation. The normalisation
	 * is equal to the lexical base form if it has been set in a word object's lexical entry. 
	 * Otherwise - if no lexical baseform has been set or it takes uninformative values such as 
	 * 'unknown' or '-' - it is set to the word string.
	 * Note that normalisations are not unambiguous as several words in the same sentence may 
	 * have the same normalisation, even if the corresponding word strings themselves are different 
	 * from each other. This is a feature rather than a bug since it allows us to model lexical 
	 * polysemy. 
	 * 
	 * Contrary to the lexicalisations of concepts, which have been modelled as sets of 
	 * lexicalisation strings, the normalisation of a word is assumed to be a single string.
	 * The normalisation is saved in the word object's hashmap 'word_features' under the 
	 * key 'normalisation'.
	 * 
	*/
	
	private void set_Normalisation()
	{	
		/*
		 * Note that not all word classes have a lexical base form in their lexicon entry. It may hence 
		 * be necessary to construct a base form for those entries. For the words that don't have 
		 * a base form entry in the lexicon or whose baseform is 'unknown' or '-' (typically homonyms 
		 * generated from pattern matches in CDG's grammar) the normalisation is set to the word's 
		 * surface string. Capitalisation resulting from sentence-initial occurrence of a word 
		 * must be removed in those cases unless the word's part of speech is NN, NE or FM.
		 */

		// Set a default value for lex and overwrite it in case there is a word feature 'base'.
		
		String lex = "-";	
		
		if ( word_features.containsKey( "base" ) )
		{
			lex = (String) word_features.get( "base" );
		}
		
		if ( lex.equals( "-" ) | lex.equals( "unknown") )
		{
			lex = this.get_Word_String();
		}
		
//		System.err.println( "Attempting to set the normalisation for word >" + this.get_Description() + "< to >" + lex + "<." );

		
		/* 	Remove capitalisation resulting from sentence initial occurrence in the normalisation for:
		 * 
		 * 	(1) words of category other than NN, NE or FM. 
		 * 
		 * 	(2) words which have been created via a pattern match of a lowercase pattern (i.e.: words that have the word 
		 * 	feature 'pattern' and whose pattern name starts with a lowercase letter).
		 * 	
		*/
	
		if ( this.get_Slot_Number() == 1 )
		{
			boolean a = ! ( this.get_Feature_Value( "cat" ).equals( "NN" ) | this.get_Feature_Value( "cat" ).equals( "NE" ) | this.get_Feature_Value( "cat" ).equals( "FM" ) );
			
			String first_letter = this.get_Description().substring( 0, 1 );
//			System.err.println( "The first letter is: " + first_letter );		
			boolean b = this.word_features.keySet().contains( "pattern" ) & first_letter.equals( first_letter.toLowerCase() );

			if ( a | b )
			{
				lex = lex.toLowerCase();
//				System.err.println( "Switched case for " + this.get_Description() + "." );
			}			
		}
											
		// Normalise German umlauts and special characters.

		lex = lex.replaceAll( "ä", "ae" );
		lex = lex.replaceAll( "ö", "oe" );
		lex = lex.replaceAll( "ü", "ue" );
		lex = lex.replaceAll( "Ä", "Ae" );
		lex = lex.replaceAll( "Ö", "Oe" );
		lex = lex.replaceAll( "Ü", "Ue" );		
		lex = lex.replaceAll( "ß", "ss" );
		
		word_features.put( "normalisation", lex );
		
//		System.err.println( "Have set the normalisation for homonym '" + this.get_Description() + "' with sem_val '" + word_features.get( "sem_val" ) + "' to '" + word_features.get( "normalisation" ) + "'.\n" );
	}


	public String get_Normalisation()
	{
		return (String) word_features.get( "normalisation" );
	}
		
	
	public Set <OWLClass> get_Conceptualisations()
	{
		return (Set<OWLClass>) word_features.get( "conceptualisations" );
	}
		
	
	public String get_Sem_Val()
	{
		String result = (String) word_features.get( "sem_val" );
	
		if ( result == null )
		{
			return "";
		}
		
		return result;
	}


	public void set_Contextually_Compatible_Conceptualisations( Set <OWLClass> set )
	{
		// System.out.println( "Word '" + this.get_Description() + "' has " + ( (Set <OWLClass>) word_features.get( "compatible_conceptualisations" ) ).size() + " pre-saved conceptualisations." );
		compatible_conceptualisations = set;
	}

	
	/**
	 * This method returns the set of all concepts that conceptualise the current word.
	 */
	
	public Set <OWLClass> get_Contextually_Compatible_Conceptualisations()
	{
		// System.out.println( "Word '" + this.get_Description() + "' has " + ( (Set <OWLClass>) word_features.get( "compatible_conceptualisations" ) ).size() + " pre-saved conceptualisations." );
		return compatible_conceptualisations;
	
//		System.out.println( "The word '" + get_Description() + "' has " +  ( (Set <OWLClass>) word_features.get( "compatible_sets" ) ).size() + " compatible concepts." );
//		for ( OWLClass concept : (Set <OWLClass>) word_features.get( "compatible_sets" ) )
//		{
//			System.out.println( "Compatible concept: " + concept );
//		}
	}
	
	
	/**
	 * This method adds an OWLClass instantiated in visual context to the set of concepts that 
	 * one of the word's conceptualisations is compatible with.
	 * 
	 * @param word_concept is the word's concept which is compatible with one or more concepts
	 * instantiated in visual context.
	 * 
	 * @param context_concept is the concept instantiated in visual context that the current
	 * word's conceptualisation is compatible with.
	 */
	
	public void add_Contextually_Compatible_Conceptualisation( OWLClass concept )
	{
		compatible_conceptualisations.add( concept );
		// System.out.println( "\nAdded concept '" + concept + " to the list of contextually compatible concepts of word '" + this.get_Description() + "'." );
	}
}