#!/usr/bin/env perl
###############################################################################
# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Author: Michael Daum 
#
# This is ment to be usefull in order to trick itck/itk sources into doxygen,
# by converting itcl into something that resembles c++. So for instance the
# common member declaration in itcl classes is translated to static members,
# itk_options are translated to methods of the options name etc.
# Additionally some extra commands have been added to doxygen mostly because
# tcl is an untyped language whereas c++ is not. But even in tcl you actually
# like to know more about the function arguments and its return values. So only
# for documentational reasons we provide the \type command to enrich the tcl 
# sources a little. 
#
# Most importand the tcl comments beginning with a double hash ## are translated
# into //! which are special doxygen relevant comments. The single hashed comments
# are translated into // left untouched by doxygen. Also line comments are
# translaged this way, so that ...; ## ... becomes ...; //! ...
#
# Here is the list of all extra commands unknown to doxygen
# itself.
#
#
# \type <arg-type>[, <arg-type> ... 
#	Specify the type of the function arguments of procs and methods giving
#	a list of argument types matching the order of the function arguments.
#	The \type command must be in the same line as the function declaration
#	or implementation. Examples:
#	
#	proc hello_world {text} {; ## \type TclString
#	}
#
#	body MyClass::thisMethod {x y args} {; ## \type TclNumber, TclNumber, TclList
#
#	Also this must match the declaration as doxygen is very picky about
#	matching declaration and implementation which tcl isn't.
#
#	In general if you don't specify the type of the function arguments
#	this filter will default to a TclRef default type, whatever that might
#	mean.
#
# \virtual
#	Specify that a method in an itcl class is going to be overwriteable.
#	No such declarations exist in itcl. Everything of a base class can be 
#	overwritten even when the base class gets brocken then
#
# \purevirtual
#	Specify that a method is not implemented in that itcl class making the
#	class abstract. Again no such declarations are needed in itcl. Even a
#	intended abstract class can be instanciated. Not implemented methods
#	can be implemented compiling and arbitrary other source file. Unthinkeable
#	in any other OO language.
#
# \comment
#	translated into #if 0 comment out hughe areas of code. This needs a
#	matching \endcomment for obvious reasons
#
# \endcomment
#	translated into #endif matching a previous \comment
#
###############################################################################
use warnings;
use strict;

my $debug = 0;
my @outQueue;
my %memberTypes;
my $lineNumber = 0;
my $currentLine = "";
my $errorNumber = 0;

###############################################################################
sub printArgs {
  my ($types, $args) = @_;
  
  my @args;
  foreach my $arg (split(/({.+?})|\s+/, $args)) {
    next if ! $arg;
    push  @args, $arg;
  }

  if (scalar(@{$types}) != scalar(@args)) {
    print STDERR "ERROR in $lineNumber: unbalanced number of args and types\n";
    print STDERR "\n" . $currentLine . "\n\n";
    print STDERR "got " . scalar(@args) . " args";
    if (@args) {
      print STDERR ": " .  join(", ", @args) . "\n";
    } else {
      print STDERR "\n";
    }
    
    print STDERR "got " . scalar(@{$types}) . " types";
    if (@{$types}) {
      print STDERR ": " .  join(", ", @{$types}) . "\n\n";
    } else {
      print STDERR "\n";
    }
    $errorNumber++;
    return;
  }

  my $i = 0;
  foreach my $arg (@args) {
    print STDERR "DEBUG: arg=$arg\n" if $debug;
    &printOut(", ") if $i > 0;
    if ($arg =~/{(\S+)}/) {
      &printOut("$$types[$i] $1");
    } elsif ($arg =~ /{(\S+)\s(\S+)}/) {
      &printOut("$$types[$i] $1 = $2");
    } else {
      &printOut("$$types[$i] $arg");
    }
    $i++;
  }
}

###############################################################################
sub printOut {
  my $message = shift;
  my $item;

  while (defined($item = pop(@outQueue))) {
    #print STDERR "DEBUG: printOut $item\n" if $debug;
    print $item;
  }
  #print STDERR "DEBUG: printOut $message\n" if $debug;
  print $message;
}

###############################################################################
sub pushOut {
  my $item = shift;
  push @outQueue, $item;
}

###############################################################################
sub popOut {
  my $pattern = shift;
  my @newQueue;


  foreach my $item (@outQueue) {
    next if $item =~ /$pattern/;
    push @newQueue, $item;
  }

  @outQueue = @newQueue;
}


###############################################################################
sub main {
  my $scopeDecl = "";
  my $className = "";
  my @argTypes;

  while (my $line = <>) {
    chomp($line);
    $lineNumber++;
    $currentLine = $line;

    # skip empty lines
    next if $line =~ /^\s*$/;

    # comment conversion
    $line =~ s/^\s*##\s*--+\s*$//;
    $line =~ s/^(\s*)##/$1\/\/!/;
    $line =~ s/^(\s*)#[^!]/$1\/\//;
    $line =~ s/;(\s*)##/;$1\/\/!/; 
    $line =~ s/;(\s*)#/;$1\/\//; 

    # preprocessor comment
    $line =~ s/^\s*\/\/!\s*\\comment\b.*$/#if 0/;
    $line =~ s/^\s*\/\/!\s*\\endcomment\b.*$/#endif/;

    # keep track of a class definition
    if ($line =~ /^class ([a-zA-Z0-9_]*) { *$/) {
      $className = $1;
      &pushOut("class $className {\n");
      next;
    }

    # now, print the class definition and inheritance (public by default)
    if ($line =~ /^\s*inherit\s+(.*)/) {
      &popOut("^class");
      my $classDefinition = "class $className : public ";
      $classDefinition .= join " , public ", split /\s+/, $1;
      $classDefinition .= " {\n";
      &pushOut($classDefinition);
      next;
    }

    # keep track of the declaration scope
    if ($line =~ /^\s*(public|protected|private) (method|variable|proc)/) {
      if ($1 ne $scopeDecl) {
	$scopeDecl = $1;
	&printOut("$scopeDecl:\n");
      }
    }

    # get type definitions, but don't store them yet into memberTypes
    # as we don't have the member name yet
    if ($line =~ s/(.*);\s*\/\/!.*\\type\s*([^\\]+)/$1/) {
      @argTypes = split(/\s*,\s*/, $2);
    } else {
      @argTypes = ();
    }

    # constructor declaration
    if ($line =~ /^(\s*)constructor\s*{(.*)}\s*{}/) {
      $memberTypes{"constructor"} = [ @argTypes ];
      &printOut("$1$className(");
      &printArgs(\@argTypes, $2);
      &printOut(");\n");
      next;
    }
    
    # constructor implementation
    if ($line =~ /^body ${className}::constructor\s*{(.*)}\s*{/) {
      my $args = $1;
      &printOut("${className}::${className}(");
      if (defined $memberTypes{"constructor"}) {
	&printArgs($memberTypes{"constructor"}, $args);
      } else {
	&printArgs(\@argTypes, $args);
      }
      &printOut(") {\n");
      next;
    }
    
    # destructor declaration
    $line =~ s/^(\s*)destructor\s*{}\s*$/$1~$className();/;

    # destructor implementation
    $line =~ s/^body ${className}::destructor\s*{}\s*{\s*$/${className}::~$className() {/;

    # method declaration
    if ($line =~ /^(\s*)$scopeDecl\s+method\s+([a-zA-Z0-9_]*) +{(.*)}/) {
      my $indent = $1;
      my $methodName = $2;
      my $args = $3;
      my $isPureVirtual = 0;
      
      $memberTypes{$methodName} = [ @argTypes ];
      &printOut($indent);

      # get possible virual declaration flag
      if ($line =~ /.*;\s*\/\/!.*\\purevirtual/) {
	&printOut("virtual ");
	$isPureVirtual = 1;
      }
      if ($line =~ /.*;\s*\/\/!.*\\virtual/) {
	&printOut("virtual ");
      }

      &printOut("$methodName(");
      &printArgs(\@argTypes, $args);

      if ($isPureVirtual) {
	&printOut(") = 0;\n");
      } else {
	&printOut(");\n");
      }
      next;
    }
    
    # method implementation
    if ($line =~ /^\s*body\s*${className}::([a-zA-Z0-9_]*)\s+{(.*)}\s*/) {
      my $methodName = $1;
      my $args = $2;

      &printOut("${className}::$1(");

      if (defined $memberTypes{$methodName}) {
	&printArgs($memberTypes{$methodName}, $args);
      } else {
	&printArgs(\@argTypes, $args);
      }
      &printOut(") {\n");
      next;
    }
    
    # proc declaration
    if ($line =~ /^(\s*)$scopeDecl\s+proc\s+([a-zA-Z0-9_]*) +{(.*)}/) {
      my $indent = $1;
      my $procName = $2;
      my $args = $3;

      &printOut("${indent}static $procName(");
      &printArgs(\@argTypes, $args);
      &printOut(");\n");
      $memberTypes{$procName} = [ @argTypes ];
      next;
    }

    # class proc implementation
    if ($line =~ /^\s*proc\s*${className}::([a-zA-Z0-9_]*)\s+{(.*)}\s*/) {
      my $procName = $1;
      my $args = $2;

      &printOut("static ${className}::$procName(");
      if (defined $memberTypes{$procName}) {
	&printArgs($memberTypes{$procName}, $args);
      } else {
	&printArgs(\@argTypes, $2);
      }
      &printOut(") {\n");
      next;
    }

    # normal proc implementation
    if ($line =~ /^\s*proc\s*([a-zA-Z0-9_]*)\s+{(.*)}\s*/) {
      &printOut("$1(");
      &printArgs(\@argTypes, $2);
      &printOut(") {\n");
      next;
    }

    # class variable declaration
    my $arg0 = $argTypes[0];

    if (!$arg0) {
      if($line =~ s/^(\s*)common\s+([a-zA-Z0-9_]+)\s+({.*})\s*$/$1static TclList $2 = $3;/) {
	$memberTypes{$1} = ["TclList"];
      }
      if ($line =~ s/^(\s*)common\s+([a-zA-Z0-9_]+)\s+(".*")\s*$/$1static TclString $2 = $3;/) {
	$memberTypes{$2} = ["TclString"];
      }
      if ($line =~ s/^(\s*)common\s+([a-zA-Z0-9_]+)\s+(\d+|\d.\d+|.\d+)\s*$/$1static TclNumber $2 = $3;/) {
	$memberTypes{$2} = ["TclNumber"];
      }
      if ($line =~ s/^(\s*)$scopeDecl\s+variable\s+([a-zA-Z0-9_]+)\s+({.*})\s*$/$1TclList $2 = $3;/) {
	$memberTypes{$2} = ["TclList"];
      }
      if ($line =~ s/^(\s*)$scopeDecl\s+variable\s+([a-zA-Z0-9_]+)\s+(".*")\s*$/$1TclString $2 = $3;/) {
	$memberTypes{$2} = ["TclString"];
      }
      if ($line =~ s/^(\s*)$scopeDecl\s+variable\s+([a-zA-Z0-9_]+)\s+(\d+|\d.\d+|.\d+)\s*$/$1TclNumber $2 = $3;/) {
	$memberTypes{$2} = ["TclNumber"];
      }
      $arg0 = "TclRef"
    }
    $line =~ s/^(\s*)common\s+([a-zA-Z0-9_]+).*$/$1static $arg0 $2;/;
    $line =~ s/^(\s*)$scopeDecl\s+variable\s+([a-zA-Z0-9_]+)\s+(.*)$/$1$arg0 $2 = $3;/;
    $line =~ s/^(\s*)$scopeDecl\s+variable\s+([a-zA-Z0-9_]+)\s*$/$1$arg0 $2;/;

    # tcl variable declaration
    if ($line =~ s/^(\s*)set\s+([a-zA-Z0-9_]+)\s+({.*})\s*$/$1TclList $2 = $3;/) {
      $memberTypes{$2} = ["TclList"];
    }
    if ($line =~ s/^(\s*)set\s+([a-zA-Z0-9_]+)\s+(".*")\s*$/$1TclString $2 = $3;/) {
      $memberTypes{$2} = ["TclString"];
    }
    if ($line =~ s/^(\s*)set\s+([a-zA-Z0-9_]+)\s+(\d+|\d.\d+|.\d+)\s*$/$1TclNumber $2 = $3;/) {
      $memberTypes{$2} = ["TclNumber"];
    }
    $line =~ s/^(\s*)set\s+([a-zA-Z0-9_]+)\s+(.*)$/$1$arg0 $2 = $3;/;

    # option declaration
    if ($line =~ /^(\s*)itk_option\s+define -(\S+)\s+(\S+)\s+(\S+)\s+(.*)$/) {
      my $indent = $1;
      my $varName = $2;
      my $initValue = $5;
      if ($scopeDecl ne "public") {
	$scopeDecl = "public";
	&printOut("$scopeDecl:\n");
      }
      if ($initValue =~ /^(\d+|\d.\d+|.\d+)$/) {
	&printOut("$indent$varName(TclNumber value = $initValue);\n");
	$memberTypes{$varName} = [ "TclNumber" ];
      } elsif ($initValue =~ /(".*")/) {
	&printOut("$indent$varName(TclString value = $initValue);\n") ;
	$memberTypes{$varName} = [ "TclString" ];
      } elsif ($initValue =~ /({.*})/) {
	&printOut("$indent$varName(TclList value = $initValue);\n");
	$memberTypes{$varName} = [ "TclList" ];
      } else {
	&printOut("$1$2(TclRef value = $initValue);\n");
	$memberTypes{$varName} = [ "TclRef" ];
      }
      next;
    }
    
    # option implementation
    if ($line =~ /^(\s*)configbody\s+${className}::([a-zA-Z0-9_]*)\s+{/) {
      my $indent = $1;
      my $varName = $2;

      &printOut("$indent${className}::$varName("); 
      if (defined $memberTypes{$varName}) {
	&printArgs($memberTypes{$varName}, "value");
      } elsif (@argTypes) {
	&printArgs(\@argTypes, "value");
      } else {
	&printArgs([$arg0], "value");
      }
      &printOut(") {\n");
      next;
    }


    &printOut("$line\n");
  }
  return $errorNumber;
}

###############################################################################

exit (&main() != 0); 
