/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define DEBUGSEARCH
#define DEBUGCOMBINE


/* -- INCLUDES ------------------------------------------------------------- */
#include <stdio.h>
#include <string.h>
#include "cdg.h"
#include "chart.h"
#include "hook.h"
#include "input.h"
#include "tagger.h"
#include "transform.h"

/*

  This module provides a solution method for CDG constraint nets that is
  modeled on the widely-used chart mechanism. Basically, it constructs
  subtrees (consistent sets of LVs whose dependents are adjacent) and tries
  to construct a complete solution by recursively recombining existing
  subtrees.

  Although this method is a variant of search and therefore just as
  infeasible as all the complete solution methods, it is hoped that it will
  nevertheless work better than naive search. For one, it is capable of
  working with incomplete constraint nets -- that is, nets that do not
  contain all the LVs theoretically possible. The idea is that if we
  succeed in finding a good solution that consists mainly of local
  attachments, most of the other LVs need never be built at all, which
  speeds up the creation and possibly also the search of a net.

  Also, subtrees can be viewed as an approximation of constituents as used
  in normal charts, e.g. a subtree headed by an NN corresponds to an NP,
  etc. It is hoped that this kind of classification allows statistical
  heuristics to operate better than the unstructured `set of LVs' approach
  used so far. For instance, it is easy to see from even a very small tree
  bank that the `Fragment' penalty caused by the head word of such an `NP'
  is spurious (i.e. it will probably disappear after recombination) and
  should be disregarded, whereas an agreement mismatch will always remain a
  mismatch.   

*/

/* -- MACROS --------------------------------------------------------------- */

/* -- TYPE DEFINITIONS ----------------------------------------------------- */


/* A Tree is essentially an Analysis limited to a particular part of the
   sentence, augmented with information about which word is the head.

   HEAD is the value of ->arc->from of the word that is dominates all other
   words.

   The span of the tree is held in ->a->subStart and ->a->subStop. subStart
   is the value of ->arc->from of the first word, and subStop is the value
   of ->arc->to of the last word.

*/

   

typedef struct {
  Analysis a;
  Badness b;
  LexemNode head;
  String type;
  Number score;
  int no;
} TreeStruct;
typedef TreeStruct *Tree;

/* -- VARIABLES ------------------------------------------------------------ */


/* List of subtrees ordered by score */
static List agenda = NULL;

/* subtrees ordered by time points */
static Array chart = NULL;

/* the level declared as the main level */
static Level Main =  NULL;

/* score we should reach before terminating */
static Number target = 0.5;

/* counter of all trees ever created */
static int tree_counter = 0;

/* ratio of length vs. Badness for computing tree scores */
static float bias = 0.1;

/* table of constraints that often fail on fragments */
static Hashtable fragments = NULL;

/* -- FUNCTIONS ------------------------------------------------------------ */

/* -- IMPLEMENTATION ------------------------------------------------------- */
/* ----------------------------------------------------------------------
   Read table of category/constraint pairs that often fail in fragments.
*/
void read_fragment_table() {
  FILE *file;
  char *filename = "/home/foth/dawai/txt/fragment-constraints";
  char buffer[1024];
  int nrConstraints = hashSize(inputCurrentGrammar->constraints);

  fragments =
    hashNew(10, 0.7, hashStringHashFunction, hashStringEqualFunction);

  if (NULL == (file = fopen(filename, "r"))) return;

  while(!feof(file)) {
    char *s;
    char *cat;
    char *id;
    int i;
    Constraint c;
    Vector v;

    /* read `ADJA:Fragment:3156' */
    fgets(buffer, 1024, file);
    if(NULL == (s = strtok(buffer, ":"))) continue;
    cat = strRegister(s);
    if(NULL == (v = hashGet(fragments, cat))) {
      v = vectorNew(nrConstraints);
      vectorSetAllElements(v, (Pointer)0);
      hashSet(fragments, cat, v);
    }
    if(NULL == (s = strtok(NULL, ":"))) continue;
    id = strRegister(s);
    if(NULL == (c = hashGet(inputCurrentGrammar->constraints, id))) {
      cdgPrintf(CDG_INFO, "WARNING: constraint `%s' not defined!\n", id);
      continue;
    }
    if(NULL == (s = strtok(NULL, ":"))) continue;
    if(1 != sscanf(s, "%d", &i)) continue;
    vectorSetElement(v, (Pointer)i, c->no);
#ifdef DEBUGFRAGMENTS
    cdgPrintf(CDG_INFO, "%s fragments violate `%s' %d times.\n",
              cat, c->id, i);
#endif
  }
  fclose(file);
}

/* ----------------------------------------------------------------------
   Create empty Tree structure.
*/
Tree newTree() {
  Tree t = memMalloc(sizeof(TreeStruct));
  t->no = tree_counter++;
  t->b = bestBadness();
  return t;
}

/* ----------------------------------------------------------------------
   Compute score of a tree.
*/
void tScore(Tree t) {

  /* If the head of a subtree is forbidden by POS information, this may be
     overlooked by the constraints since there is no actual LV binding this
     word. We still want to toss it out, though, so we add an extra penalty
     for combinatorially improbable heads. */
  if(lgTagscore(t->head) < 1.0) {
    bAdd(t->a->b, lgTagscore(t->head));
  }
  
  bCopy(t->b, t->a->b);

  /* subtract the values spurious constraint failures */
  if(fragments && t->type) {
    Vector v = hashGet(fragments, t->type);
    List l;
    for(l = t->a->conflicts; l != NULL; l = l->next) {
      ConstraintViolation cv = l->item;
      if(((int)vectorElement(v, cv->constraint->no)) > 0) {
        bSubtract(t->b, cv->penalty);
      }
    }
  }

  t->score =
    t->b->soft - t->b->hard +
    bias * ((float)(t->a->subStop - t->a->subStart));
}

/* ----------------------------------------------------------------------
   Try to optimize the auxiliary levels of T.
*/
void tComputeAux(Tree t) {
  int i;
  for(i = t->a->subStart; i < t->a->subStop; i++) {
    aComputeAux(t->a, i, TRUE, TRUE);
  }
}

/* ----------------------------------------------------------------------
   Read the value of the taggerCategoryPath in LN.
*/
String find_cat(LexemNode ln) {
  if(-1 != taggerCategoryIndex) {
    Value v = ln->lexem->values[taggerCategoryIndex];
    if(v->type == VTString) {
      return v->data.string;
    }
    return NULL;
  }
  return NULL;
}

/* ----------------------------------------------------------------------
   Create Tree from LVs.
*/
Tree treeFromLVs(Problem P, Vector LVs, LexemNode head) {
  Tree result = newTree();
  Analysis a = aNew(P);
  int i;

  aBuildPath(a);
 
  a->subStart = head->arc->from;
  a->subStop = head->arc->to;

  if(LVs) {
    a->subStart = P->lg->max;
    a->subStop = P->lg->min;
    for(i = 0; i < vectorSize(LVs); i++) {
      LevelValue lv = vectorElement(LVs, i);
      if(!lv) continue;
      if(((LexemNode)(lv->dependents->item))->arc->from < a->subStart) {
        a->subStart = ((LexemNode)(lv->dependents->item))->arc->from;
      }
      if(((LexemNode)(lv->dependents->item))->arc->to > a->subStop) {
        a->subStop = ((LexemNode)(lv->dependents->item))->arc->to;
      }
      aInsert(a, lv);
    }

    for(i = a->subStart; i < a->subStop; i++) {
      aComputeAux(a, i, TRUE, TRUE);
    }
  }

  result->a = a;
  result->head = head;
  result->type = find_cat(head);
  tScore(result);
  return result;
}

/* ----------------------------------------------------------------------
   Deallocate a Tree.
*/
void deleteTree(Tree t) {
  aDelete(t->a);
  bDelete(t->b);
  memFree(t);
}


/* ----------------------------------------------------------------------
   Is this a subtree that can be part of the annotation?
*/
Boolean isViable(Tree t)
{

  int i;
  LevelValue lva, lvb;
  Problem P = t->a->P;

  if (!P->annotation) {
    return FALSE;
  }

  /* If T has a head whose correct regent is within the tree's span, and its
   * LV is not present, reject the tree. */
  i = t->head->arc->from * P->nrLevels + Main->no;
  lva = vectorElement(P->annotation->slots, i);
  if (lva->regents &&
      ((LexemNode)(lva->regents->item))->arc->from >= t->a->subStart &&
      ((LexemNode)(lva->regents->item))->arc->to < t->a->subStop &&
      !vectorElement(t->a->slots, i)) {
    return FALSE;
  }

  for (i = 0; i < P->nrSlots; i++) {
    lva = vectorElement(t->a->slots, i);
    lvb = vectorElement(P->annotation->slots, i);
    if (!lvb)
      continue;
    if (!lva)
      continue;
    if (!isValid(P->annotation, lva, FALSE))
      return FALSE;
    if (((LexemNode)(lva->dependents->item))->arc->from !=
	((LexemNode)(lvb->dependents->item))->arc->from) {
      return FALSE;
    }

    /* allow modifiee/label mismatch for the NIL subordination of a Tree's
     * head if that head is actually subordinated outside this tree */
    if (lva->label != lvb->label ||
	((LexemNode)(lva->regents->item))->arc->from !=
	((LexemNode)(lvb->regents->item))->arc->from) {
      if (!lva->regents &&
	  ((LexemNode)(lva->dependents->item))->arc->from ==
	    t->head->arc->from && lvb->regents &&
	  (((LexemNode)(lvb->regents->item))->arc->from < t->a->subStart || 
	  ((LexemNode)(lvb->regents->item))->arc->from > t->a->subStop)) {
	continue;
      }
      return FALSE;
    }
  }

  return TRUE;

}

/* ----------------------------------------------------------------------
   Print representation of a Tree.
*/
void printTree(unsigned long int mode, Tree t, int verbosity) {
  LexemGraph lg = t->a->P->lg;
  int i, j;

  /* print correctness marker */
  if(isViable(t)) {
    cdgPrintf(mode, "*");
  }
  
  /* print index */
  cdgPrintf(mode, "[%d", t->no);

  /* print corresponding words */
  if(verbosity > 1) {
    for(i = t->a->subStart; i < t->a->subStop; i++) {
      for(j = 0; j < vectorSize(lg->graphemnodes); j++) {
        GraphemNode gn = vectorElement(lg->graphemnodes, j);
        if(i != gn->arc->from) {
          continue;
        } else {
          cdgPrintf(mode, " %s%s",
                    t->head->arc->from == i ? "*" : "",
                    gn->arc->word);
          break;
        }
      }
    }
  }
  cdgPrintf(mode, "]");
  if(verbosity > 1) {
    cdgPrintf(mode, " %4.3e", t->score);
    cdgPrintf(mode, " (%d--%d)", t->a->subStart, t->a->subStop);
  }

  if(verbosity > 2) {
    cdgPrintf(mode, "\n");
    for(i = 0; i < vectorSize(t->a->slots); i++) {
      LevelValue lv = vectorElement(t->a->slots, i);
      if(!slotActive(t->a, i)) continue;
      if(!lv) continue;
      cdgPrintf(mode, "  #%3d: ", lv->indexWRTNet);
      lvPrint(mode, lv, NULL, 1);
      cdgPrintf(mode, "\n");
    }
  }
  if(verbosity > 3) {
    printConflicts(mode, t->a);
  }
}


/* ----------------------------------------------------------------------
   Compare two trees.
*/
Boolean compareTree(Tree a, Tree b) {
  return a->score > b->score;
}

/* ----------------------------------------------------------------------
   Try to complete R so that it can be exported as a Parse.

   This simply adds the most suitable binding for the head. It only
   fails if there is no possible subordination. It is not guaranteed that
   this is a NIL binding, but of course it will almost always be.
*/
Boolean complete(Tree t) {
  int index = inputCurrentGrammar->noOfLevels * t->head->arc->from + Main->no;
  Patch p = optimizeDomains(t->a, (Pointer)index, OD_INT | OD_ALLOWALL);
  if(!p) return FALSE;
  apply(t->a, p);

  aCheck(t->a, 3);
  
  return TRUE;
}

/* ----------------------------------------------------------------------
   Add a new tree to the search space.
*/
void add(Tree t) {
  List l;
  agenda = listInsertSorted(agenda, t, compareTree);
  l = arrayElement(chart, t->a->subStart, t->a->subStop);
  l = listPrependElement(l, t);
  arraySetElement(chart, l, t->a->subStart, t->a->subStop);
#ifdef DEBUGADD
  cdgPrintf(CDG_INFO, "add ");
  printTree(CDG_INFO, t, 4);
  cdgPrintf(CDG_INFO, "\n");
#else
  cdgPrintf(CDG_INFO, "add ");
  printTree(CDG_INFO, t, 2);
  cdgPrintf(CDG_INFO, "\n");
#endif
}


/* ----------------------------------------------------------------------
   Decide what to do with a freshly constructed Tree, and do it.

   If the Tree violates hard constraints, it is discarded.
   
   If the Tree is complete and better than the previous record, it is used
   as P->best.

   If the Tree is complete but worse than P->best, it is discarded.

   If the Tree is incomplete, it is added to the chart.
*/
void maybe_add(Tree t) {
  Problem P = t->a->P;

  if(t->b->hard) {
    deleteTree(t);
    return;
  }

  if(t->a->subStart == t->a->P->lg->min &&
     t->a->subStop == t->a->P->lg->max &&
     complete(t)) {
    if(!P->best || aCompare(t->a, P->best)) {
#ifdef DEBUGSEARCH
      cdgPrintf(CDG_INFO, "new best: [%d] ", t->no);
      bPrint(CDG_INFO, t->b);
      cdgPrintf(CDG_INFO, "\n");
#endif
      newMaximum(t->a);
    }
    deleteTree(t);
  } else {            
    add(t);
  }
}



/* ----------------------------------------------------------------------
   Combine two Trees into new Trees.

   A must always be the Tree on the left. The following recombinations are
   possible:

   (1) attach A: the head of A is subordinated under the head of B.
   (2) root A: the head of A is subordinated under NIL
   (3) attach B: the head of B is subordinated under the head of A.
   (4) root B: the head of B is subordinated under NIL
   (5) join: if both Trees have the same head, they can be merged without
       addition ov LVs.
   
   If more than one Tree is possible, all of them are tried out.

   All Trees produced are disposed of via maybe_add().

*/
void combine(Tree a, Tree b) {
  Problem P = a->a->P;
  List l;
  int i, j, index = 0;
  Analysis new;
  Boolean can_join = FALSE;
  List domain;
  
  /* reject misordered trees */
  if(a->a->subStart >= b->a->subStart) {
    return;
  }

  /* reject incompatible trees */
  for(i = a->a->subStart; i < a->a->subStop; i++) {
    for(j = 0; j < P->nrLevels; j++) {
      int index = P->nrLevels * i + j;
      LevelValue lv = vectorElement(a->a->slots, index);
      if(lv && !isValid(b->a, lv, FALSE)) {
        return;
      }
    }
  }

  /* Detect overlapping trees.

     There are slight complications about whether overlapping trees should
     be allowed to combine or not. Basically, we should only combine Trees
     that cover adjacent spans, e.g. one that covers the interval [2-5)
     with one that covers [5-9). But we should allow an overlap of one if
     both trees have the same head, and it is on the verge of both trees.

     Example: Suppose we have already built the trees

       John--Subject-->loves   and   loves<--Object--Mary

     and want to combine them. Since both of them have `loves' as the
     head, we should allow them to combine to form the obvious tree. This
     is what possibility (5) is for. But to allow it, we demand that at the
     spans of of both trees must be different, and that both heads are the
     same. 
  */
  if(a->a->subStop != b->a->subStart) {
    if(a->head->arc->from == b->head->arc->from &&
       a->head->arc->to == a->a->subStop &&
       b->head->arc->from == b->a->subStart) {
      can_join = TRUE;
    } else {
      return;
    }
  }
  
  /* create initial version of the new tree  */
  new = aClone(b->a);
  new->subStart = min(a->a->subStart, b->a->subStart);
  new->subStop = max(a->a->subStop, b->a->subStop);
#ifdef DEBUGCOMBINE
  cdgPrintf(CDG_INFO, "combine ");
  printTree(CDG_INFO, a, 0);
  cdgPrintf(CDG_INFO, " and ");
  printTree(CDG_INFO, b, 0);
  cdgPrintf(CDG_INFO, "\n");
#endif
  
  for(i = a->a->subStart; i < a->a->subStop; i++) {
    for(j = 0; j < P->nrLevels; j++) {
      LevelValue lv = vectorElement(a->a->slots, P->nrLevels * i + j);
      if(lv) {
        aInsert(new, lv);
      }
    }
  }
  aJudge(new);

  if(can_join) {
  }

  else {
    /* maybe build suitable connecting LVs */
    i = P->net->totalNumberOfValues;
    if(cnBuildTriple(P->net,
                     a->head->arc->from,
                     b->head->arc->from,
                     Main->no)) {
      importLVs(P, i);
    }
    i = P->net->totalNumberOfValues;
    if(cnBuildTriple(P->net,
                     b->head->arc->from,
                     a->head->arc->from,
                     Main->no)) {
      importLVs(P, i);
    }
  
    /* (1), (2) */
    index = P->nrLevels * a->head->arc->from + Main->no;

    /* need to clone P->domains[index] because newMaximum() can change
       that List during the loop */
    domain = listClone (P->domains[index]);
    for (l = domain; l; l = l->next) {
      LevelValue lv = l->item;
      Tree t;
      if(lv->isDeleted ||
         (lv->regents && ((LexemNode)(lv->regents->item))->arc->from != b->head->arc->from))
        continue;
      t = newTree();
      t->a = aClone(new);
      t->head = b->head;
      t->type = b->type;
      aInsert(t->a, lv);
      tComputeAux(t);
      tScore(t);
#ifdef DEBUGCOMBINE
      cdgPrintf(CDG_INFO, "(%d)\n", lv->regents ? 1 : 2);
#endif
      maybe_add(t);
    }
    listDelete(domain);

    /* (3), (4) */
    index = P->nrLevels * b->head->arc->from + Main->no;
    domain = listClone (P->domains[index]);
    for (l = domain; l; l = l->next) {
      LevelValue lv = l->item;
      Tree t;
      if(lv->isDeleted ||
         (lv->regents && ((LexemNode)(lv->regents->item))->arc->from != a->head->arc->from))
        continue;
      t = newTree();
      t->a = aClone(new);
      t->head = a->head;
      t->type = a->type;
      aInsert(t->a, lv);
      tComputeAux(t);
      tScore(t);
#ifdef DEBUGCOMBINE
      cdgPrintf(CDG_INFO, "(%d)\n", lv->regents ? 3 : 4);
#endif
      maybe_add(t);
    }
    listDelete(domain);
  }  
  aDelete(new);
}


/* ----------------------------------------------------------------------
   Perform chart search.
*/
void search(Problem P) {
  int i;
  Analysis a = aNew(P);
  
  /* Try to combine T with already known subtrees
     that cover the span X--Y. */
  void search_combinations(Tree t, int x, int y) {
    List l = arrayElement(chart, x, y);
    List n;
#ifdef DEBUGSEARCH
    cdgPrintf(CDG_INFO, "search_combinations([%d], (%d--%d))\n", t->no, x, y);
#endif
    for(n = l; n != NULL; n = n->next) {
      Tree right = n->item;
      combine(t, right);
    }
  }
  
  agenda = NULL;
  chart = arrayNew(P->lg->max+1, P->lg->max+1, 0);
  
  /* create trivial trees for LVs that connect adjacent nodes */
  a = aNew(P);
  aBuildPath(a);
  for(i = 0; i < P->nrSlots; i++) {
    List l;
    if(i % P->nrLevels != Main->no) continue;
    for(l = P->domains[i]; l != NULL; l = l->next) {
      LevelValue lv = l->item;
      Tree t;

      if(lv->regents &&
         ((LexemNode)(lv->regents->item))->arc->from != 
	 ((LexemNode)(lv->dependents->item))->arc->to &&
         ((LexemNode)(lv->regents->item))->arc->to != 
	 ((LexemNode)(lv->dependents->item))->arc->from) {
        continue;
      }
      t = newTree();
      t->a = aClone(a);
      aInsert(t->a, lv);
      t->a->subStart = ((LexemNode)(lv->dependents->item))->arc->from;
      t->a->subStop = ((LexemNode)(lv->dependents->item))->arc->to;
      if(lv->regents && ((LexemNode)(lv->regents->item))->arc->to > t->a->subStop) {
        t->a->subStop = ((LexemNode)(lv->regents->item))->arc->to;
      }
      if(lv->regents && ((LexemNode)(lv->regents->item))->arc->from < t->a->subStart) {
        t->a->subStart = ((LexemNode)(lv->regents->item))->arc->from;
      }
      t->head = lv->regents ?
        lv->regents->item :
        lv->dependents->item;
      t->type = find_cat(t->head);
      tScore(t);
      add(t);
    }
  }
    
  /* recombine existing trees into new ones */
  while (agenda) {
    Tree t = agenda->item;
    agenda = agenda->next;
#ifdef DEBUGSEARCH
    cdgPrintf(CDG_INFO, "read ");
    printTree(CDG_INFO, t, 0);
    cdgPrintf(CDG_INFO, "\n");
#endif

    /* Combine this tree with other subtrees that are already known.
       Basically we need to search all trees with spans that start where t's
       span ends. */ 
    for(i = P->lg->min; i <= t->a->subStart; i++) {
      search_combinations(t, i, t->a->subStart);
    }
    for(i = t->a->subStop+1; i <= P->lg->max; i++) {
      search_combinations(t, t->a->subStop, i);
    }
    
    /* But if t ends with its head, we also need to search spans
       that start one word earlier to allow combinations with trees that
       start with that word. combine() will check that this is the case
       before attempting a join. */
/*     if(t->head->arc->to == t->a->subStop) { */
/*       for(i = t->a->subStop; i <= P->lg->max; i++) { */
/*         search_combinations(t, t->head->arc->from, i); */
/*       } */
/*     } */
    
    if(P->best && P->best->b->hard == 0 &&
       P->best->b->soft > target) {
      break;
    }
    
    if(cdgCtrlCTrapped || timerExpired) break;
  }

  if(P->best) {
    addParse(P->best);
    aJudge(P->best);
    aPrint(CDG_INFO, P->best);
    printConflicts(CDG_INFO, P->best);
  }
  
  listForEachDelete(agenda, deleteTree);
  arrayDelete(chart);
}

/* ----------------------------------------------------------------------
   Gather arguments and call the actual search routine.
*/
Boolean chart_search(int no, char **args) {
  Lattice lat;
  ConstraintNet net;
  int index, i, j;
  List l;
  Problem P;
  
  /* interpret args */
  if (no < 1) {
    if (inputMostRecentlyCreatedLattice == NULL) {
      cdgPrintf(CDG_ERROR,
		"ERROR: `chart' needs a lattice.\n");
      return FALSE;
    }
    cdgPrintf(CDG_INFO,
	      "INFO: using most recently created word graph `%s'\n",
	      inputMostRecentlyCreatedLattice->id);
    lat = inputMostRecentlyCreatedLattice;
  } else {

    if (!(lat = findLattice(args[0]))) {
      cdgPrintf(CDG_ERROR,
                "ERROR: `chart': wordgraph `%s' unknown\n",
                args[0]);
      return FALSE;
    }
  }

  /* create structures */
  Main = inputGetMainlevel(inputCurrentGrammar);
  net = cnBuild(lat, FALSE, TRUE);
  index = net->totalNumberOfValues;

  read_fragment_table();
  
  /* build NIL and adjacent LVs for the main levels */
  for(i = net->lexemgraph->min; i < net->lexemgraph->max; i++) {
    if(i > net->lexemgraph->min) {
      cnBuildTriple(net, i, i-1, Main->no);
    }
    cnBuildTriple(net, i, -1, Main->no);
    if(i < net->lexemgraph->max-1) {
      cnBuildTriple(net, i, i+1, Main->no);
    }
  }


  /* build all LVs for auxiliary levels */
  for(i = net->lexemgraph->min; i < net->lexemgraph->max; i++) {
    for(j = -1; j < net->lexemgraph->max; j++) {
      for(l = inputCurrentGrammar->levels; l != NULL; l = l->next) {
        Level level = l->item;
        if(level == Main) continue;
        cnBuildTriple(net, i, j, level->no);
      }
    }
  }
  cnOptimizeNet(net);
  
  if(!(P = net2problem(net, 0.0, NULL))) {
    return FALSE;
  }

  cdgCtrlCTrapped = FALSE;
  cdgCtrlCAllowed = TRUE;
  search(P);
  cdgCtrlCTrapped = FALSE;
  cdgCtrlCAllowed = FALSE;

  return TRUE;
}


/* ------------------------------------------------------------------------- */
/* -- ENDOFFILE ------------------------------------------------------------ */
/** @} */


