/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/*
 * File: incrementalcompletion.h
 * Author: Ingo Schroeder
 * Purpose: header file for incrementalcompletion
 * Birth: Aug/99
 * 
 */

#ifndef INCREMENTALCOMPLETION_H
#define INCREMENTALCOMPLETION_H

/* ---------------------------------------------------------------------- */

#include "input.h"
#include "constraintnet.h"

/* ---------------------------------------------------------------------- */
#ifndef SWIG

typedef struct {
  String agendasize;   /* size of agenda used in search */
  String graph_file;   /* root names of the files used for graphical display     */
  String info_file;    /* and for table display; they will have added '-i-j.fig' */
  /* and '-i-j.tex', where  'i' is the current incremental iteration and */
  /* 'j' is the current solution number at this iteration */
  int heurclass1;     /* heuristics; there are five classes, */
  int heurclass1dist; /* each with an optional distance argument */
  int heurclass2;    
  int heurclass2dist;
  int heurclass3;    
  int heurclass3dist;
  int heurclass4;    
  int heurclass4dist;
  int heurclass5;    
  int heurclass5dist;
  float frob_time;     /* time to run frobbing, as multiple of time since last run */
  float frob_pressure; /* pressure to apply to frobbing */
  int frob_agendasize; /* agenda for frobbing */
  int frob_beamwidth;  /* beamwith for frobbing */
  int frob_iter;       /* iteration frequency frobbing is allowed to run */
  float hotspot_val;   /* consider all nodes with score in [hot, hot + val] */
  Boolean hotspot;     /* work with hotspot(s) */
  Boolean hotpath;     /* work with hotpath */
  Boolean method_ns;   /* allow netsearch */
  Boolean method_fr;   /* allow frobbing */
} IcParamsStruct, *IcParams;
#endif

/* ---------------------------------------------------------------------- */
/* -- VARIABLES --------------------------------------------------------- */

/* Default IC parameters string */
extern String icParamsStr; 

#ifndef SWIG
/* Default IC parameters structure */
extern IcParams icParams;    

/* -- FUNCTIONS --------------------------------------------------------- */

/* Module initialization */
extern void icInitialize(void);
extern Boolean icValidateIcParams(String name, String value, String *var);

/* Module termination */
extern void icFinalize(void);

/* Incremental Completion main entry point */
extern int icIncrementalCompletion(Lattice latice, IcParams params, int no, char **args);

/* Functions that run a specific IC method */
extern void icInc(Lattice lattice, IcParams icParams, int no, char **args);
extern void icIncInteractive(IcParams icParams, int no, char **args);

/* Functions that provide elementary functionality to different IC method functions */
extern ConstraintNet icInitializeNet(Lattice lattice, IcParams icParams);
extern Boolean icAddArcsToNet(ConstraintNet net, List listArcs, IcParams icParams);
extern Boolean icAddArcsToNetSmaller(ConstraintNet net, List listArcs, IcParams icParams);
extern Boolean icSearchNet(ConstraintNet net, IcParams icParams, Boolean lastIteration, int no, char **args);
extern void icPrintSolutions(ConstraintNet net, IcParams icParams);

/* These should be used together: icDeleteWord downgrades the
 * constraint net with the ONE last word, and icAddWord adds ONE word
 * to the constraint net and does the search on it */
extern Boolean icDeleteWord(ConstraintNet net);
extern Boolean icAddWord(ConstraintNet net, String word, IcParams icParams, int no, char **args);

/* set ic parameters */
extern IcParams icSetParams(String string);

#endif

/* ----------------------------------------------------------------------
Description of IC heuristics

The five heuristics are:

(1) Allow links FROM new node.
(2) Allow links TO new node from nodes in the previous solution.
(3) Allow changing the lexical reading of nodes in the previous solution.
(4) Allow changing the label of node values in the previous solution.
(5) Allow changing the governor of nodes in the previous solution.

The value <m> of the heur3class<i> field define the area of action of the 
heuristic (i): 

(1) Operation takes place on last node of previous solution. 
(2) Operation takes place on last node of all trees of previous solution.
(3) Operation takes place on nodes on the last path of previous solution.
(4) Operation takes place on nodes on the last path of all trees of previous 
    solution.
(5) Operation takes place on all nodes.

The value <n> of the heur3class<i>dist field, if non-zero, define an
additional constraint: the operation of heuristic (i) is allowed only on 
nodes within a distance of <n> of the new node.

The values of <m> and <n> for heuristic (i) are entered as command line
arguments, like  

     -<i> <m> <n>

or, if no <n> is needed,

     -<i> <m>
------------------------------------------------------------------------- */

/* ---------------------------------------------------------------------- */
#endif /* don't insert anything after this #endif */
