/*

    The BLAH library, a container library
    Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
    Contact: blah@nats.informatik.uni-hamburg.de

    
*/

/* ------------------------------------------------------------------- */
/** \mainpage The BLAH Reference Manual.

\author Ingo Schrder
\author Kilian A. Foth
\author Michael Daum

\section Introduction Introduction

This is the reference manual for the BLAH library for the C programming
language. BLAH stands for bitstrings, lists, arrays and hashes --- admittedly a poorly
chosen giving the set of container data types most usually used. Actually
there are more data types than these but who cares if the name is funky enuf.

The BLAH library is copyright by The CDG Team; it is distributed under
the GNU General Publice License Version 2. You should have received a
copy of the GPL with the software.

The maintainers of this software can be contacted at the following
email address:
\verbatim blah@nats.informatik.uni-hamburg.de \endverbatim

\section Usage Usage

The BLAH library is installed as both shared and static versions of the
library by default. In order to use it your C code must include the
header file \code blah.h \endcode

For instance, if you installed the header file in a system directory,
a minimal C program might look as follows:

\code
#include <stdio.h>
#include <blah.h>

int main(int argc, char **argv)
{
  List squares=listNew();
  int i;

  for (i=0; i<100; i++)
    {
      squares=listPrependElement(squares, (Pointer)(i*i));
    }
  printf("42th square is %d\n", (int)listNthElement(squares, 42));
  listDelete(squares);
}
\endcode

You have to link against the BLAH library as well as the math library
to create a binary:

\verbatim
$ gcc -o blah-example -lblah -lm blah-example.c
$ ./blah-example
42th square is 3364
$
\endverbatim

\section Overview Overview 

The BLAH library defines the following container data types:

  - \ref Array:
    An array stores information that is accessed using a number of indices.
    Arrays usually have two or more dimensions. The number of dimensions as
    well as the size of each dimension must be provided at creation time of
    the array and stay fixed. The required time to access a specific
    information given the corresponding indices is constant; the memory is
    proportional to the product of all dimension sizes.
  - \ref BitString:
    A bitstring stores a sequence of Booleans. The design emphasis is on
    memory efficiency. 
  - \ref ByteVector:
    A bitvector also stores a sequence of Booleans. However, one byte is used
    to store a single Boolean. This data type is more time efficient than a
    bitstring.
  - \ref Hashtable:
    A hashtable stores key-value pairs. The key is used to access the
    actual information in the value. Access time is (almost) constant for
    arbitrary sizes of the hashtable and arbitrary keys.
  - \ref List:
    A list stores a sequence of objects. Access to the head of the list
    is efficiently possible while in general access to an arbitrary
    object requires linear time.
  - \ref ListAgenda:
    An agenda is created by sorting the items according to priority using
    a simple linked list as its storage medium.
  - \ref Memory:
    Some basic Operations with the memory are done in this module.
  - \ref Prime:
    Prime is a module that makes use of Rabin's Probablistic Primetest-Algorithm
    for generating the prime numbers equal to the Hash table entries.
  - \ref Ringbuffer:
     It is similar to the vector that stores a series of objects, though
     the head and the tail are attached inorder to enhance cyclic operations.
  - \ref String:
    Strings in C are represented by arrays of characters. The end of the string
    is marked with a special character, the null character, which is simply the
    character with the value 0. Whenever we write a string, enclosed in double
    quotes, C automatically creates an array of characters for us, containing that
    string, terminated by the \a NULL character.
  - \ref TreeAgenda:
    An agenda is created by sorting the items according to the order of the
    priority using an unbalanced binary tree.
  - \ref Vector:
    A vector stores objects at a given index. The size of the vector is
    linear in the largest used index. Time to access an object given the
    index is constant.
*/

#ifndef BLAH_H
#define BLAH_H

#include <stdarg.h>
#include <stdlib.h>
#include "config.h"

/* CONSTANTS ---------------------------------------------------------- */
#ifndef NULL
#define NULL 0
#endif

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef max
#define max(a, b) (a) > (b) ? (a) : (b)
#endif

#ifndef min
#define min(a, b) (a) < (b) ? (a) : (b)
#endif

typedef unsigned char Boolean;
typedef void *Pointer;

#ifndef SWIG
typedef void VoidFunction ();
typedef Boolean BooleanFunction ();
typedef int IntFunction ();
typedef Pointer PointerFunction ();
#endif


/* FORWARD DECLARATIONS ---------------------------------------------- */
struct ArrayStruct;
typedef struct ArrayStruct ArrayStruct, *Array;
struct BitStringStruct;
typedef struct BitStringStruct BitStringStruct, *BitString;
struct ByteVectorStruct;
typedef struct ByteVectorStruct ByteVectorStruct, *ByteVector;
struct HashtableEntryStruct;
typedef struct HashtableEntryStruct HashtableEntryStruct, *HashtableEntry;
struct HashtableStruct;
typedef struct HashtableStruct HashtableStruct, *Hashtable;
struct HashIteratorStruct;
typedef struct HashIteratorStruct HashIteratorStruct, *HashIterator;
struct ListStruct {
  Pointer item;	            /**< void element link to member of list */
  struct ListStruct *next;  /**< link to the next element in the list */
};
typedef struct ListStruct ListStruct, *List;
struct ListAgendaEntryStruct;
typedef struct ListAgendaEntryStruct ListAgendaEntryStruct, *ListAgendaEntry;
struct ListAgendaEntryStruct;
typedef struct ListAgendaStruct ListAgendaStruct, *ListAgenda;
typedef List ListAgendaIterator;
struct RingBufferStruct;
typedef struct RingBufferStruct RingBufferStruct, *RingBuffer;
typedef char *String;
struct TANodeStruct;
typedef struct TANodeStruct TANodeStruct, *TANode;
struct TreeAgendaStruct;
typedef struct TreeAgendaStruct TreeAgendaStruct, *TreeAgenda;
struct TreeAgendaIteratorStruct;
typedef struct TreeAgendaIteratorStruct TreeAgendaIteratorStruct, *TreeAgendaIterator;
struct VectorStruct;
typedef struct VectorStruct VectorStruct, *Vector;

struct StringTreeStruct {
  String value;
  struct StringTreeStruct *next;
  struct StringTreeStruct *firstchild;
};
typedef struct StringTreeStruct StringTreeStruct, *StringTree;

struct StringTreePosStruct {
  struct StringTreeStruct *st;
  String pos;
};
typedef struct StringTreePosStruct StringTreePosStruct, *StringTreePos;

/* MODULES ----------------------------------------------------------- */

/* MAIN -------------------------------------------------------------- */
void blahInitialize(void);
void blahFinalize(void);

/* AGENDA ------------------------------------------------------------ */
#ifndef SWIG
#ifdef TREEAGENDA
#define AgendaStruct TreeAgendaStruct
#define Agenda TreeAgenda

#define agNew taNew
#define agSetVerbosity taSetVerbosity
#define agVerbosity taVerbosity
#define agSize taSize
#define agMaxSize taMaxSize
#define agMaxSizeSoFar taMaxSizeSoFar
#define agIsEmpty taIsEmpty
#define agIsTruncated taIsTruncated
#define agResetTruncated taResetTruncated
#define agInsert taInsert
#define agBest taBest
#define agRemoveBest taRemoveBest
#define agDelete taDelete

#define AgendaIterator TreeAgendaIterator

#define agIteratorNew taIteratorNew
#define agIteratorNextElement taIteratorNextElement 
#define agIteratorDelete taIteratorDelete

#else

#define AgendaStruct ListAgendaStruct
#define Agenda ListAgenda

#define agNew laNew
#define agSetVerbosity laSetVerbosity
#define agVerbosity laVerbosity
#define agSize laSize
#define agMaxSize laMaxSize
#define agMaxSizeSoFar laMaxSizeSoFar
#define agIsEmpty laIsEmpty
#define agIsTruncated laIsTruncated
#define agResetTruncated laResetTruncated
#define agInsert laInsert
#define agBest laBest
#define agRemoveBest laRemoveBest
#define agDelete laDelete

#define AgendaIterator ListAgendaIterator

#define agIteratorNew laIteratorNew
#define agIteratorNextElement laIteratorNextElement 
#define agIteratorDelete laIteratorDelete

#endif
#endif


/* ARRAY ------------------------------------------------------------- */
#ifndef SWIG
extern Array arrayNew(int i, ...);
extern Pointer arraySetElement(Array a, Pointer element, ...);
extern Pointer arrayElement(Array a, ...);
#endif

extern Array arrayClone(Array a);
extern void arrayDelete(Array a);
extern void arraySetAllElements(Array a, Pointer element);
extern int arrayDimension(Array a, int dim);


/* BITSTRING --------------------------------------------------------- */
extern BitString bitNew(int size);
extern BitString bitClone(BitString bs);
extern BitString bitCopy(BitString dst, BitString src);
extern void bitDelete(BitString bs);
extern int bitSize(BitString bs);
extern void bitPrint(BitString bs);
extern void bitSet(BitString a, int no);
extern void bitClear(BitString a, int no);
extern void bitSetAll(BitString a);
extern void bitClearAll(BitString a);
extern BitString bitAnd(BitString a, BitString b);
extern BitString bitOr(BitString a, BitString b);
extern Boolean bitCheck(BitString a, BitString b);
extern Boolean bitGet(BitString bs, int no);
extern Boolean bitIsAllCleared(BitString bs);
extern Boolean bitIsAllSet(BitString bs);

/* BITVECTOR --------------------------------------------------------- */
extern ByteVector bvNew(int capacity);
extern void bvDelete(ByteVector a);
extern int bvAddElement(ByteVector v, char element);
extern char bvElement(ByteVector v, int index);
extern char bvRemoveElement(ByteVector v, int index);
extern char bvInsertElement(ByteVector v, char element, int index);
extern char bvSetElement(ByteVector v, char element, int index);
extern void bvSetElements(ByteVector v, char element, int from, int to);
extern void bvSetAllElements(ByteVector v, char element);
extern int bvCapacity(ByteVector v);
extern int bvSize(ByteVector v);
extern Boolean bvIsEmpty(ByteVector v);
extern ByteVector bvCopy(ByteVector dst, ByteVector src);
extern ByteVector bvClone(ByteVector v);
extern void bvExtend(ByteVector bv, int n, char value);
extern char bvAndElement(ByteVector v, char element, int index);
extern char bvOrElement(ByteVector v, char element, int index);
extern char bvNotElement(ByteVector v, int index);

/* HASHTABLE --------------------------------------------------------- */
extern Hashtable hashNew(int capacity,
		     double loadFactor,
		     IntFunction *hashFunction,
		     IntFunction *keyEqualFunction);
extern void hashDelete(Hashtable ht);
extern void hashForEach(Hashtable ht, VoidFunction *f);
extern void hashForEachWithData(Hashtable ht, VoidFunction *f, Pointer data);
extern void hashForEachFree(Hashtable ht, VoidFunction *f);
extern void hashForEachFreeValue(Hashtable ht, VoidFunction *f);
extern Pointer hashSet(Hashtable ht, Pointer key, Pointer value);
extern Pointer hashGet(Hashtable ht, Pointer key);
extern Pointer hashRemove(Hashtable ht, Pointer key);
extern int hashSize(Hashtable ht);
extern Boolean hashIsEmpty(Hashtable ht);
extern Boolean hashContainsKey(Hashtable ht, Pointer key);
extern Boolean hashContainsValue(Hashtable ht, Pointer value);
extern List hashListOfKeys(Hashtable ht);
extern HashIterator hashIteratorNew(Hashtable ht);
extern Pointer hashIteratorNextKey(HashIterator hi);
extern Pointer hashIteratorNextValue(HashIterator hi);
extern void hashIteratorDelete(HashIterator hi);
extern int hashStringHashFunction(char *s);
extern int hashStringEqualFunction(char *s, char *t);

/* LIST -------------------------------------------------------------- */
extern List listNew();
extern List listClone(List l);
extern List listDeepClone(List l, PointerFunction *p);
extern List listCopy(List dst, List src);
extern List listAppendList(List front, List rear);
extern List listAddUniqueElement(List l, Pointer item);
extern List listAppendElement(List l, Pointer item);
extern List listPrependElement(List l, Pointer item);
extern List listInsertSorted(List l, Pointer item, BooleanFunction *f);
extern List listInsertSortedWithData(List l, Pointer item, BooleanFunction *f, Pointer data);
extern List listSort(List l, BooleanFunction *f);
extern List listSortWithData(List l, BooleanFunction *f, void *data);
extern int listSize(List l);
extern Pointer listNthElement(List l, int n);
extern Pointer listLastElement(List l);
extern Boolean listContains(List l, Pointer item);
extern void listForEach(List l, VoidFunction *f);
extern void listForEachDelete(List l, VoidFunction *f);
extern List listFilter(List l, BooleanFunction *f);
extern void listDelete(List l);
extern List listDeleteElement(List l, Pointer item);
extern List listDeleteLastElement(List l);
extern Vector listToVector(List l);
extern int listIndex(List l, Pointer item);
extern List listReverse(List l);
extern Boolean listIsEqual(List l1, List l2);

/* STRINGTREE -------------------------------------------------------------- */
extern StringTree stNew();
extern void stFree(StringTree this);
extern Boolean stEquals(StringTree a, StringTree b);
extern StringTree stClone(StringTree st);
extern void stInsert(StringTree this, String item);
extern void stRemove(StringTree this, String item);
extern StringTreePos stFind(StringTree st, String item);
extern void stFreeStringTreePos(StringTreePos stp);
extern Boolean stIsIn(StringTree this, String item);

/* LISTAGENDA -------------------------------------------------------- */
extern ListAgenda laNew(int, VoidFunction *f);
extern Boolean laSetVerbosity(ListAgenda, Boolean);
extern Boolean laVerbosity(ListAgenda);
extern int laSize(ListAgenda);
extern int laMaxSize(ListAgenda);
extern int laMaxSizeSoFar(ListAgenda);
extern Boolean laIsEmpty(ListAgenda);
extern Boolean laIsTruncated(ListAgenda);
extern Boolean laResetTruncated(ListAgenda);
extern Boolean laInsert(ListAgenda, double, Pointer);
extern Pointer laBest(ListAgenda);
extern Pointer laRemoveBest(ListAgenda);
extern void laDelete(ListAgenda);
extern ListAgendaIterator laIteratorNew(ListAgenda);
extern Pointer laIteratorNextElement(ListAgendaIterator);
extern void laIteratorDelete(ListAgendaIterator);



/* MEMORY ------------------------------------------------------------ */
#ifndef SWIG


#if GC_DEBUG
   #define memMalloc(size) memMallocCheck(GC_debug_malloc(size, __FILE__, __LINE__), __FILE__, __LINE__)
   #define memFree(ptr) (ptr) = NULL
#define memRealloc(ptr, size) memMallocCheck(GC_debug_realloc(ptr, size, __FILE__, __LINE__), __FILE__, __LINE__)
#else
   #define memMalloc(size) memMallocCheck(GC_malloc(size), __FILE__, __LINE__)
   #define memFree(ptr) (ptr) = NULL
   #define memRealloc(ptr, size) memMallocCheck(GC_realloc(ptr, size), __FILE__, __LINE__)
#endif
#include <gc/gc.h>

#endif

extern void *memMallocCheck(void *, char*, int);
extern void memFreeFunction(void *);

/* PRIMES ------------------------------------------------------------ */
int primeRabin(unsigned long number, unsigned long times);
unsigned long primeNext(unsigned long number, unsigned long times);

/* RINGBUFFER -------------------------------------------------------- */
extern RingBuffer rbNew(int capacity);
extern void rbDelete(RingBuffer rb);
extern int rbAddTopElement(RingBuffer rb, Pointer element);
extern int rbAddBottomElement(RingBuffer rb, Pointer element);
extern Boolean rbIsEmpty(RingBuffer rb);
extern Boolean rbIsFull(RingBuffer rb);
extern Pointer rbTopPeek(RingBuffer rb);
extern int rbSize(RingBuffer rb);
extern int rbCapacity(RingBuffer rb);
extern Pointer rbBottomPeek(RingBuffer rb);
extern Pointer rbRemoveTopElement(RingBuffer rb);
extern Pointer rbRemoveBottomElement(RingBuffer rb);
extern Boolean rbContains(RingBuffer rb, Pointer element);
extern RingBuffer rbCopy(RingBuffer dst, RingBuffer src);
extern RingBuffer rbClone(RingBuffer src);
extern void rbForEachWithData(RingBuffer rb,
					  VoidFunction *f,
					  Pointer clientData);
extern void rbClear(RingBuffer rb);

/* STRING ------------------------------------------------------------ */
extern void strInitialize(void);
extern void strFinalize(void);
#ifndef SWIG
extern void strDelete(String);
extern String strRegister(const String);
extern String strCopy(String s);
extern String strCat(String a, String b);
extern String strFromList(List list);
extern String strAppend(String head, ...);
extern String strPrintf(String fmt, ...);
extern String strVPrintf(String fmt, va_list);
extern Boolean strCompare(String a, String b);
extern Boolean strVersionCompare(String a, String b);
extern int strStoreSize(void);
extern String strDecode(String word);
extern String strSubStr(String a, int start, int length);
extern int strFindSubStr(String haystack, String needle);
extern Boolean strIsPrefix(String a, String sub);
extern Boolean strIsSuffix(String a, String sub);
#endif

/* TREEAGENDA ---------------------------------------------------------*/
extern TreeAgenda taNew(int, VoidFunction *);
extern Boolean taSetVerbosity(TreeAgenda, Boolean);
extern Boolean taVerbosity(TreeAgenda);
extern int taSize(TreeAgenda);
extern int taMaxSize(TreeAgenda);
extern int taMaxSizeSoFar(TreeAgenda);
extern Boolean taIsEmpty(TreeAgenda);
extern Boolean taIsTruncated(TreeAgenda);
extern Boolean taResetTruncated(TreeAgenda);
extern Boolean taInsert(TreeAgenda, double, Pointer);
extern Pointer taBest(TreeAgenda);
extern Pointer taRemoveBest(TreeAgenda);
extern void taDelete(TreeAgenda);
extern TreeAgendaIterator taIteratorNew(TreeAgenda);
extern Pointer taIteratorNextElement(TreeAgendaIterator);
extern void taIteratorDelete(TreeAgendaIterator);

/* VECTOR ------------------------------------------------------------ */
extern Vector vectorNew(int capacity);
extern void vectorDelete(Vector a);
extern int vectorAddElement(Vector v, void *element);
extern void *vectorElement(Vector v, int index);
extern void *vectorRemoveElementAt(Vector v, int index);
extern int vectorRemoveElement(Vector v, void *element);
extern void *vectorInsertElement (Vector v, void *element, int index);
extern void *vectorSetElement(Vector v, void *element, int index);
extern void vectorSetElements(Vector v, void *element, int from, int to);
extern void vectorSetAllElements(Vector v, void *element);
extern int vectorIndexOf(Vector v, void *element, int index);
extern int vectorCapacity (Vector v);
extern int vectorSize(Vector v);
extern Boolean vectorIsEmpty(Vector v);
extern Boolean vectorContains(Vector v, void *element);
extern Vector vectorCopy(Vector dst, Vector src);
extern Vector vectorClone(Vector v);
extern Vector vectorSort(Vector v, BooleanFunction *f);
extern Vector vectorSortWithData(Vector v, BooleanFunction *f, void *data);
extern List vectorToList(Vector v);

/* ------------------------------------------------------------------- */
#endif /* #ifndef BLAH_H */

