/*

    The BLAH library, a container library
    Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
    Contact: blah@nats.informatik.uni-hamburg.de

    
*/

/* ----------------------------------------------------------------------
   @defgroup ByteVector ByteVectors

  A bytevector is a variant of vector with char instead of Pointer as
  its element type. There are no other differences.

  @{
 */

/* ---INCLUDES-------------------------------------------------------- */
#include <stdio.h>
#include <string.h>
#include "blah.h"

/* ------------------------------------------------------------------- */
#define RESIZEFACTOR 2

/* -- TYPE DEFINITIONS ----------------------------------------------- */
/* ----------------------------------------------------------------------
 * internal representation of a bit vector
 */
struct ByteVectorStruct
{
#ifdef BITVECTOR_DEBUG
  int magic;		     /**< identification string */
#endif
  int count;		     /**< number of entries */
  int capacity;		     /**< capacity of table */
  char *entries;	     /**< table of entries */
};

/* -- VARIABLES ------------------------------------------------------ */
/* ------------------------------------------------------------------- */
#ifdef BITVECTOR_DEBUG
static int magic_constant=8755156;
#endif

/* -- FUNCTIONS ------------------------------------------------------ */
#ifdef BITVECTOR_DEBUG
static void check_magic_function(ByteVector v, char *file, int no)
{
  if (!v)
    {
      fprintf(stderr, "%s:%d: ByteVector must not be NULL\n", file, no);
      abort();
    }
  if (v->magic!=magic_constant)
    {
      fprintf(stderr, "%s:%d: not a ByteVector\n", file, no);
      abort();
    }
}
#define check_magic(v) check_magic_function(v, __FILE__, __LINE__)
#else /* #ifdef BITVECTOR_DEBUG */
#define check_magic(v)
#endif /* #ifdef BITVECTOR_DEBUG */

/* ----------------------------------------------------------------------
 * doubles capacity of ByteVector.
 *
 * @param v the bitvector that has to be resized.
 * @returns the bitvector after increasing its capacity.
 */
static void resize(ByteVector v)
{
  check_magic(v);

  v->capacity=(int)(RESIZEFACTOR*v->capacity);
  v->entries=(char *)memRealloc(v->entries, sizeof(char)*v->capacity);
  memset(&v->entries[v->count], 0, (v->capacity-v->count)*sizeof(char));
}

/* ----------------------------------------------------------------------
 * creates a new empty bitvector with an initial capacity.
 *
 * Specifying a correct or nearly correct capacity slightly improves the
 * efficiency. The bitvector roughly needs capacity bytes.
 *
 * @param capacity the capacity of the new vector to be created.
 * @returns a new and empty bit vector of the specified capacity.
 */
ByteVector bvNew(int capacity)
{
  ByteVector v;

  if (capacity<8) { capacity=8; }
  v=(ByteVector)memMalloc(sizeof(ByteVectorStruct));
#ifdef BITVECTOR_DEBUG
  v->magic=magic_constant;
#endif
  v->count=0;
  v->capacity=capacity;
  v->entries=(char *)memMalloc(sizeof(char)*capacity);

  memset(v->entries, 0, capacity*sizeof(char));

  return(v);
}

/* ----------------------------------------------------------------------
 * deletes a bitvector and frees the associated memory.
 *
 * note that nothing user-defined is stored in a bitvector.
 * therefore, the user does not need to free anything himself.
 *
 * @param v the bitvector that has to be deleted.
 */
void bvDelete(ByteVector v)
{
  check_magic(s);

  memFree(v->entries);
  memFree(v);
}

/* ----------------------------------------------------------------------
 * adds a new element to the end of the ByteVector.
 *
 * automatically increases the capacity of the ByteVector if necessary
 *
 * @param v the vector to which the new element has to be added.
 * @param element the element that has to be added to the end of the bitvector v
 * @returns index of the vector
 */
int bvAddElement(ByteVector v, char element)
{
  check_magic(v);

  /* The number of entries exceeds the capacity. Resize the table. */
  while (v->count>=v->capacity) { resize(v); }

  v->entries[v->count]=element;
  return v->count++;
}

/* ----------------------------------------------------------------------
 * sets the element at the specific index to a new element.
 *
 * if necessary the bitVector automatically increases its capacity.
 *
 * @param v the bitvector in which the specified element has to be set to a new element.
 * @param index the index of the element that has to be set.
 * @returns the old element or \a False at the specified index.
 */
char bvElement(ByteVector v, int index)
{
  check_magic(v);

  if (index>=0 && index<v->count) { return v->entries[index]; }

  fprintf(stderr, "ERROR: bvElement: index %d out of range\n", index);
  abort();
}

/* ----------------------------------------------------------------------
 * removes element at the specified index.
 *
 * size decreases and all later elements move one position to the front.
 * inefficient method, not recommended.
 *
 * @param v the bitvector from which the element has to be removed.
 * @param index the index at which the element to be removed is present.
 * @returns old element at the specified index.
 */
char bvRemoveElement(ByteVector v, int index)
{
   char p;

   check_magic(v);

   if (index<0 || index>=v->count)
     {
       fprintf(stderr,
	       "ERROR: bvRemoveElement: index %d out of range\n",
	       index);
       abort();
     }

   p=v->entries[index];
   /* TODO: check boundaries */
   if (v->count-index>1)
     {
       memmove(&v->entries[index],
	       &v->entries[index+1],
	       (v->count-index)*sizeof(char));
     }
   v->count--;
   return p;
}

/* ----------------------------------------------------------------------
 * inserts a new element at the given index.
 *
 * if necessary the bitvector automatically increases its capacity.
 * inefficient method,not recommended.
 *
 * @param v the bitvector into which the new element has to be inserted.
 * @param element the new element that has to be inserted into the bitvector v.
 * @param index the index at which the new element has to be inserted in the bitvector v.
 * @returns the old element at the index.
 */
char bvInsertElement(ByteVector v, char element, int index)
{
  char p;

  check_magic(v);

  if (index<0)
    {
      fprintf(stderr,
	      "ERROR: bvInsertElement: index %d out of range\n",
	      index);
      abort();
    }

  /*
   * v->capacity==index is too small because index is counted
   * from zero.
   */
  while (v->capacity<=index) { resize(v); }

  /*
   * if v->count==v->capacity then the vector is
   * completely full; v->count>v->capacity shouldn't be possible.
   */
  if (v->count>=v->capacity) { resize(v); }

  if (v->count-index>0)
    {
      memmove(&v->entries[index+1],
	      &v->entries[index],
	      (v->count-index)*sizeof(char));
    }

  p=v->entries[index];
  v->entries[index]=element;
  v->count=max(v->count+1, index+1);

  return p;
}

/* ----------------------------------------------------------------------
 * sets element at the given index.
 *
 * if necessary the vector automatically increases its capacity.
 *
 * @param v the vector in which the element has to be set at the specified index.
 * @param element the element that has to be set in the bitvector.
 * @param index the index at which the element to be set is present.
 * @returns The old element at the index.
 */
char bvSetElement(ByteVector v, char element, int index)
{
  char p;

  check_magic(v);

  if (index<0)
    {
      fprintf(stderr,
	      "ERROR: bvSetElement: index %d out of range\n",
	      index);
      abort();
    }

  /* TODO: check boundaries */
  while (v->capacity<=index) { resize(v); }

  p=v->entries[index];
  v->entries[index]=element;
  v->count=max(v->count, index+1);
  return p;
}

/* ----------------------------------------------------------------------
 * sets all the elements between index \a from and \a to (excluding to) to a new value.
 *
 * if necessary the bit vector automatically increases in capacity.
 *
 * @param v the bitvector in which the elements have to be set.
 * @param element the value of the bits at the corresponding indices.
 * @param from the source index from which all the elements in the bitvector have to be set.
 * @param to the destination index until which all the elements in the bitvector have to be set.
 * @returns the bitvector v after setting all the elements at the specified indices.
 *
 */
void bvSetElements(ByteVector v, char element, int from, int to)
{
   check_magic(v);

   if (from<0 || to<0)
     {
       fprintf(stderr,
	       "ERROR: bvSetElement: index %d out of range\n",
	       from<0 ? from : to);
       abort();
     }

   if (from>to) { int tmp=from; from=to; to=tmp; }

   /* TODO: check boundaries */
   /* to is excluded, so v->capacity==to is enough */
   while (v->capacity<to) { resize(v); }

   memset(&v->entries[from], element, (to-from)*sizeof(char));
   v->count=max(v->count, to);
}

/* ----------------------------------------------------------------------
 * sets all the elements to a new value.
 *
 * the size increases to the current capacity of the bit vector.
 *
 * @param v the bitvector in which all the elements have to be set.
 * @param element the elements at the specified indices.
 *
 * @returns the bitvector after setting all the elements.
 */
void bvSetAllElements(ByteVector v, char element)
{
   memset(v->entries, element, v->capacity*sizeof(char));
   v->count=v->capacity;
}

/* ----------------------------------------------------------------------
 * finds the current capacity of bitvector.
 *
 * @param v the bitvector whose capacity has to be retrieved.
 * @returns the capacity of the bitvector v.
 */
int bvCapacity(ByteVector v)
{
  check_magic(v);

  return v->capacity;
}

/* ----------------------------------------------------------------------
 * finds the number of entries(size) in the bitvector.
 *
 * @param v the bitvector whose size has to be retrieved.
 * @returns the size of the bitvector v.
 */
int bvSize(ByteVector v)
{
  check_magic(v);

  return v->count;
}

/* ----------------------------------------------------------------------
 * finds if the bit vector is empty or not.
 *
 * @param v the bitvector for which emptiness is checked.
 * @returns \a TRUE if bitvector is empty and  \a FALSE otherwise.
 */
Boolean bvIsEmpty(ByteVector v)
{
  check_magic(v);

  return v->count==0 ? TRUE : FALSE;
}

/* ----------------------------------------------------------------------
 * creates an exact copy of the specified bitvector.
 *
 * @param v the bitvector for which cloning has to be performed.
 * @returns a complete independent \a CLONE of the bit vector v.
 */
ByteVector bvClone(ByteVector v)
{
  ByteVector new;

  check_magic(v);

  new=bvNew(bvCapacity(v));
  bvCopy(new, v);
  return new;
}

/* ----------------------------------------------------------------------
 *
 * copies entries of one bitvector to another.
 *
 * copying is done from bitVector \a src to the bitvector \a dst.
 * \a dst vector automatically increases its capacity.
 * @param src source index or the index from which copying has to be initiated.
 * @param dst destination index or the index until which the copying has to be done.
 * @returns the new \a dst vector with the elements copied into it.
 */
ByteVector bvCopy(ByteVector dst, ByteVector src)
{
   if (dst->capacity<src->capacity)
     {
       dst->entries=(char *)memRealloc(dst->entries, sizeof(char)*src->capacity);
     }

   memcpy(dst->entries, src->entries, sizeof(char)*src->capacity);

   dst->count=src->count;
   dst->capacity=src->capacity;

   return dst;
}


/**
 * extend a byte vector by a certain size
 * @param value is inserted
 * @param n times at the end of
 * @param bv
 */
void bvExtend(ByteVector bv, int n, char value) {
	int i;
	int size=bvSize(bv);
	for(i=0; i<n; i++) {
		bvInsertElement(bv,value,size+i);
	}
}

/* ----------------------------------------------------------------------
 * computes AND to the element at the given index.
 *
 * if necessary the vector automatically increases its capacity.
 *
 * @param v the vector in which the element has to be set at the specified index.
 * @param element the element that has to be set in the bitvector.
 * @param index the index at which the element to be set is present.
 * @returns The new value at the index.
 */
char bvAndElement(ByteVector v, char element, int index)
{
  check_magic(v);

  if (index < 0) {
    fprintf(stderr, "ERROR: bvSetElement: index %d out of range\n", index);
    abort();
  }

  /* TODO: check boundaries */
  while (v->capacity <= index) {
    resize(v);
  }

  return (v->entries[index] &= element);
}

/* ----------------------------------------------------------------------
 * computes OR to the element at the given index.
 *
 * if necessary the vector automatically increases its capacity.
 *
 * @param v the vector in which the element has to be set at the specified index.
 * @param element the element that has to be set in the bitvector.
 * @param index the index at which the element to be set is present.
 * @returns The new value at the index.
 */
char bvOrElement(ByteVector v, char element, int index)
{
  check_magic(v);

  if (index < 0) {
    fprintf(stderr, "ERROR: bvSetElement: index %d out of range\n", index);
    abort();
  }

  /* TODO: check boundaries */
  while (v->capacity <= index) {
    resize(v);
  }

  return (v->entries[index] |= element);
}

/* ----------------------------------------------------------------------
 * computes NOT to the element at the given index.
 *
 * if necessary the vector automatically increases its capacity.
 *
 * @param v the vector in which the element has to be set at the specified index.
 * @param index the index at which the element to be set is present.
 * @returns The new value at the index.
 */
char bvNotElement(ByteVector v, int index)
{
  check_magic(v);

  if (index < 0) {
    fprintf(stderr, "ERROR: bvSetElement: index %d out of range\n", index);
    abort();
  }

  /* TODO: check boundaries */
  while (v->capacity <= index) {
    resize(v);
  }

  return (v->entries[index] = ~v->entries[index]);
}


/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/** @} */
