/*

    The BLAH library, a container library
    Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
    Contact: blah@nats.informatik.uni-hamburg.de

    
*/

/* ----------------------------------------------------------------------
 * @defgroup List Lists
 * A list container. A list is a sequence of data objects where the data items are usually
 * access from the beginning of list. The head of a list is the first
 * data item and the tail is a list containing the remaining objects.
 * @{
 */

/* ----INCLUDES------------------------------------------------------- */
#include <stdio.h>
#include <stdarg.h>
#include "blah.h"

/* ---------------------------------------------------------------------- 
 * this is used to debug the cell allocation for Lists.
 *
 * It expands to _newCell when BLAH is compiled with -DLIST_DEBUG or to the normal
 * memMalloc function.
 *
 * @see _newCell, freeListCell 
 */
#ifdef LIST_DEBUG
#define newListCell _newCell()

/** @see newListCell */
List _newCell(void) {
  List result = memMalloc(sizeof(ListStruct));
  printf("+ cell %p\n", result);
  return result;
}

#else
#define newListCell (List)memMalloc(sizeof(ListStruct))
#endif

/* ---------------------------------------------------------------------- 
 * this is used to debug the cell deallocation for Lists.
 * It expands to _freeCell when BLAH is compiled with -DLIST_DEBUG or to the normal
 * memFree function.
 *
 * @see newCell
 */
#ifdef LIST_DEBUG
#define freeListCell(x) _freeCell(x)

/** @see freeListCell */
void _freeCell(List cell) {
  printf("- cell %p\n",cell);
  memFree(cell);
}

#else
#define freeListCell memFree
#endif

/* ---------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * get a new list.
 *
 * A new List is represented by NULL.
 *
 * @returns NULL
 */
List listNew()
{
  return NULL;
}

/* ----------------------------------------------------------------------
 * clones a list.
 *
 * This doesn't clone the items.
 * 
 * @param l the list that has to be cloned.
 * @returns a newly allocated list after cloning.
 */
List listClone(List l)
{
  List m = NULL;

  if ( l == NULL )
    return( NULL );

  m = newListCell;

  m->item = l->item;
  m->next = listClone( l->next );

  return( m );
}

/* ----------------------------------------------------------------------
   Clones a list, performing a deep copy of all items via P.
*/
List listDeepClone(List l, PointerFunction *p)
{
  List m;

  if (l == NULL) {
    return NULL;
  }

  m = newListCell;
  m->item = (*p)(l->item);
  m->next = listDeepClone(l->next, p);

  return m;
}

/* ----------------------------------------------------------------------
 * set list-items of dst to those of src by reusing old buckets.
 *
 * It works as follows:
 * - if dst has more buckets than src then these are freed
 * - if dst has less buckets than src then new once are allocated
 * - if dst is NULL then listCopy is equivalent to listClone
 * - if src is NULL then listCopy is equivalent to listDelete
 *
 * @param dst the resulting list
 * @param src the originating list
 * @returns the dst List
 */
List listCopy(List dst, List src)
{
  List ldst, lsrc, prev;

  /* empty dst = listClone */
  if (!dst) {
    return listClone(src);
  }

  /* empty src = listDelete */
  if (!src) {
    listDelete(dst);
    return NULL;
  }
    
  /* real work */
  for (prev = NULL, ldst = dst, lsrc = src; 
      ldst && lsrc; 
      prev = ldst, ldst = ldst->next, lsrc = lsrc->next) 
  {
    ldst->item = lsrc->item;
  }

  /* more buckets in dst than in src */
  if (ldst) {
    listDelete(ldst);

  /* prev should never bee NULL here because dst != NULL */
    prev->next = NULL;
  } else {
    prev->next = listClone(lsrc);
  }

  return dst;
}

/* ----------------------------------------------------------------------
 * appends a list to another list.
 *
 * @param front a list to whose end another list rear is appended.
 * @param rear a list which is appended to the tail of the front list
 * @returns the new list after appending.
 */
List listAppendList(List front, List rear)

 {
  List l;

  if ( front == NULL ) 
    return( rear );

  l = front;
  while ( l->next != NULL )
    l = l->next;
  l->next = rear;
  return( front );
}

/* ----------------------------------------------------------------------
 * appends an item to the end of a list.
 *
 * @param l the list to which an item has to be appended.
 * @param item the item that has to be appended to the end of the list l.
 * @returns the new list after appending.
 */
List listAppendElement(List l, Pointer item)
{
  List m, n;

  m = newListCell;

  m->next = NULL;
  m->item = item;

  if (l == NULL) {
    return (m);
  } else {
    n = l;
    while (n->next != NULL)
      n = n->next;
    n->next = m;
    return (l);
  }
}

/* ----------------------------------------------------------------------
 * prepends an item to a list.
 *
 * @param l the list to which an item has to be prepended.
 * @param item the item that has to be prepended to the list l.
 * @returns the new list after prepending.
 */
List listPrependElement(List l, Pointer item)
{
  List m;

  m = newListCell;

  m->next = l;
  m->item = item;

  return( m );
}

/* ----------------------------------------------------------------------
 * appends a couple of items to the end of a list.
 *
 * @param oldList the list to which a couple of items are to be appended.
 * @param ... the list of items are to be appended to the end of the list.
 */
List listAppendElements(List oldList, ...)
{
  va_list ap;
  Pointer item;
  List newList = oldList;
  
  va_start(ap, oldList);
  while ((item = va_arg(ap, Pointer)) != NULL) {
    newList = listAppendElement(newList, item);
  }
  va_end(ap);

  return newList;
}

/* ----------------------------------------------------------------------
 * prepends a couple of items to the end of a list.
 *
 * @param oldList the list to which the items are prepended.
 * @param ... the list of items that are to be prepended to the list l.
 * @returns the new list after prepending.
 */
List listPrependElements(List oldList, ...)
{
  va_list ap;
  Pointer item;
  List newList = oldList;
  
  va_start(ap, oldList);
  while ((item = va_arg(ap, Pointer)) != NULL) {
    newList = listPrependElement(newList, item);
  }
  va_end(ap);

  return newList;
}

/* ----------------------------------------------------------------------
 * inserts an item keeping an order defined by f.
 * 
 * @param list the list into which a specified item has to be inserted.
 * @param item the item that has to be inserted into the list l.
 * @param f the function that defines the ordering of the during insertion.
 * @returns the new list after insertion operation.
 */
List listInsertSorted(List list, Pointer item, BooleanFunction *f)
{
   List l, prev, newList;

   newList = newListCell;

   newList->item = item;
   newList->next = NULL;

   for (prev = NULL, l = list; l; prev = l, l = l->next) {
      if ((*f) (item, l->item))
	 break;
   }
   newList->next = l;
   if (prev) {
      prev->next = newList;
      return (list);
   } else {
      return (newList);
   }
}

/* ----------------------------------------------------------------------
 * inserts an item keeping an order defined by f and some extra data.
 *
 * @param list the list into which a specified item has to be inserted.
 * @param item the item that has to be inserted into the list l.
 * @param 'f' the function that defines the ordering of the during insertion.
 * @param clientData the data that determines the ordering along with the function 'f'
 * @returns the new list after insertion operation.
 */
List listInsertSortedWithData(List list, Pointer item, BooleanFunction * f,
			      Pointer clientData)
{
  List l, prev, newList;

  newList = newListCell;

  newList->item = item;
  newList->next = NULL;

  for (prev = NULL, l = list; l; prev = l, l = l->next) {
    if ((*f) (item, l->item, clientData))
      break;
  }
  newList->next = l;
  if (prev) {
    prev->next = newList;
    return (list);
  } else {
    return (newList);
  }
}

/* ----------------------------------------------------------------------
 * adds an item to the list if and only if it is not already present.
 *
 * @param l the list to which the specified item has to be added.
 * @param item the item that has to be added to the list l.
 * @returns the new list after the addition of the item if its unique.
 */
List listAddUniqueElement(List l, Pointer item)
{
  if(listContains(l,item)) {
    return l;
  } else {
    return listAppendElement(l,item);
  }
}

/* ----------------------------------------------------------------------
 * retrieves number of items in the list.
 *
 * @param l the list whose size has to be retrieved.
 * @returns the total number of items in the list l.
 */
int listSize(List l)
{
  int i = 0;

  List ll = l;

  while ( ll != NULL ) {
    i++;
    ll = ll->next;
  }
  return( i );
}

/* ----------------------------------------------------------------------
 * returns nth item in list.
 *
 * @param l the list whose 'n'th item has to be retrieved.
 * @param n the number that idicates the item that has to be retrieved.
 * @returns the item in the 'n'th position of the list.
 */
Pointer listNthElement(List l, int n)
{
  if ( l == NULL || n < 1 ) {
    fprintf( stderr, "WARNING: listNthElement: list is empty or n < 1\n" );
    abort();
  } else if ( n == 1 ) {
    return( l->item );
  } else {
    return( listNthElement( l->next, n - 1 ) );
  }
}

/* ----------------------------------------------------------------------
 * returns last item in list.
 *
 * @param l the list from which the last item has to be retrieved.
 * @returns the last item of the list l.
 */
Pointer listLastElement(List l )
{
  if ( l == NULL ) {
    fprintf( stderr, "WARNING: listLastElement: list is empty\n" );
    abort();
  } else {
    while ( l->next != NULL )
      l = l->next;
    return( l->item );
  }
}

/* ----------------------------------------------------------------------
 * checks if the list contains a particular item.
 *
 * @param l the list in which the existance of the specified item has to be checked.
 * @param p specifies the item that has to be checked for in the list l.
 * @returns \a TRUE if the list contains the item or \a FALSE otherwise.
 */
Boolean listContains(List l, Pointer p)
{
  for ( ; l != NULL; l = l->next )
    if ( l->item == p )
      return( TRUE );
  return( FALSE );
}

/* ----------------------------------------------------------------------
 * calls function `f' for each list element.
 *
 * @param l the list in which the function 'f' is called.
 * @param f the function which is called for each element in the list l.
 */
void listForEach(List l, VoidFunction *f)
{
  while ( l != NULL) {
    (*f)( l->item );
    l = l->next;
  }
}

/* ----------------------------------------------------------------------
 * like listForEach, but frees list, list becomes inaccessible.
 *
 * @param l the list in which the function 'f' has to be called.
 * @param f the function that is called for each element in the list l.
 */
void listForEachDelete(List l, VoidFunction *f)
{
  List m;

  while ( l != NULL ) {
    (*f)( l->item );
    m = l;
    l = l->next;

    freeListCell( m );
  }
}

/* ----------------------------------------------------------------------
 * filters list, returns new (sub-)list.
 *
 * @param l the list on which the function is evaluated.
 * @param f the function that is evaluated in the list l to generate a (sub-) list
 * @returns an independent list consisting of all the elements of the given list l that make
 * function 'f' evaluate to \a TRUE
 */
List listFilter(List l, BooleanFunction *f)
{
  List m = NULL;

  while ( l != NULL) {
    if ( (*f)( l->item ) ) {
      m = listAppendElement( m, l->item );
    }
    l = l->next;
  }
  return( m );
}

/* ----------------------------------------------------------------------
 * frees list, does NOT free elements.
 *
 * @param l the list that needs to be freed.
 * @returns nothing.
 */
void listDelete(List l)
{
  List m;
  while ( l != NULL ) {
    m = l;
    l = l->next;

    freeListCell( m );
  }
}

/* ----------------------------------------------------------------------
 * deletes all occurences of item from list.
 * 
 * @param l the list from which all the occurances of a specified item has to be deleted.
 * @param p the particular item whose occurances in the list l has to be deleted.
 * @returns the new list after deleting the item from the list.
 */
List listDeleteElement(List l, Pointer p)
{
  List m, n;
  
  if ( l == NULL ) 
    return( l );

  if ( l->item == p ) {
    m = l->next;

    freeListCell( l );
    return( listDeleteElement( m, p ) );
  } else {
    m = l;
    while ( m->next != NULL ) {
      while ( m->next != NULL && m->next->item != p )
	m = m->next;
      
      if ( m->next == NULL ) {
	return( l );
      } else {
	n = m->next;
	m->next = m->next->next;

	freeListCell( n );
      }
    }
    return( l );
  }
}
/* ----------------------------------------------------------------------
 * deletes the last item from the list.
 *
 * @param l the list from which the last item has to be deleted.
 * @returns the new list after deleting the last item form the list l.
 */
List listDeleteLastElement(List l)
{
  List m = NULL; /* The initialisation is done due to prohibit a compiler warning. */
  List orig;

  if ( l == NULL )
    return( l );

  if(l->next == NULL){
    freeListCell( l );
    return NULL;
  } else {
    orig=l;
    while ( l->next != NULL ) {
      m = l;
      l = l->next;
    }
    freeListCell( m->next );
    m->next = NULL;

    return( orig );
  }
}

/* ----------------------------------------------------------------------
 * converts a list into a vector.
 * 
 * @param l the list that has to be converted into a vector v.
 * @returns the vector v corresponding to the list l.
 */
Vector listToVector(List l)
{
   Vector v;
   List ll;
   int i;

   if (l == NULL)
      return NULL;

   v = vectorNew(listSize(l));
   for (ll=l, i=0; ll != NULL; ll=ll->next, i++) {
      vectorSetElement(v, ll->item, i);
   }

   return v;
}

/* ----------------------------------------------------------------------
 * find a particular item in a list .
 *
 * @param l the list from which the specified item has to be found.
 * @param item the item that has to be found from the list l.
 * @returns the index of the item if found else zero.
 */
int listIndex(List l, Pointer item)
{

  int result = 1;
  List m = l;

  for (m = l; m != NULL; m = m->next, result++) {
    if (m->item == item) {
      return result;
    }
  }

  return 0;

}

/* ----------------------------------------------------------------------
 * compare two lists item per item.
 * 
 * Two lists are equal if they have the same size and 
 * contain the same elements in the same order.
 * 
 * @param list1 the first list whose items are to be compared.
 * @param list2 the second list whose items are compared with the list 1.
 * @returns \a TRUE if the items are all equal and \a FALSE otherwise.
 */
Boolean listIsEqual(List list1, List list2)
{
  List l, m;

  if (listSize(list1) != listSize(list2)) 
    return FALSE;

  for (l = list1, m = list2; l && m; l = l->next, m = m->next) {
    if (l->item != m->item)
      return FALSE;
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
 * sorts a list, using a user-specified compare function.
 *
 * @param l the list that has to be sorted using the function 'f'
 * @param f the compare function that is used for the sorting of the list l.
 * @returns the new list after sorting using the function 'f'
 */
List listSort(List l, BooleanFunction *f)
{
  if (l) {
    Vector v = listToVector(l);
    vectorSort(v, f);
    l = vectorToList(v);
    vectorDelete(v);
    return l;
  } else {
    return NULL;
  }
}

/* ----------------------------------------------------------------------
 * sorts a list, using a user-specified compare function and some data.
 *
 * @param l the list that has to be sorted using the function 'f'
 * @param f the compare function that is used for the sorting of the list l.
 * @param data the data that defines the sorting along with the function 'f'
 * @returns the new list after sorting using the function 'f' and the data.
 */
List listSortWithData(List l, BooleanFunction *f, void *data) 
{
  if (l) {
    Vector v = listToVector(l);
    vectorSortWithData(v, f, data);
    l = vectorToList(v);
    vectorDelete(v);
    return l;
  } else {
    return NULL;
  }
}

/* ----------------------------------------------------------------------
 * return a new reverse list.
 *
 * @param l the list whose data items are to be reversed.
 * @returns a new list after reversing the list l.
 */
List listReverse(List l)
{
  List reverse = NULL;

  while (l) {
    reverse = listPrependElement(reverse, l->item);
    l = l->next;
  }

  return reverse;
}
/* ---------------------------------------------------------------------- */
/** @} */


