/*

    The BLAH library, a container library
    Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: blah@nats.informatik.uni-hamburg.de


*/

/* ------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * @defgroup Vector Vectors
 * implementation of vectors.
 *
 * A Vector is a one-dimensional array but allows for automatic resizing.i.e,the size
 * need not be known in advance.
 * @{
 */

/* ----INCLUDES------------------------------------------------------ */

#include <stdio.h>
#include <string.h>
#include "blah.h"

/* ------------------------------------------------------------------- */
#define RESIZEFACTOR 2

/* ----------------------------------------------------------------------
 * internal representation of a vector
 */
struct VectorStruct {
#if VECTOR_DEBUG
  char *id;		     /**< identification string */
#endif
  int count;		     /**< number of entries */
  int capacity;		     /**< capacity of table */
  void **entries;	     /**< table of entries */
};

/* ------------------------------------------------------------------- */
#if VECTOR_DEBUG
static char *vectorID = "Ingo";

static void checkVector(Vector v, char *s)
{
  if (!v)
    {
      fprintf( stderr, "ERROR: %s: vector must not be NULL\n", s );
      abort();
    }
  if (v->id != vectorID)
    {
      fprintf( stderr, "ERROR: %s: not a vector\n", s );
      abort();
    }
}
#else
#define checkVector(v, s)
#endif

/* ----------------------------------------------------------------------
 * resize vector, double capacity.
 *
 * @param v the vector that has to be resized.
 * @returns the new vector after resizing.
 */
static void resizeVector(Vector v)
{
  checkVector(v, "resizeVector");

  v->capacity=(int)(RESIZEFACTOR*v->capacity);
  v->entries=(Pointer *)memRealloc(v->entries, sizeof(Pointer)*v->capacity);
  memset(&(v->entries[v->count]),
	 0,
	 (v->capacity - v->count) * sizeof(Pointer));
}

/*-----------------------------------------------------------------------
 * creates a new vector with an initial capacity.
 *
 * specifying a correct or nearly correct capacity may slightly improve the efficiency.
 *
 * @param capacity the capacity with which the new vector has to be created.
 * @returns the new vector created with size capacity.
 */
Vector vectorNew(int capacity)
{
  Vector v;
  int i;

  v = (Vector)memMalloc( sizeof(VectorStruct) );
#if VECTOR_DEBUG
  v->id = vectorID;
#endif
  v->count = 0;
  if(capacity <= 0) {
    capacity = 1;
  }
  v->capacity = capacity;
  v->entries = (Pointer *)memMalloc( sizeof(Pointer) * capacity );
  for ( i = 0; i < capacity; i++ )
    v->entries[i] = NULL;

  return( v );
}

/* ----------------------------------------------------------------------
 * deletes vector, but can't free memory for the content.
 *
 * @param v the vector that has to be deleted.
 */
void vectorDelete(Vector v)
{
  checkVector(v, "vectorDelete" );

  memFree( v->entries );
  memFree( v );
}

/* ----------------------------------------------------------------------
 * adds a new element to the end of the vector.
 *
 * if necessary the vector automatically increases its capacity.
 *
 * @param v the vector to which the new element has to be added.
 * @param element the element that has to be added to the vector v.
 * @returns the index of the vector after addition.
 */
int vectorAddElement(Vector v, void *element)
{
  checkVector(v, "vectorAddElement");

  /* The number of entries exceeds the capacity. Resize the table. */
  if ( v->count >= v->capacity ) { resizeVector( v ); }

  v->entries[v->count] = element;
  return( v->count++ );
}


/* ----------------------------------------------------------------------
 * sets the element at a specific index to a new element.
 *
 * if necessary the vector automatically increases its capacity.
 *
 * @param v the vector in which the element has to be set.
 * @param index the index at which the element that has to be set is located.
 * @returns the old element (or \a NULL) at that index.
 */
Pointer vectorElement(Vector v, int index)
{
  checkVector( v, "vectorElement" );

  if ( index >= 0 && index < v->count )
    return( v->entries[index] );

  fprintf( stderr, "WARNING: vectorElement: index %d out of range\n", index);
  abort();
}

/* ----------------------------------------------------------------------
 * removes element at the index.
 *
 * all later elements move one position in front and size decreases.
 * inefficient, not recommended.
 *
 * @param v the vector from which the element has to be removed.
 * @param index the index at which the element that has to be removed is located.
 * @returns the old element at the specified index.
 */
Pointer vectorRemoveElementAt(Vector v, int index)
{
  checkVector(v, "vectorRemoveElementAt");

  if ( index >= 0 && index < v->count ) {
    Pointer p = v->entries[index];
    int i;

    for ( i = index + 1; i < v->count; i++)
      v->entries[i - 1] = v->entries[i];
    v->entries[v->count - 1] = NULL;
    v->count--;
    return( p );
  }

  fprintf( stderr,
	   "WARNING: vectorRemoveElementAt: index %d out of range\n",
	   index);
  abort();
}

/* ----------------------------------------------------------------------
 * removes the specified element from the vector.
 *
 * If the element occurs multiple times, the first item will be removed.
 * If the element does not belong to the vector the program is aborted.
 * Not efficient,not recommended.
 *
 * @param v the vector from which the element to be removed is present.
 * @param element the element that has to be removed from the vector v.
 * @returns the associated index of the vector after removal.
 */
int vectorRemoveElement(Vector v, void *element)
{
  int i;

  checkVector(v, "vectorRemoveElement");

  i = vectorIndexOf( v, element, 0 );
  if ( i < 0 ) {
    fprintf( stderr, "WARNING: vectorRemoveElement: no such element\n" );
    abort();
  } else {
    vectorRemoveElementAt( v, i );
  }
  return( i );
}


/* ----------------------------------------------------------------------
 * inserts a new element at the specified index.
 *
 * Size increases automatically if necessary.
 * Inefficient method,Not recommended.
 *
 * @param v the vector into which the new element has to be added.
 * @param element the element that has to be added in the vector v.
 * @param index the specified index at which the insertion has to be done.
 * @returns the old element at that index (or \a NULL).
 */
Pointer vectorInsertElement(Vector v,
                              void *element,
			      int index )
{
  int i;
  Pointer p;

  checkVector( v, "vectorInsertElement" );

  if ( index < 0 ) {
    fprintf( stderr,
	     "WARNING: vectorInsertElement: index %d out of range\n",
	     index);
    abort();
  }

  while (v->capacity < v->count+1 || v->capacity <= index ) { resizeVector(v); }

  for (i=v->count-1; i>=index; i--) { v->entries[i+1]=v->entries[i]; }

  p = v->entries[index];
  v->entries[index] = element;
  v->count = max( v->count + 1, index + 1);
  return( p );
}

/* ----------------------------------------------------------------------
 * sets the element at the given index.
 *
 * size may increase automatically if necessary.
 *
 * @param v the vector in which the specified element has to be set.
 * @param element the element that has to be set in the vector v.
 * @param index the index at which the element has to be set.
 * @returns the old element at that index.
  */
Pointer vectorSetElement(Vector v, void *element, int index)
{
  Pointer p;

  checkVector( v, "vectorSetElement" );

  if ( index < 0 ) {
    fprintf( stderr, "WARNING: vectorSetElement: index %d out of range\n",
	     index);
    abort();
  }

  while (v->capacity<=index ) { resizeVector(v); }

  p = v->entries[index];
  v->entries[index] = element;
  v->count = max( v->count, index + 1);
  return( p );
}

/* ----------------------------------------------------------------------
 * sets all elements between \a FROM and \a TO(excluding to) to a new value.
 *
 * size may increase automatically if necessary.
 *
 * @param v the vector in which the elements have to be set.
 * @param element determines the element that is being set.
 * @param from this is the source index from which the elements are set.
 * @param to this is the destination index until which the elements are set.
 * @returns the new vector after setting the elements.
 */
void vectorSetElements(Vector v, void *element, int from, int to)
{
   int i;

   checkVector(v, "vectorSetElements");

   if (from<0 || to<0)
     {
       fprintf(stderr,
	       "ERROR: vectorSetElements: region from %d to %d out of range\n",
	       from, to);
      abort();
     }

   if (from>to)
     {
       int tmp;

       tmp = from;
       from = to;
       to = tmp;
     }

   while (v->capacity<to) { resizeVector(v); }

   if (!element)
     {
       memset(&(v->entries[from]), 0, (to-from)*sizeof(void *));
     }
   else
     {
       for (i=from; i<to; i++) { v->entries[i]=element; }
     }
   
   v->count=max(v->count, to);
}

/* ----------------------------------------------------------------------
 * sets all the elements to a new value.
 *
 * Size increases to the current capacity of the vector.
 *
 * @param v the vector in which all the elements have to be set to the new value.
 * @param element determines the element that is being set in vector v.
 */
void vectorSetAllElements(Vector v, void *element)
{
  checkVector(v, "vectorSetAllElements");

  vectorSetElements(v, element, 0, v->capacity);
}

/* ----------------------------------------------------------------------
 * finds the index of an entry in the vector.
 *
 * @param v the vector in which the index of the element has to be retrieved.
 * @param element the element whose index has to be retrieved in the vector v.
 * @param index determines the index of all the elements in the vector.
 * @returns index of a given element in the vector or '-1' if the element does not belong to
 * that specified vector.
 */
int vectorIndexOf(Vector v, void *element, int index)
{
  int i;

  checkVector( v, "vectorIndexOf" );

  for ( i = index; i < v->count; i++ )
    if ( v->entries[i] == element )
      return( i );

  return( -1 );
}

/* ----------------------------------------------------------------------
 * returns the current capacity of vector.
 *
 * @param v the vector whose capacity has to be retrieved.
 * @returns the capacity of the vector v.
 */
int vectorCapacity(Vector v)
{
  checkVector( v, "vectorCapacity" );

  return( v->capacity );
}

/* ----------------------------------------------------------------------
 * returns the current number of entries in the vector.
 *
 * @param v the vector whose current number of entries has to be determined.
 * @returns the total number of entries in the vector v.
 */
int vectorSize(Vector v)
{
  checkVector( v, "vectorSize" );

  return( v->count );
}

/* ----------------------------------------------------------------------
 * checks if the vector is empty or not.
 *
 * @param v the vector for which the emptiness is checked.
 * @returns \a TRUE if the vector is empty and \a FALSE otherwise.
 */
Boolean vectorIsEmpty(Vector v)
{
  checkVector( v, "vectorIsEmpty" );

  return( v->count == 0 ? TRUE : FALSE );
}

/* ----------------------------------------------------------------------
 * checks if a vector contains a given element.
 *
 * @param v the vector in which a particular number has to be searched.
 * @param element the element which has to be checked in the vector v.
 * @returns \a TRUE if vector contains the element and \a FALSE otherwise.
 */
Boolean vectorContains(Vector v, void *element)
{
  int i;

  checkVector( v, "vectorContains" );

  for ( i = 0; i < v->count; i++ )
    if ( v->entries[i] == element )
      return( TRUE );

  return( FALSE );
}

/* ----------------------------------------------------------------------
 * clones a vector.
 *
 * @param v the vector that has to be cloned.
 * @returns a new vector after cloning is done.
 */
Vector vectorClone(Vector v)
{
  Vector new;

  checkVector( v, "vectorClone" );

  new = vectorNew( vectorCapacity( v ) );
  vectorCopy(new, v);
  return( new );
}

/* ----------------------------------------------------------------------
 * copies all the entries of one vector \a src to another vector \a dst.
 *
 * @param src source vector from which the copying is done.
 * @param dst destination vector to which copying is done.
 * @returns the dst vector with the entries copied.
 */
Vector vectorCopy(Vector dst, Vector src)
{
  checkVector( dst, "vectorCopy dst" );
  checkVector( src, "vectorCopy src" );

  if (dst->capacity < src->capacity)
    {
      dst->entries = (Pointer *)
	memRealloc(dst->entries, sizeof(Pointer) * src->capacity);
    }

  memcpy(dst->entries, src->entries, sizeof(Pointer) * src->capacity);

  dst->count = src->count;
  dst->capacity = src->capacity;

  return dst;
}

/* ----------------------------------------------------------------------
 * internal helper function for sorting.
 *
 * @param v the vector in which the sorting operation has to be done.
 * @param f the function that is used for the sorting of the vector f.
 * @param data some additional data that is handed over to the comparison function.
 * @param use determines the usage in the program
 */
static void doSorting(Vector v, BooleanFunction *f, void *data, char use)
{
  Pointer a, b;
  int i, j;
  int size=vectorSize(v);

  /* straight insertion, quickly coded, slowly running */
  for (i=1; i<size; i++)
    {
      for (j=i-1; j>=0; j--)
	{
	  a=vectorElement(v, j+1);
	  b=vectorElement(v, j);
	  /* compare */
	  if (use) {
	    if (!f(a, b, data))
	      break;
	  } else {
	    if (!f(a, b))
	      break;
	  }
	  /* swap a and b */
	  vectorSetElement(v, a, j);
	  vectorSetElement(v, b, j+1);
	}
    }
}

/* ----------------------------------------------------------------------
 * sorts a vector using a user-specified compare function.
 *
 * The function f is called as f(a,b)
 * @param v the vector that has to be sorted.
 * @param f the function that is used for sorting.
 * @returns should return \a TRUE if the element \a a should be before element \a b.
 */
Vector vectorSort(Vector v, BooleanFunction *f)
{
  checkVector(v, "vectorSort");

  doSorting(v, f, NULL, FALSE);
  return v;
}

/* ----------------------------------------------------------------------
 * sorts a vector, using a user-specified compare function and some extra data.
 *
 * The function f is called as f(a,b,)and some data.
 *
 * @param v the vector that has to be sorted.
 * @param f the function that is used for sorting.
 * @param clientData the additional data that is handed over to the comparison function.
 * @returns should return \a TRUE if the element \a a should be before element \a b
 */
Vector vectorSortWithData(Vector v, BooleanFunction *f, void *clientData)
{
  checkVector(v, "vectorSortWithData");

  doSorting(v, f, clientData, TRUE);
  return v;
}

/* ----------------------------------------------------------------------
 * converts a vector into a list.
 *
 * @param v the vector that has to be converted into a list l.
 * @returns the list l corresponding to the vector v.
 */
List vectorToList(Vector v)
{
   List l;
   int i;

   if (v == NULL)
      return NULL;

   l = NULL;
   for (i = vectorSize(v)-1; i >= 0; i--) {
     l = listPrependElement(l, vectorElement(v, i));
   }

   return l;
}

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/** @} */
