#!/usr/bin/env perl
# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  
# cdg/utils/PP-attach.pl
#
# Make predictions about which prepositions in a sentence are likely
# to attach to which regents. Use this in either of three ways:
#
# 1) as a predictor in CDG. The `predict' command must read something like:
#
#  cdg> predict PP 'PP-attach.pl -s <database name> cat
#
# so that CDG gives the correct input.
#
# 2) to solve the classical quadruple task. Call it like this:
#
#  PP-attach.pl -q  <test file name>
#
# where the test file contains specifications such as
#
#  warnt Regierung vor V
#  gibt Ratschlger fr N
#
# 3) To test multiway-ambiguity resolution on complete syntax trees.
# Call it like this:
#
#  PP-attach.pl -e negra:1-1000
use warnings;
use strict;

BEGIN {
  use File::Basename;
  unshift @INC, dirname(readlink($0) || $0);
  require 'cdgsetup.pl';
}

use CDG;
use CDG::German;
use CDG::Utils qw(cda2ae);
use DB_File;

use Getopt::Std;
our $opt_3;
our $opt_b;
our $opt_d;
our $opt_e;
our $opt_f;
our $opt_h;
our $opt_m;
our $opt_q;
our $opt_s;
our $opt_t;
getopts('3b:de:f:hm:qs:t:');

die "

Usage: $0
   -3        use triple statistics in addition to pairs
   -b <num>  boost noun attachment counts by NUM
   -d        print Debug messages
   -e <spec> Evaluate on these trees (e.g. `heiseticker:1-10000')
   -f        attenuation Factor (50)
   -h        read this
   -m        specify Minimal penalty for PP constraints (0.8)
   -q        solve the quadruple task instead of full parsing
   -s <spec> use these sources, e.g. `source1,souce2,source3'
   -t <num>  back off if word occurs fewer than N times in a source

" if $opt_h;

our @boost_factor;

# Hash of statistics sources. Each source is identified by its name
# and points to another hash with the following keys:
#
# sentences             # sentences in the source
# nrwords               # words in the source
# pp|<foo>|<bar>        # PP built from prep `foo' and the kernel `bar'
# word|<foo>            # occurrences of the word `foo'
# tag|<bar>             # occurrences of the tag `bar'
# wpf|<foo>|<bar>       # co-occurrences of word `foo' with prep `bar'
# tpf|<foo>|<bar>       # co-occurrences of tag `foo' with prep `bar'
# wpp|<x>|<y>|<z>       # triples of word x+prep y+kernel z
# tpp|<x>|<y>|<z>       # triples of tag x+prep y+kernel z
#
# This hash is actually tied to the DB file of the same name.
our %h;

# List of names of statistics sources. If more than one is given,
# they are processed in this order, not combined.
our @sources;

# Limit of occurrences before we switch to the next larger source
our $threshold = $opt_t || 1000;

# minimum value of any prediction
our $min_penalty = $opt_m || 0.8;

# factor by which the difference to 1 is attenuated
our $factor = $opt_f || 50;

our @sentence;

our %predictions;

sub debug {
  warn "DEBUG: @_\n" if $opt_d;
}

# Attach Perl hashes to database files.
sub read_stats {

  if ($opt_s) {
    @sources = split(/,/, $opt_s);
  } else {
    @sources = ('default');
  }

  for my $s (@sources) {
    my %hash;
    my $filename = $s;

    # search in the CWD
    $filename .= '.pp';

    # search alongside this script
    if (! -r $filename) {
      my $here = $0;
      while( -l $here) {
        $here = readlink $here;
      }
      $here =~ s!/[^/]+$!!;
      if(-r "$here/$filename") {
        $filename = "$here/$filename";
      }
      elsif($here =~ s!/bin!/lib! &&
         -r "$here/$filename") {
        $filename = "$here/$filename";
      }
    }

    # search in CDGDIR
    if(! -r $filename && exists($ENV{CDGDIR})) {
      if(-r "$ENV{CDGDIR}/lib/$filename") {
        $filename = "$ENV{CDGDIR}/lib/$filename";
      }
    }

    tie %hash, "DB_File", $filename, O_RDONLY
      or die "Can't open `$filename': $!\n";

    $h{$s} = \%hash;

  }

  @boost_factor = (1 x ($#sources+1));

  if ($opt_b) {
    if ($opt_b =~ /^\d+$/) {
      @boost_factor = ($opt_b x ($#sources + 1));
    } else {
      my $n = 0;
      my $spec = $opt_b;
      while ($spec =~ s/^(\d+),?//) {
        $boost_factor[$n++] = $1;
      }
    }
  }
}

# Read info printed by cdgp.
sub read_input {
  while (<>) {
    chomp;
    debug "read `$_'";

    # ignore all metadata
    next if /^\#[a-z]+:/;

    my($word, $tag) = split(/\t/);
    next unless defined $tag;

    my ($base, undef) = normalize_word ($word, $tag);
    my $prep = '';
    if($tag =~ /^(APPR|APPRART|PROAV|PWAV|KOKOM|APPO)$/) {
      $prep = base_prep $word;
    }

    if ($tag !~ /NE|NN/) {
      $base = deutsch_lc $base;
    }
    $base = supernormalize ($base, $tag);

    push @sentence,
      { word => $word,
        base => $base,
        tag => $tag,
        prep => $prep };
  }
}

sub guess_kernels {

  # Guess one kernel
  sub guess {
    my $i = shift;
    my $limit = shift;
    my $w = $sentence[$i];
    return unless $$w{tag} =~ /^APPR/;

    # already done?
    next if exists $$w{kernel};

    # assume the next noun is the kernel
    for my $j ($i+1..$limit) {
      my $x = $sentence[$j];

      # If we find another preposition while looking for kernel, this
      # might mean that the PP are truly nested, or that the first has a
      # non-noun kernel (`seit 2002', `bis dahin'). The latter is much
      # more likely, so we just repeat the search and limit it to this
      # position.
      if ($$x{tag} =~ /APPR|\$\.|FIN|INF|IZU|^KO/) {

        # make exception for pseudo-prepositions: `bis auf', `auer in' etc.
        if ($j == $i+1 &&
            $$x{word} =~ /^(ausser|auer|bis|statt)$/) {
          $$w{kernel} = $$x{word};
          return;
        }

        guess($i, $j-1);
        return;
      }

      if ($$x{tag} =~ /^(NN|NE|FM|P.*S|PPER|PRF|TRUNC)$/ &&
          $$x{word} !~ /^(Million|Milliarde|Billion)$/) {
        $$w{kernel} = $$x{word};
        return;
      }
    }

    # If we didn't find anything, we must try harder and allow other
    # categories such as ADV and stuff.
    for my $j ($i+1..$limit) {
      my $x = $sentence[$j];
      if ($$x{tag} =~ /^(CARD|ADJA|ADJD|ADV|PWAV|PROAV)$/) {
        $$w{kernel} = $$x{word};
        return;
      }
    }

  }

  for my $i (0..$#sentence) {
    guess($i, $#sentence);
  }

}

# Enrich the sentence with information that we can compute ourselves.
sub preprocess {

  # Reassemble prefixed verbs
  # Assume that the next PTKVZ belongs to each verb.
  # This is slightly uncertain, but usually a good thing to do because
  # verb prefixing changes the PP preferences just like the valence
  # frame.
  my $v = undef;
  for (@sentence) {
    if ($$_{tag} =~ /^(VVFIN|VVIMP)/) {
      $v = $_;
    }
    if ($v && $$_{tag} eq 'PTKVZ') {
      $$v{base} = $$_{base} . $$v{base};
      debug "reassemble `$$v{base}'";
      $v = undef;
    }
  }

  # Consider some auxiliary verbs full verbs like in PP-extract.pl
  # by detecting that no full verb appears after or shortly before it.
  for my $i (0..$#sentence) {
    my $w = $sentence[$i];
    if ($$w{tag} =~ /^VA/) {
      my $aux = 0;
      for my $j ($i-3..$i+40) {
        next unless $j >= 0 && $j <= $#sentence && $j != $i;
        my $w2 = $sentence[$j];
        if ($$w2{tag} =~ /^(V.INF|V.IZU|V.PP)$/) {
          $aux = 1;
          last;
        }
      }
      if (!$aux) {
        debug "$$w{word}($i) is quasi-full.";
        $$w{quasifull} = 1;
      }
    }
  }

  # Guess PP kernels
  if ($opt_3) {
    guess_kernels();
  }

}

sub _log {
  my $e = shift;
  my $n = shift;
  if ($n == 0) {
    return -1000;
  }
  return log($n)/log($e);
}

# Does a form of `werden' occur in a position where it could serve to
# form a passive together with word X?
sub find_werden {
  my $x = shift;

  for ($x-20 .. $x+3) {
    if ($_ >= 0 && $_ <= $#sentence) {
      if ($sentence[$_]{base} eq 'werden') {
        return 1;
      }
    }
  }
  return 0;
}

# Compute LA score for one word pair.
sub compute_score {

  my $g = shift;
  my $p = shift;
  my $kernel = shift;

  my $gov = $$g{base};
  my $prep = $$p{prep};

  my $source;
  my $score;

# Try the given sources in order.
 SOURCE:
  for my $k (0..$#sources) {
    my $last = ($k == $#sources);
    $source = $sources[$k];
    my $fx = $h{$source}{"word|$gov"};
    my $fy = $h{$source}{"word|$prep"};
    my $fpp = $h{$source}{"pp|$prep|$kernel"};
    my $pair = $h{$source}{"wpf|$gov|$prep"};
    my $tpair = $h{$source}{"tpf|$$g{tag}|$prep"};
    my $sentences = $h{$source}{sentences};

    # make sure all values are defined
    if (!defined $tpair) {
      $tpair = 0
    }
    if (!defined $pair) {
      $pair = 0
    }
    if (!defined $fx) {
      # write this warning only once per sentence
#          debug "Never saw word `$gov'" if $i == 0;
      $fx = 1;
    }
    if (!defined $fy) {
#          debug "Never saw preposition `$prep'" if $i == 0;
      $fy = 1;
    }

    if (defined $fpp) {
#          debug "PP $prep $kernel occurs $fpp times";
    } else {
#          debug "Never saw PP $prep $kernel";
      $fpp = 0;
    }

    # maybe backoff to the next source if the word is too rare
    if ($fx < $threshold && !$last) {
      next SOURCE;
    }

#        debug "f($gov)==$fx  f($prep)==$fy pair==$pair";

    # If there are no pair statistics, it might mean either that the
    # collocation is really unlikely or that the words themselves
    # are unlikely.
    if (!$pair) {
      my $expected = sprintf("%.2f", $fx * $fy / $sentences);
#          debug "f($gov/$prep) undefined (expected $expected)";

      # For `Sprecher gegenber' we should notice that it is really
      # the collocation that is rare, and not the components.
      if ($gov ne 'unknown' && $expected > 3) {
        $pair = 0;
      }

      # But for `Tischlerei jenseits' we expect only few
      # occurrences anyway, so back off to word/tag pairs
      else {
        debug "Backing off from $gov/$prep to $$g{tag}/$prep ($tpair)";
        $pair = $tpair;
        $fx = $h{$source}{"tag|$$g{tag}"};
      }
    }

    $score = $sentences * $pair / ($fx * $fy);
    if ($opt_b && $$g{tag} =~ /^(NN|NE)$/) {
      die if $score eq '';
      die "bad k == $k: boost_factor[$k] == $boost_factor[$k]" if $boost_factor[$k] eq '';
      $score = $score * $boost_factor[$k];
    }
    return ($score, $source, "$fx/$fy/$pair");
  }

}


# Munge the values so as to produce sensible predictions
sub make_predictions {

 PREP:
  for my $i (0..$#sentence) {
    my $p = $sentence[$i];
    next unless($$p{tag} =~ /^APPR|PROAV|KOKOM/);
    my $max = 0;

#    debug "Assume kernel $$p{prep} $$p{kernel}" if exists $$p{kernel};
    my $kernel = $$p{kernel} || '';

  GOV:
    for my $j (0..$#sentence) {
      next if $i == $j;
      my $g = $sentence[$j];
      next unless $$g{tag} =~ /^(VV|NN|NE|CARD|ADJ[AD])/ or
        exists$$g{quasifull};

      # Ooookay. We've got a possible PP subordination here.
      # Deduce a score from the data that we know.
      my $prep = $$p{prep};
      my $gov = $$g{base};

      my ($score,$source,$stat) = compute_score($g, $p, $kernel);

      # `von' and `durch' are passive morphemes. Therefore, if a
      # passive auxiliary+participle occur, raise the score for verb
      # attachment to at least unity.
      if ($score < 1 &&
          $$g{tag} =~ /PP/ &&
          $prep =~ /^(von|durch)$/ &&
          find_werden($j)) {
        $score = 1;
      }

      # Do not assign high scores to left-modification of nouns.
      # This is almost always wrong, and might wrongly set a false
      # maximum so that the real subordination is needlessly
      # punished. Instead, assign no score at all. That way, if the
      # subordination is possible after all, e.g. in a coordinate
      # cluster, it receives no penalty, but the second best
      # alternative receives no penalty either.
      if ($j > $i && $$g{tag} =~ /^(NN|NE|FM)$/) {
        next GOV;
      }

      debug sprintf("%5.2f  %15s %s  (from $source: $stat)", $score, $gov, $prep);

      # compute penalty
      my $pen = 1 - ( (2 - _log(3, $score) ) / $factor);
      $predictions{$i}{$j} = $pen;

      if ($pen > $max) {
        $max = $pen;
      }
    }

    # normalize and limit to a sane range
    for my $j (0..$#sentence) {
      if (exists $predictions{$i}{$j}) {
        my $x = $predictions{$i}{$j};
        if ($max) {
          $x /= $max;
        }
        if ($x > 1) {
          $x = 1;
        }
        if ($x < $min_penalty) {
          $x = $min_penalty;
        }
        $predictions{$i}{$j} = $x;
      }
    }

  }
}


# Talk to cdgp
sub write_predictions {
  my $i = 0;
  for my $w (@sentence) {
    print "$$w{word}";
    for my $j (0..$#sentence) {
      print "\t$j\t";
      if (exists $predictions{$i}{$j}) {
        printf "%.3f", $predictions{$i}{$j};
      } else {
        print '1';
      }
    }
    print "\n";
    $i++;
  }
}

# Read lines of quadruples and judge then in isolation
sub solve_quadruples {

  my ($good,$bad,$ngood,$nbad,$vgood,$vbad);

  while (<>) {
    my($v,$n1,$p,$sol) = /^(\S+)\s+(\S+)\s+(\S+)\s+([NV])/;
    if ($sol) {

      $v = deutsch_lc $v;
      $p = deutsch_lc $p;

      (my $vbase,undef) = normalize_word($v, 'VVFIN');
      (my $n1base,undef) = normalize_word($n1, 'NN');
      $n1base = supernormalize($n1base, 'NN');

      (my $pr,undef) = normalize_word($p, 'APPR');

      # judge noun attachment
      my $gov = { word => $n1, base => $n1base, tag => 'NN', prep => '' };
      my $prep = { word => $p, base => $pr, tag => 'APPR', prep => $pr };
      my ($nscore, undef, undef) = compute_score($gov,$prep,'');

      # judge verb attachment
      $gov = { word => $v, base => $vbase, tag => 'VVFIN', prep => '' };
      my ($vscore, undef, undef) = compute_score($gov,$prep,'');

      my $answer;
      if($nscore > $vscore) {
        $answer = 'N';
      } else {
        $answer = 'V';
      }

      if($sol eq $answer) {
        $good++;
        if($sol eq 'N') {
          $ngood++;
        } else {
          $vgood++;
        }
      } else {
        if($sol eq 'N') {
          $nbad++;
        } else {
          $vbad++;
        }
        $bad++;
        print "$v $n1 $p: predict $answer, correct $sol";
        printf "  (N %.2f V %.2f)\n", $nscore, $vscore;
      }
    }
  }

  printf "   on V: %.1f%% ($vgood:$vbad)\n", 100 * $vgood / ($vgood+$vbad);
  printf "   on N: %.1f%% ($ngood:$nbad)\n", 100 * $ngood / ($ngood+$nbad);
  printf "overall: %.1f%% ($good:$bad)\n", 100 * $good / ($good+$bad);

}

# measure how well our model predicts the actual state
sub eval_tree {
  my $file = shift;
  my $stat = shift;

  my $ae = cda2ae $file;
  return unless $ae;

  @sentence = ();

  # extract relevant data from .cda
  for (my $annos = AnnoEntryStruct_annos_get($ae);
       $annos;
       $annos = ListStruct_next_get($annos)) {
    my $anno = ListStruct_item_get($annos);
    my($word, $base, $tag, $prep, $label, $gov);
    $word = AnnotationStruct_word_get $anno;
    for (my $specs = AnnotationStruct_specs_get($anno);
         $specs;
         $specs = ListStruct_next_get($specs)) {
      my $spec = ListStruct_item_get($specs);

      if (SpecificationStruct_type_get($spec) eq $CDG::STTag) {
        my $att = SpecificationStruct_kind_get($spec);
        my $value = SpecificationStruct_name_get($spec);
        if('cat' eq $att) {
          $tag = $value
        } elsif('base' eq $att) {
          $base = $value
        } elsif('prep' eq $att) {
          $prep = $value
        }
      } else {
        my $level = SpecificationStruct_kind_get($spec);
        if('SYN' eq $level) {
          $label = SpecificationStruct_name_get($spec);
          $gov = SpecificationStruct_position_get($spec);
        }
      }
    }

    if(!$base) {
      $base = $word;
    }
    if(!$prep) {
      $prep = $base;
    }
    warn "no tag for `$word' in $file!\n" unless $tag;

    push @sentence,
    { word => $word,
      base => $base,
      tag => $tag,
      prep => $prep,
      label => $label,
      gov => $gov-1 };
  }
  commandEval 'reset';

  # Make the predictions
  preprocess();
  %predictions = ();
  make_predictions();

  # test against known structure
  my $i = 0;
  for(@sentence) {
    if($$_{tag} =~ /^APPR|PROAV|KOKOM/) {
      $$stat{no_tries}++;

      my $h = $predictions{$i};
      $$stat{no_alternatives} += scalar keys %$h;

      for my $g (keys %$h) {
        if(1 == $$h{$g}) {
          # and now, for the big moment:
          # is the predicted regent the actual regent?
          if($g == $$_{gov}) {
            $$stat{no_hits}++;
          }
          debug "$$_{word}--->$sentence[$g]{word}";
          if($g != $$_{gov}) {
            debug " (should be $sentence[$$_{gov}]{word})";
          }
          debug "\n";
          last;
        }
      }
    }
    $i++;
  }
}

sub eval_trees {

  my %stat =
    (
     no_tries        => 0,
     no_alternatives => 0,
     no_hits         => 0
    );

  commandEval 'set verbosity 0';

  $|=1;
  for my $spec (split(/,/, $opt_e)) {
    my $ae;
    if (-r $spec) {
      eval_tree($spec);
    } else {
      my ($min, $max, $corpus) = (0,1000, $spec);
      if ($spec =~ /^(.+):(\d+)-(\d+)/) {
        ($corpus,$min,$max) = ($1,$2,$3);
      }

      my $dir = "/data/corpus/ccc/$corpus";
      for my $i ($min..$max) {
        my $subdir = sprintf "%03d", ($i / 10000);
        my $file = "$dir/$subdir/$corpus-s$i.cda";
        next unless -f $file;
        print "$corpus-s$i\r";
        eval_tree($file, \%stat);
      }
    }
  }

  my $h = $stat{no_hits};
  my $m = $stat{no_tries} - $h;
  my $a = sprintf '%.1f', ($stat{no_alternatives} / $stat{no_tries});
  my $p = sprintf '%.2f', 100 * $h / $stat{no_tries};
  print "$p% ($h hits, $m misses, ambiguity $a)\n";
}


### MAIN ###

read_stats();

if($opt_e) {
  eval_trees();
  exit 0;
}

if ($opt_q) {
  solve_quadruples();
  exit 0;
}

read_input();
preprocess();
make_predictions();
write_predictions();
