#!/usr/bin/env perl
# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  
# cdg/utils/PP-extract.pl
#
# Extract cooccurrence statistics for content words and prepositions.
# Either CDA annotations or tagged text can be used as input.
BEGIN { use lib "$ENV{CDGDIR}/bindings", "$ENV{CDGDIR}/lib"; }

use warnings;
use strict;

use CDG;
use CDG::German;
use CDG::Utils qw(:statistics :functions);

use DB_File;

our $invocation = "$0 @ARGV";

our $interrupted = 0;
$SIG{INT} = sub { $interrupted++ };


use Getopt::Std;
our ($opt_c, $opt_h, $opt_u, $opt_v, $opt_3, $opt_n, $opt_r);
getopts('c:hn:ruv:3');

my $dbname = shift;

our $usage = "

Usage: $0 <database file name> < <tagged text file>
   -c <spec> use CDG annotations (default tagged text)

             SPEC can look like this: negra:1-1000,heiseticker:2000-4000
                        or like this: sentence1.cda,sentence2.cda

   -h        read this
   -n        specify window for assuming noun modification (default 1..3)
   -r        write Record of unknown nouns encountered
   -u        Unsupervised: ignore syntax structure even in .cda files
   -v        specify window for assuming Verb modification (default -3..3)
   -3        extract triples as well as pairs

";

die $usage if $opt_h;
die $usage unless defined $dbname;
die "Can't write $dbname: $!" if(-f $dbname && ! -w $dbname);

# input buffer
our @words = ();

our $sentences = 0;
our $words = 0;

# count occurrences of (normalized) words, tags, and co-occurrences
our %word_freq;
our %tag_freq;
our %pp_freq;

our %word_prep_freq;
our %tag_prep_freq;

our %word_pp_freq;
our %tag_pp_freq;


our $verb_window_start = -3;
our $verb_window_end = 3;

our $noun_window_start = 1;
our $noun_window_end = 3;

if($opt_v) {
  ($verb_window_start, $verb_window_end) =
    ($opt_v =~ /^([-0-9]+)\.\.([-0-9]+)$/)
      or die "Bad verb window spec: $opt_v\n";
}

if($opt_n) {
  ($noun_window_start, $noun_window_end) =
    ($opt_n =~ /^([-0-9]+)\.\.([-0-9]+)$/)
      or die "Bad noun window spec: $opt_v\n";
}


our %unknown_nouns;


# Enter a subordination into the frequency tables.
sub count {
  my $x = shift;
  my $y = shift;

  # pairs like `angestrichen von' are more likely passives than
  # collocations, ignore them
  if($$x{tag} =~ /PP$/ && $$y{word} =~ /^(durch|von)$/) {
    return;
  }

  my $kernel = $$y{kernel} || '';

  $word_prep_freq{$$x{word}}{$$y{word}}++;
  $tag_prep_freq{$$x{tag}}{$$y{word}}++;
  if($kernel) {
    $word_pp_freq{$$x{word}}{$$y{word}}{$kernel} ++;
    $tag_pp_freq{$$x{tag}}{$$y{word}}{$kernel} ++;
  }
}

# Print statistics about the input.
sub write_statistics {

  my %h;
  unlink $dbname;
  tie %h, "DB_File", $dbname or die "Can't write `$dbname': $!\n";

  $h{invocation} = $invocation;
  $h{sentences} = $sentences;
  $h{words} = $words;

  for (keys %word_freq) {
    $h{"word|$_"} = $word_freq{$_};
  }

  for (keys %tag_freq) {
    $h{"tag|$_"} = $tag_freq{$_};
  }

  for (keys %pp_freq) {
    $h{"pp|$_"} = $pp_freq{$_};
  }

  for my $x (keys %word_prep_freq) {
    for my $y (keys %{$word_prep_freq{$x}}) {
      my $pair = $word_prep_freq{$x}{$y};
      $h{"wpf|$x|$y"} = $pair;
    }
  }

  for my $x (keys %tag_prep_freq) {
    for my $y (keys %{$tag_prep_freq{$x}}) {
      my $pair = $tag_prep_freq{$x}{$y};
      $h{"tpf|$x|$y"} = $pair;
    }
  }

  if($opt_3) {
    for my $x (keys %word_pp_freq) {
      for my $y (keys %{$word_pp_freq{$x}}) {
        for my $z(keys %{$word_pp_freq{$x}{$y}}) {
          my $triple = $word_pp_freq{$x}{$y}{$z};
          $h{"w3|$x|$y|$z"} = $triple;
        }
      }
    }

    for my $x (keys %tag_pp_freq) {
      for my $y (keys %{$tag_pp_freq{$x}}) {
        for my $z(keys %{$tag_pp_freq{$x}{$y}}) {
          my $triple = $tag_pp_freq{$x}{$y}{$z};
          $h{"t3|$x|$y|$z"} = $triple;
        }
      }
    }
  }
}

# Find instances of PP attachment and count() them.
sub count_PP_attachments {

  if ($opt_c && !$opt_u) {
    for my $i (0..$#words) {
      my $w = $words[$i];

      if($$w{tag} =~ /^APPR|PROAV|KOKOM/ &&
         $$w{label} =~ /^(PP|OBJP|KOM|PRED)$/) {
        my $g = $words[$$w{gov}-1];
        count $g, $w;
      }

    }

  } else {

    for my $i (0..$#words) {
      my $w = $words[$i];

      next if $words[$i]{word} eq '<unk>';

      if ($$w{tag} =~ /APPR|PROAV|KOKOM/ && is_prep($$w{word})) {
        for my $j (0..$#words) {

          my $w2 = $words[$j];

          if ($$w2{tag} =~ /^VV/ |
               ($$w2{tag} =~ /^VA/ &&
                (exists $$w2{quasifull} ||
                 exists $$w2{prefix}))) {

            next if $i-$j < $verb_window_start;
            next if $i-$j > $verb_window_end;

            next unless is_verb($$w2{word});
            count($w2, $w);
          }
          if ($$w2{tag} eq 'NN') {
            next if $i-$j < $noun_window_start;
            next if $i-$j > $noun_window_end;
            if (is_noun($$w2{word})) {
              count($w2, $w);
            } else {
              $unknown_nouns{$$w2{word}}++;
            }
          }
          if ($$w2{tag} =~ /^(P.*S|PPER|CARD|FM|NE|TRUNC|ADJD)$/) {
            next if $i-$j < $noun_window_start;
            next if $i-$j > $noun_window_end;
            count($w2, $w);
          }
        }
      }
    }
  }
}

# The verbs `sein', `haben', `werden' are semantically empty when they
# occur with full verbs. Usually the grammar forces attachments to the
# full verb anyway. But sometimes they appear without full verb and
# then have a status that is similar to a full verb. Try to find these
# instances and mark them.
sub find_quasi_full_verbs {

  if($opt_c) {

    # Detect that an auxiliary verb has no AUX dependency.
    for (@words) {
      if ($$_{tag} =~ /^VA/) {
        my $aux = 0;
        for my $w(@words) {
          next unless $words[$$w{gov}-1] == $_;
          next unless $$w{label} eq 'AUX';
          $aux = 1;
          last;
        }
        if(!$aux) {
          $$_{quasifull} = 1;
        }
      }
    }
  }

  else {

    # Detect that no full verb appears after or shortly before it.
    for my $i (0..$#words) {
      my $w = $words[$i];
      if($$w{tag} =~ /^VA/) {
        my $aux = 0;
        for my $j($i-3..$i+40) {
          next unless $j >= 0 && $j <= $#words && $j != $i;
          my $w2 = $words[$j];
          if($$w2{tag} =~ /^(V.INF|V.IZU|V.PP)$/) {
            $aux = 1;
            last;
          }
        }
        if(!$aux) {
          $$w{quasifull} = 1;
        }
      }
    }
  }
}

# Try to reconstruct the true base form of split verbs.
sub fix_prefixed_verbs {

  if($opt_c) {
    for (@words) {
      if ($$_{label} eq 'AVZ') {
        my $w = $words[$$_{gov}-1];
        #          warn "reassembling `$$_{word}' + `$$w{word}'\n";
        $$w{word} = $$_{word} . $$w{word};
        $$w{prefix} = $$_{word};
      }
    }
  } else {
    my $v = undef;
    for (@words) {
      if ($$_{tag} =~ /^(VVFIN|VVIMP)/) {
        $v = $_;
      }
      if ($v && $$_{tag} eq 'PTKVZ') {
        $$v{word} = $$_{word} . $$v{word};
        $$v{prefix} = $$_{word};
        $v = undef;
      }
    }
  }
}

# Increase the unary %freq hashes.
sub count_words() {

  for (@words) {

    # Most words can be counted indifferently. Nouns and full verbs
    # are possible attachment sites, so we need them; and articles
    # etc. never are, so they will never have any pair counts, and
    # it's not necessary to filter them out. But auxiliaries are only
    # sometimes possible regents, so we must count only those
    # instances which function as full verbs; otherwise pairs such as
    # `sein gegen' or `werden zu' would be drastically underestimated.
    next if($$_{tag} =~ /^VA/ && !exists $$_{quasifull});

    $word_freq{$$_{word}}++;
    $tag_freq{$$_{tag}}++;
    if(exists $$_{kernel}) {
      $pp_freq{"$$_{word}|$$_{kernel}"}++;
    }
  }

}

# Try to pair prepositions with their noun kernels.
our($correct_kernels,$wrong_kernels);
sub guess_kernels {

  # Guess one kernel
  sub guess {
    my $i = shift;
    my $limit = shift;
    my $w = $words[$i];
    return unless $$w{tag} =~ /^APPR/;

    # already done?
    next if exists $$w{kernel};

    # assume the next noun is the kernel
    for my $j ($i+1..$limit) {
      my $x = $words[$j];

      # If we find another preposition while looking for kernel, this
      # might mean that the PP are truly nested, or that the first has a
      # non-noun kernel (`seit 2002', `bis dahin'). The latter is much
      # more likely, so we just repeat the search and limit it to this
      # position.
      if ($$x{tag} =~ /APPR|\$\.|FIN|INF|IZU|^KO/) {

        # make exception for pseudo-prepositions: `bis auf', `auer in' etc.
        if($j == $i+1 &&
           $$x{word} =~ /^(ausser|auer|bis|statt)$/) {
          $$w{kernel} = $$x{word};
          return;
        }

        guess($i, $j-1);
        return;
      }

      if ($$x{tag} =~ /^(NN|NE|FM|P.*S|PPER|PRF|TRUNC)$/ &&
          $$x{word} !~ /^(Million|Milliarde|Billion)$/) {
        $$w{kernel} = $$x{word};
        return;
      }
    }

    # If we didn't find anything, we must try harder and allow other
    # categories such as ADV and stuff.
    for my $j ($i+1..$limit) {
      my $x = $words[$j];
      if($$x{tag} =~ /^(CARD|ADJA|ADJD|ADV|PWAV|PROAV)$/) {
        $$w{kernel} = $$x{word};
        return;
      }
    }

    # give up: don't assign a kernel
  }


  if($opt_c) {

    # read kernels from `PN' labels
    for(@words) {
      if($$_{label} eq 'PN') {
        my $x = $words[$$_{gov}-1];
        $$x{kernel} = $$_{word};
      }
    }
  } else {

    # guess kernels
    for my $i (0..$#words) {
      guess($i, $#words);
    }
  }
}


# Deal with an entire sentence.
sub sentence {

  fix_prefixed_verbs();
  find_quasi_full_verbs();
  if($opt_3) {
    guess_kernels();
  }
  count_words();
  count_PP_attachments();

  @words = ();
  $sentences++;

}


# Print listing of things tagged `NN' that our lexicon doesn't have;
# it might be a good idea to add the most common ones.
sub write_unknown_nouns {

  open(UNK, ">unknown_nouns") or die "Can't write `unknown_nouns': $!\n";
  for (sort { $unknown_nouns{$b} <=> $unknown_nouns{$a} } keys %unknown_nouns) {
    print UNK "$unknown_nouns{$_}: $_\n";
  }
  close(UNK);
}

# Read syntax tree from a CDG annotation
sub read_cda {
 my $file = shift;

  my $id;
  my $i;
  my $form;
  my $word;
  my $cat;
  my $base = undef;
  my $gov;
  my $label;

  open(CDA, $file) or die "Can't read `$file': $!\n";

  $base = undef;
  while (defined (my $line = <CDA>)) {

    # recognize tree name
    if ($line =~ /^ *(\S+) *: *\S+ *<-> *$/) {
      $id = parse_string $1;
      print STDERR "[$id]\r";
    }

    # recognize word
    elsif ($line =~ /^ *(\d+) +(\d+) +(\S+)/) {
      $i = $1;
      $form = parse_string $3;
      $word = $form;
    }

    # recognize attributes
    elsif ($line =~ /^ *(\S+) *\/ *(\S+) *$/) {
      my($att, $val) = (parse_string($1), parse_string($2));
      if ($att =~ /^(base|baseform|infinitive)/) {
        $base = $val;
      } elsif ($att eq 'cat') {
        $cat = $val;
      }
    }

    # recognize subordination
    elsif ($line =~ /^ *(\S+) *-> *(\S+) *-> *(\d+)/) {
      my ($g, $level, $l);
      ($level,$l,$g) = (parse_string($1), parse_string($2), $3);
      if ($level eq 'SYN') {
        $gov = $g;
        $label = $l;
      }
    }

    # process word
    elsif ($line =~ /^ *[,;]/) {
      $words++;


      if ($cat =~ /^(APPR|PROAV|APPR|NN|NE|CARD|NE|V)/) {
        if (defined $base) {
          $word = $base
        } else {
          ($word, $cat) = normalize_word($word, $cat);
        }
        $word = supernormalize ($word, $cat);
      }

      $words[$i] = { form => $form,
                     word => $word,
                     tag => $cat,
                     gov => $gov,
                     label => $label };
      $base = undef;
    }

    # process entire tree
    if ($line =~ /^ *;/) {
      sentence();
    }
  }
  close(CDA);
}


### MAIN ###

# supervised mode: read CDG annotations
if($opt_c) {

  for my $spec (split(/,/, $opt_c)) {

    if (-r $spec) {
      read_cda $spec
    } else {

      my ($min, $max, $corpus) = (0,1000, $spec);
      if ($spec =~ /^(.+):(\d+)-(\d+)/) {
        ($corpus,$min,$max) = ($1,$2,$3);
      }

      my $dir = "/data/corpus/ccc/$corpus";
      for my $i ($min..$max) {
        my $subdir = sprintf "%03d", ($i / 10000);
        my $file= "$dir/$subdir/$corpus-s$i.cda";
        if (! -f $file) {
          $file= "$dir/$subdir/auto-$corpus-s$i.cda";
        }
        next unless -f $file;
        read_cda $file;
      }
    }
  }
}

# unsupervised mode: read tagged text
else {

  while (<>) {
    chomp;

    # process sentence
    if (/^$/) {
      sentence \@words;
      next;
    }

    # give progress info
    if (0 == $. % 10000) {
      print STDERR "[$.]\r";
    }

    # process word
    if (/^([^\t]+)\t([A-Z\$\.\,\(]+)/) {
      my($form, $tag) = ($1,$2);
      my $word = $form;

      # count
      $words++;

      ($word, $tag) = normalize_word($word, $tag);
      $word = supernormalize ($word, $tag);

      push @words, { form => $form, word => $word, tag =>$tag };

    } else {
      warn "Can't match `$_'\n";
    }
    last if $interrupted;
  }
}

write_statistics();
write_unknown_nouns() if $opt_r;

