#!/usr/bin/env perl
# cdg/utils/batch-eval.pl
#
# Read .cda files and generate error statistics.

## Boilerplate
use warnings;
use strict;
BEGIN {
  use File::Basename;
  unshift @INC, dirname(readlink($0) || $0);
  require 'cdgsetup.pl';
}
use CDG;
use CDG::Utils qw(:statistics :functions);

## Invocation
use Getopt::Std;
use vars qw ($opt_R $opt_g $opt_h $opt_l $opt_m $opt_r $opt_0 $opt_c);
getopts("0cg:hl:m:r:R:");
our ($first,$last);
our ($min,$max);

if($opt_l) {
  $opt_l =~ /(\d+)-(\d+)/;
  ($min,$max) = ($1,$2);
}

if($opt_r) {
  $opt_r =~ /(\d+)-(\d+)/;
  ($first,$last) = ($1,$2);
}

if($opt_R) {
  $opt_R =~ /(\d+)-(\d+)/;
  ($first,$last) = ($1,$2);
}

my $strlab = "";
if ($opt_c) {
	$strlab = "          struc+label corr";
}

our $mainlevel = $opt_m || 'SYN';

my %stat;
my $file = shift;

die "

Usage:  $0 <filename>

        <filename> contains the Annotations to be evaluated in .cda format.

        -0          count parses of mismatched length as 0 hits (default: skip)
	-c          show also correctness for structure AND label
        -g <file>   read Golden trees from file (default: from /data/corpus/ccc)
        -h          read this Help
        -l <i-j>    give range of permissible Lengths, e.g. `-l 1-40'
        -m <STRING> specify Main level (default `SYN')
        -r <i-j>    give Range of numbers to compare, e.g. `-r 1-1000'
        -R <i-j>    as -r, but count missing parses from that range as 0 hits'

           -R and -r are mutually exclusive.
           -R requires sentences parses be named <corpus>-s<number>.

" if ($opt_h or !defined $file or $opt_R && $opt_r);


## Data acquisition

# load parse results
commandEval "load $file";

# deduce annotation directory
our $dir;
{
  my $annotations = InputStruct_annotations_get($CDG::inputCurrentGrammar);
  my $x = hashListOfKeys $annotations;
  my $a = hashGet($annotations, ListStruct_item_get $x);
  my $id = AnnoEntryStruct_id_get $a;

  my($corpus,$n) = ($id =~ /([a-z]+)-s(\d+)/i);
  if(!defined $n) {
    die "Can't deduce annotation directory from parse name `id'!\n";
  }
  ($corpus,$n) = ($1,$2);
  $dir = sprintf '/data/corpus/ccc/%s', $corpus, $n / 10000;
}

# load annotations
if($opt_g) {
  commandEval "load $opt_g";
} else {
  our $ANNOS = "/tmp/$$.cdg";
  if(-d $dir) {
    system "cd $dir; find | grep cda\$ | grep -v auto- | xargs cat >> $ANNOS";
  } else {
    die "Can't read directory `$dir': $!\n";
  }
  commandEval "load $ANNOS";
  unlink $ANNOS;
  $opt_g = $ANNOS;
}

# canonicalize file name
if($opt_g !~ m!^/!) {
  our $cwd = `pwd`;
  chomp $cwd;
  $opt_g = "$cwd/$opt_g";
}

our $total_edges;
our $total_structure;
our $total_labelled;
our %all;

# make Perl hash of all trees
my $annotations = InputStruct_annotations_get($CDG::inputCurrentGrammar);
for(my $x = hashListOfKeys $annotations; $x; $x = ListStruct_next_get $x) {
  my $a = hashGet($annotations, ListStruct_item_get $x);
  my $id = AnnoEntryStruct_id_get $a;
  $all{$id} = $a;
}
print "Got ", scalar keys %all, " annos.\n";


## Processing proper

sub get_syntax_edge {
  my $anno = shift;

  for (my $l = AnnotationStruct_specs_get $anno; $l; $l = ListStruct_next_get $l) {
    my $spec = ListStruct_item_get $l;
  my $type = SpecificationStruct_type_get $spec;
  my $kind = SpecificationStruct_kind_get $spec;
    if ($type eq $CDG::STDep &&
        $kind eq $mainlevel) {
      return $spec;
    }
  }
  print "Falling off the edge!\n";
}


sub do_comparison {
  my($a,$ref) = @_;

  my (@test, @ref);

  # if $a is 0, just leave @test empty...
  if($a) {
    for (my $l = AnnoEntryStruct_annos_get $a; $l; $l = ListStruct_next_get $l) {
      push @test, ListStruct_item_get $l;
    }
  }
  for (my $l = AnnoEntryStruct_annos_get $ref; $l; $l = ListStruct_next_get $l) {
    push @ref, ListStruct_item_get $l;
  }

  # ... and it will be counted as a bunch of errors.
  if($#test != $#ref) {
    my $id = AnnoEntryStruct_id_get $ref;
    printf STDERR
    "WARNING: annotation `$id' has %d edges, parse has %d edges!\n",
    $#test+1, $#ref+1;
    if($opt_0 or
       $opt_R && -1 == $#test) {
      $total_edges += $#ref / 2;
    }
    return;
  }

  # suppress sentences with undesired lengths
  if($min) {
    return if $#test < $min;
    return if $#test > $max;
  }

  # compute matches
  for (0..$#test) {
    my $t = get_syntax_edge $test[$_];
    my $r = get_syntax_edge $ref[$_];
    $total_edges++;


    my $tlabel = SpecificationStruct_name_get $t;
    my $rlabel = SpecificationStruct_name_get $r;
    my $tparent = SpecificationStruct_position_get $t;
    my $rparent = SpecificationStruct_position_get $r;


    $stat{$rlabel}{occurred}++;
    $stat{$tlabel}{predicted}++;

    if ($tlabel eq $rlabel) {
      $stat{$tlabel}{correct}++;
    } else {
      $stat{$rlabel}{missed}++;
      $stat{$tlabel}{false}++;
    }

    # structural hit?
    if ($tparent == $rparent) {
      $total_structure++;

      # count structural accuracy by reference label
      $stat{$rlabel}{structure}++;

      # labelled hit?
      if($tlabel eq $rlabel) {
        $total_labelled++;
	$stat{$rlabel}{struclabel}++;
      }
    }
  }
}

our %seen;
our $corpus= '';
ANNO:
for my $a (values %all) {


  my $id = AnnoEntryStruct_id_get $a;
  my $lat = AnnoEntryStruct_lattice_get $a;
  my $filename = AnnoEntryStruct_filename_get $a;

  # don't judge the reference trees themselves
  next if $filename eq $opt_g;

  # don't measure stuff outside $opt_r
  if($first) {
    if($id =~ /(\w+)-s(\d+)/) {
      $corpus = $1;
      my $no = $2;
      $seen{$no}++;
      next if $no < $first;
      next if $no > $last;
    }
  }

  # find reference tree
  my $ref = $all{$lat};

  # If there is no tree with the same name as this sentence, try to
  # find a tree that was (1) loaded from the `gold' file and (2)
  # professes to annotate this sentence.
  if(!$ref) {
    for my $x (values %all) {
      if(AnnoEntryStruct_filename_get($x) eq $opt_g &&
         AnnoEntryStruct_lattice_get($x) eq $lat) {
        $ref = $x;
        last;
      }
    }
  }

  if (!$ref) {
    warn "WARNING: No reference annotation found for tree $id!\n";
    next ANNO;
  }

  do_comparison($a, $ref);

}

# under -R, if there were too few parses, count missing parses as errors
if($opt_R &&
   $last - $first + 1 > scalar keys %seen) {

  for my $i($first..$last) {
    next if $seen{$i};
    my $ref = $all{"$corpus-s$i"};
    do_comparison(0, $ref);
  }
}


## Ensure defined values everywhere
for(keys %stat) {

  if(!exists $stat{$_}{correct}) {
    $stat{$_}{correct} = 0;
  }
  if(!exists $stat{$_}{occurred}) {
    $stat{$_}{occurred} = 0;
  }
  if(!exists $stat{$_}{predicted}) {
    $stat{$_}{predicted} = 0;
  }
  if(!exists $stat{$_}{structure}) {
    $stat{$_}{structure} = 0;
  }

  if(0 == $stat{$_}{occurred}) {
    $stat{$_}{errors} = 0;
  } else {
    $stat{$_}{errors} = $stat{$_}{occurred} - $stat{$_}{structure};
  }
}

## Output

sub print_percentage {
  my ($a, $b) = @_;
  my $p = 1;
  $a+=0;
  if($b) {
    $p = $a / $b;
  }
  printf " %5.1f (%5d/%5d)  ", 100*$p,$a,$b;
}

print "\n       structure (by gold label)   label precision       label recall$strlab\n\n";
for(sort {$stat{$b}{errors} <=> $stat{$a}{errors}} keys %stat) {
  printf "%6s:     ", $_ || "''";
  print_percentage $stat{$_}{structure}, $stat{$_}{occurred};
  print_percentage $stat{$_}{correct}, $stat{$_}{predicted};
  print_percentage $stat{$_}{correct}, $stat{$_}{occurred};
  if ($opt_c) {
    print_percentage $stat{$_}{struclabel}, $stat{$_}{occurred};
  }
  print "\n";

}
printf "\n\n Total structure:";
print_percentage $total_structure, $total_edges;
printf "\n\n Total labelled:";
print_percentage $total_labelled, $total_edges;
print "\n\n";
