#!/usr/bin/env perl
# Copyright (C) 1997-2006 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
# 
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# 
# extract sentences from negra export format or lattice format in different formats
# author: Klaus Dalinghaus <kdaling@nats.informatik.uni-hamburg.de>
# initial revision: 2006-04-10

use Getopt::Long;
use strict;

my $help;
my $debug;
my $first_sentence = 1;
my $last_sentence = "last";
my $negra;
my $mask;
my $space;
my $word;
my $length = 0;

GetOptions(
	"help"     => \$help,
	"debug"    => \$debug,
	"begin:i"  => \$first_sentence,
	"end:i"    => \$last_sentence,
	"negra"    => \$negra,
	"mask"     => \$mask,
	"space"    => \$space,
	"length:i" => \$length,
	"word"     => \$word);

my $infile = shift;
my $outfile = shift;

my $error;
my $input;
my $output;

if ($infile) {
	open(INPUT, "<$infile") or $error .= "\n\nError: can not open $infile for reading!";
	print STDERR "open file \"$infile\" for reading.\n" if ($debug);
	$input = *INPUT;
} else {
	$input = *STDIN;
	print STDERR "read from STDIN.\n" if ($debug);
}
if ($outfile) {
	open(OUTPUT, ">$outfile") or $error.="\n\nError: can not open $outfile for writing!";
	print STDERR "open file \"$outfile\" for writing.\n" if ($debug);
	$output = *OUTPUT;
} else {
	$output = *STDOUT;
	print STDERR "write to STDOUT.\n" if ($debug);
}

						

die "$error

Usage:	$0 [input-file [output-file]]

	input-file	contains sentences
	output-file	file to write sentences in

	-help		read this help
	-debug		print debug information on STDERR
	-begin number	number of first sentence (default: 1)
	-end number	number of last sentence (default: last)
	-negra		the input is in negra export format (default: input in lat format)
	-length number  the maximum length for a sentence (longer sentences are printed as '. . . . .')
			(default: 0 which means no restriction)
	-word		print one word per line (default: one sentence per line)
	-mask		mask characters '(' and ')'
	-space		mask spaces with *SP*

" if ($help || $error);


my $inpart = 0;
my $first = 0;
my $linenr = 0;
my $s_cnt;

print STDERR "starting processing input file for sentences $first_sentence to $last_sentence.\n" if ($debug);

if ($negra) {
	my $sentence;
	my $cnt;
	my $sn;
	while (my $line = <$input>) {
		chomp $line;
		$linenr++;
		if ($line =~ /^\s*%/ || $line =~ /^\s*$/) { # handle comment
		} elsif ($line =~ /^#BOS\s+([0-9]+)\s+[0-9]+\s+[0-9]+\s+[0-9]+\s*$/) { # start a sentence
			$sn = $1;
			$first = 1;
			$cnt = 0;
			if (! $inpart && $sn eq $first_sentence) {
				$inpart = 1;
				$s_cnt = $sn;
				print STDERR "found begin of sentence no. $first_sentence.\n" if ($debug);
			}
			if ($inpart) {
				if ($s_cnt ne $sn) {
					die "ERROR: expected sentence no. $s_cnt but found no. $sn!\n";
				}
				print STDERR "sentence $sn  \r";
			}
		} elsif ($line =~ /^#EOS\s+([0-9]+)\s*$/) { # end a sentence
			if ($inpart) {
				$sentence .= "\n";
				if ($length && $cnt > $length) {
					print STDERR "WARNING: sentence no. $sn has more than $length words ($cnt) and is deleted.\n";
					if ($word) {
						print $output ".\n.\n.\n.\n.\n\n";
					} else {
						print $output ". . . . .\n";
					}
				} else {
					print $output $sentence;
				}
				$s_cnt++;
			}
			if ($last_sentence ne "last" && $1 >= $last_sentence) {
				$inpart = 0;
				print STDERR "found end of sentence no. $last_sentence.\n" if ($debug);
				last;
			}
		} elsif ($line =~ /^([^\s]+)\s.*$/) { # process line
			if ($inpart) {
				my $w = $1;
				if ($w =~ /^#/) {
					next;
				}
				if ($w =~ /^([\'\"])(.*)\1$/) {
					$w = $2;
					$w =~ s|\\||g;
				}
				$w =~ s/\"a//g;
				$w =~ s/\"o//g;
				$w =~ s/\"u//g;
				$w =~ s/\"A//g;
				$w =~ s/\"O//g;
				$w =~ s/\"U//g;
				$w =~ s/\"s//g;
				if ($space) {
					$w =~ s/ /*SP*/g;
				}
				if ($mask) {
					$w =~ s/\(/*LRB*/g;
					$w =~ s/\)/*RRB*/g;
				}
				if ($first) {
					$first = 0;
					$sentence = $w;
					if ($word) {
						$sentence .= "\n";
					}
				} else {
					if ($word) {
						$sentence .= $w . "\n";
					} else {
						$sentence .= " " . $w;
					}
				}
				$cnt++;
			}
		} else {
			print STDERR "syntax error in input file in line $linenr\n";
		}
				
	}
} else {
	while (my $line = <$input>) {
		chomp $line;
		$linenr++;
		if ($line =~ /^\s*$/) {
			next;
		} elsif ($line =~ /^\s*'[^'0-9]*([0-9]+)'\s*:\s*([^\s].*);\s*$/) {
			my $sn = $1;
			my $sent = $2;
			if (! $inpart && $sn eq $first_sentence) {
				$inpart = 1;
				$s_cnt = $sn;
				print STDERR "found begin of sentence no. $first_sentence.\n" if ($debug);
			}
			if ($inpart) {
				if ($s_cnt ne $sn) {
					die "ERROR: expected sentence no. $s_cnt but found no. $sn!\n";
				}
				print STDERR "sentence $sn  \r";
				my @wo = split(/,\s+/, $sent);
				my $first = 1;
				my $sentence;
				my $cnt = 0;
				foreach my $w (@wo) {
					if ($w =~ /^([\'\"])(.*)\1$/) {
						$w = $2;
						$w =~ s|\\||g;
					}
					$w =~ s/\"a//g;
					$w =~ s/\"o//g;
					$w =~ s/\"u//g;
					$w =~ s/\"A//g;
					$w =~ s/\"O//g;
					$w =~ s/\"U//g;
					$w =~ s/\"s//g;
					if ($space) {
						$w =~ s/ /*SP*/g;
					}
					if ($mask) {
						$w =~ s/\(/*LRB*/g;
						$w =~ s/\)/*RRB*/g;
					}
					if ($first) {
						$first = 0;
						$sentence = $w;
						if ($word) {
							$sentence .= "\n";
						}
					} else {
						if ($word) {
							$sentence .= $w . "\n";
						} else {
							$sentence .= " " . $w;
						}
					}
					$cnt++;
				}
				$sentence .= "\n";
				if ($length && $cnt > $length) {
					print STDERR "WARNING: sentence no. $sn has more than $length words ($cnt) and is deleted.\n";
					if ($word) {
						print $output ".\n.\n.\n.\n.\n\n";
					} else {
						print $output ". . . . .\n";
					}
				} else {
					print $output $sentence;
				}
				$s_cnt++;
			}
			if ($last_sentence ne "last" && $sn >= $last_sentence) {
				$inpart = 0;
				print STDERR "found end of sentence no. $last_sentence.\n" if ($debug);
				last;
			}
		}
	}
}

print STDERR "finishing program.\n" if ($debug);

close $input;
close $output;

exit 0;
