#!/usr/bin/perl

# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# cdg/utils/lat2tok.pl
#
# This script converts from lat-foramt to tok-format.
# The output is written to standard output.
#
# Usage: lat2tok.pl [-dh] [FILE]...
#
# author: Yvonne Kuestermann <yvonne@nats.informatik.uni-hamburg.de>

use strict;
use warnings;

use Getopt::Std;
our $opt_d; # debug
our $opt_h; # help
getopts("dh");

# Help
die "
Usage: lat2tok.pl [-dh] [FILE]...

Convert the FILEs from lat-foramt to tok-format.

  -d   print debug output
  -h   display this help and exit

" if $opt_h;


###################################################################
###################################################################
# Variables: Regular expressions


# The prefix of a lat-sentence.
# I.e.: 'Beispiele-s154' :
my $r_prefix = qr /'[^']+   # the name of the sentence (everything allowed but a ')
                \d+'        #     ending on the sentence number
                [ ]*:[ ]*   # colon
	       /x;

# The whole lat-sentence.
# $1: the bare sentence
my $r_sentence = qr /^[ ]*$r_prefix  # the sentence name
	          (.*)               # the bare sentence
		  [ ]*;[ ]*$         # the closing semicolon
		 /x;

# A bare word.
# I.e: word
my $r_word = qr /(?:      # combination of:
              [^']*       #   allowed letters
	      (?: \\' )*  #   quoted '
              )* /x;

# Words in lat-syntax:

# An unquoted word in lat-syntax.
# I.e.: word, 
# $1: the bare word without lat-syntax
my $r_latWord = qr /([^,']*)  # a word consisting of 
                              # any characters but comma
                    ,[ ]*     # a comma and optional spaces
                   /x;

# A quoted word in lat-syntax.
# I.e.: 'word\'s', 
# $1: the bare word without lat-syntax
#     but \' is still quoted
my $r_quotedLatWord = qr /\'($r_word)\'  # a word in ''
                          ,[ ]*          # a comma and optional spaces
                         /x;

###################################################################
###################################################################
# main
sub main {
  while ( <> ) {
    chomp;
    my $line = $_;
    next if $line eq "";
    
    # print for debug puposes
    print "\n" if $opt_d;
    print "-------------------\n -d" if $opt_d;
    print "die Zeile:             \"$line\"\n" if $opt_d;

    # Extract the sentence and 
    # add a comma to the last word,
    # so that every word is followed by a comma.
    unless ( $line =~ s/$r_sentence/$1, /gx ) {
      warn "Not conform to lat-syntax: $line\n";
      next;
    }
    print "der bloe lat-Satz:    \"$line\"\n" if $opt_d;

    # Replace the words in lat-syntax with normal "words"
    $line =~ s/ ( $r_quotedLatWord | $r_latWord ) /&output($2,$3)/gxe;
    print "\n\n";
  }
}

###################################################################
# Unquote one word.
# \param the word
sub unquote {
  my $expr = shift;

  # unquote every character but \ itsself
  $expr =~ s/ \\ (?! \\ ) //gx;
  # unquote a backslash
  $expr =~ s/ \\\\ /\\/gx;

  return $expr;
}

###################################################################
# Output one word per line.
# \param the word
sub output {
  my $word = shift;
  $word = shift unless defined $word;
  $word = &unquote($word);
  print "$word\n";
}


###################################################################
###################################################################

# Main
&main;
1;
