# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## AllFiles - CDG grammar file manager.
## This class is responsible for all issues regarding grammar files in the xcdg
## application, that is loading, reloading grammar files and xml annotaion files.
##
## \author Michael Daum (see also AUTHORS and THANKS for more)
##
## ----------------------------------------------------------------------------
class AllFiles {
    inherit DataBrowser

   # public methods
   public method init_data {}; ## abstract Method from DataBrowser
   public method load {args}; ## \type TclList
   public method load_dir {args}; ## \type TclList
   public method loadXml {filename}; ## \type TclString
   public method reload {}
   public method selectIdsOfFile {filename}; ## \type TclString

    constructor {args} {}; ## \type TclList


   # protected methods
   protected method _motion_action {w x y }; ## \type TclWidget,TclNumber,TclNumber

   # private methods
   private method _reload_action {}
   private method _load_action {}
   private method _run_action {}
   private method _reset_action {}
   private method _edit_action {}
   private method _load {args}; ## \type TclList
   private method _dom2anno {document}; ## \type DomNode

   # private variables

   ## array holding all loaded files.
   private variable _filenames; ## \type TclArray

   ## list if recently loaded files
   private variable _lastLoaded ""
};


## ----------------------------------------------------------------------------
## An AllFiles constructor.
## \param args arguments passed to itk_initialize
## ----------------------------------------------------------------------------
body AllFiles::constructor {args} {
   itk_component add loadbutton {
      button $itk_component(childsite).load \
		-anchor w \
		-text "Load" \
		-command [code $this _load_action]
   } { }

   itk_component add runbutton {
      button $itk_component(childsite).run \
		-anchor w \
		-text "Run" \
		-command [code $this _run_action]
   } { }

   itk_component add resetbutton {
      button $itk_component(childsite).reset \
		-anchor w \
		-text "Reset" \
		-command [code $this _reset_action]
   } { }

   itk_component add editbutton {
      button $itk_component(childsite).edit \
		-anchor w \
		-text "Edit" \
		-command [code $this _edit_action]
   } { }

   itk_component add reloadbutton {
      button $itk_component(childsite).reload \
		-anchor w \
		-text "Reload" \
		-command [code $this _reload_action]
   } { }

   pack $itk_component(loadbutton) \
	$itk_component(runbutton) \
	$itk_component(resetbutton) \
	$itk_component(editbutton) \
	$itk_component(reloadbutton) \
	-side left -padx 3 

  $itk_component(table) configure -titlerows 2

   # register the helpmessages
   .cdgmain help sethelpstr \
      $itk_component(loadbutton) \
	 "Load a new cdg-file" \
      $itk_component(runbutton) \
	 "Execute a tcl-cdg-script" \
      $itk_component(resetbutton) \
	 "Unload all files and destroy all related data" \
      $itk_component(entryfield) "Enter a filename" 

    init_data
}

## ----------------------------------------------------------------------------
## select row of a file.
## previous selections are cleared
## ----------------------------------------------------------------------------
body AllFiles::selectIdsOfFile {filename} {
   set table $itk_component(table)

   set rmax [$table cget -rows]
   $table selection clear 0,0 end
   
   set newselection ""
   for {set row 0} {$row < $rmax} {incr row} {
      if {[$table getCell $row,1] == $filename} {
	 set id [$table getCell $row,1 ]
	 lappend newselection $id
	 $table selection set $row,0
      }
   }

   set _selection $newselection
}


## ----------------------------------------------------------------------------
## load files and store additional information which could be acquired
## ----------------------------------------------------------------------------
body AllFiles::load {args} {

   set table $itk_component(table)

   if {$args == "" } {
      set types {
	 {"CDG files"		{.cd? .m4}}
	 {"CDG category"	{.cdc}}
	 {"CDG lexicon"		{.cdl}}
	 {"CDG sorts"		{.cds}}
	 {"CDG wordgraphs"	{.cdw}}
	 {"CDG scripts"		{.run .scr}}
	 {"CDG Xml"		{.xml.gz}}
	 {"All files"		*}
      }
      set names [tk_getOpenFile -filetypes $types -parent .]
      if {$names == ""} { 
	return 
      }
   } else {
      set names $args
   }

   # test arguments
   foreach filename $names {
      if {[catch {glob $filename}]} {
	 ::cmd::Puts "ERROR: can't find file \"$filename\""
	 return
      }
   }

   # load set of files
   set _lastLoaded $names
   set basenames [_load $names]
   .cdgmain busy compute "Init Files" [code $this init_data]
   return $basenames
}

## ----------------------------------------------------------------------------
## load all files in a directory.
## \todo this method does not use the busy dialog.
## ----------------------------------------------------------------------------
body AllFiles::load_dir {args} {

  set dir [tk_chooseDirectory -parent . -mustexist 1 -title "Load directory"]
  if {$dir == ""} { 
    return 
  }

  # load set of files
  set files [glob -type f -path "$dir/" *]
  commandEval "load $files"
  init_data
}

## ----------------------------------------------------------------------------
## internal load method.
## This method is called while the CdgBusy box is shown.
## \param args list of filenames to be loaded
## \returns the list of the filenames basenames
## ----------------------------------------------------------------------------
body AllFiles::_load {args} {
  set basenames ""
  foreach filename $args {
    set basename [file tail $filename]
    set basename [join [split $basename] "\n"]
    if {[regexp {.*\.xml\.gz} $basename]} {
      .cdgmain busy compute "Loading $basename" \
	[code $this loadXml $filename]
    } else {
      .cdgmain busy compute "Loading $basename" commandEval "load $filename"
    }
    lappend basenames $basename
  }

  # switch the notebook to the first page (files),
  # because the other pages may have become invalid,
  # but are only redrawn upon an expose event.
  .cdgmain tabno view 0

   return $basename
}

## ----------------------------------------------------------------------------
## load the last loaded file again
## ----------------------------------------------------------------------------
body AllFiles::reload {} {
   _reload_action
}


## ----------------------------------------------------------------------------
## initialization of the data managed by this class.
## This method refreshes the data which could have been changed elsewhere, that
## is consult the C layer and squeeze out the relevant information.
## ----------------------------------------------------------------------------
body AllFiles::init_data {} {
  global inputCurrentGrammar

  set table $itk_component(table)
  $table configure -state normal
  $table clear all
  $table delete rows 3 [$table cget -rows]
  $table tag configure filename -anchor w
  $table tag col filename 1
  $table width 0 9 1 18 3 9 4 9 5 9 6 9 7 9 8 9
  $table setCell row 0,0 { "" "Filename" "Constraints" "Levels" \
       	"Lex.Entries" "Lattices" "Annotations" "Hierarchies" "Parameters" \
       	"Total" }
  _setCount [listSize [InputStruct_files_get $inputCurrentGrammar]]

  array unset _filenames 

  # redraw table
  set count 2
  for {set l [InputStruct_files_get $inputCurrentGrammar]} {$l != "NULL"} \
      {set l [ListStruct_next_get $l]; incr count} {
     set filename [pointer2string [ListStruct_item_get $l]]
     set basename [file tail $filename]
     set _filenames($basename) $filename
     $table setCell $count,0 [expr $count -1]
     $table spans $count,1 0,8
     $table setCell $count,1 $basename
  }

  # count stuff
  $table setCell 1,2 [hashSize [InputStruct_constraints_get $inputCurrentGrammar]] \
  1,3 [listSize [InputStruct_levels_get $inputCurrentGrammar]] \
  1,4 [hashSize [InputStruct_lexicon_get $inputCurrentGrammar]] \
  1,5 [listSize [InputStruct_lattices_get $inputCurrentGrammar]] \
  1,6 [hashSize [InputStruct_annotations_get $inputCurrentGrammar]] \
  1,7 [hashSize [InputStruct_hierarchies_get $inputCurrentGrammar]] \
  1,8 [listSize [InputStruct_parameters_get $inputCurrentGrammar]]
  $table configure -state disabled
}

## ----------------------------------------------------------------------------
## load slot.
## This command is executed whenever the load button is pressed. It loads
## the selected files in AllFiles::_selection and sets the selection accordingly.
## ----------------------------------------------------------------------------
body AllFiles::_load_action {} {

   set files [eval load $_selection]
   eval setSelection $files
}

## ----------------------------------------------------------------------------
## reload slot.
## This command is executed whenever the load button is pressed. It resets the
## cdg system before loading the files in AllFiles::_lastLoaded.
## ----------------------------------------------------------------------------
body AllFiles::_reload_action {} {
  ::cmd::Reset force
   set files [eval load $_lastLoaded]
   eval setSelection $files
}

## ----------------------------------------------------------------------------
## select a file and run it as a tcl-cdg-script
## ----------------------------------------------------------------------------
body AllFiles::_run_action {} {

   set types {
      {"CDG scripts"	{.run .scr}}
      {"CDG files"	{.cd? .m4}}
      {"CDG category"	{.cdc}}
      {"CDG lexicon"	{.cdl}}
      {"CDG sorts"	{.cds}}
      {"CDG wordgraphs"	{.cdw}}
      {"All files"	*}
   }
   set filename [tk_getOpenFile -filetypes $types -parent .]
   if {$filename != ""} { 
     .cdgmain shell safeSource $filename
   }
}

## ----------------------------------------------------------------------------
## edit slot.
## This method is called on and edit button press and starts an editor on the
## selected filenames.
## ----------------------------------------------------------------------------
body AllFiles::_edit_action {} {

   set files ""
   foreach id $_selection {
      append files " $_filenames($id)"
   }
   
   regsub -all "%f" "/bin/sh -c \"[.cdgmain cget -editor]\"" $files command

   if {[catch {eval exec $command >/dev/null &} errMsg]} {
      ::cmd::Puts "ERROR: $errMsg"
   }
}

## ----------------------------------------------------------------------------
## reset slot.
## This method simply calls ::cmd::Reset.
## ----------------------------------------------------------------------------
body AllFiles::_reset_action {} {
   ::cmd::Reset
}


## ----------------------------------------------------------------------------
## motion slot.
## Moving the mouse over the table extracts displays a help string in the 
## status line of the main window.
## \param w the widget where the motion was detected
## \param x the x coords of the mouse 
## \param y the y coords of the mouse
## ----------------------------------------------------------------------------
body AllFiles::_motion_action {w x y} {
   focus $w

   set row [$w index @$x,$y row]
   if {$row < 1} return

   set file [$itk_component(table) getCell $row,1]
   if {[info exists _filenames($file)]} {
      set fullfilename $_filenames($file)
      .cdgmain help showstr $fullfilename
   }
}



## ----------------------------------------------------------------------------
## read in an xml annotation.
##
## \param filename the xml file containing the xml annotation
##
## The provided file is read, and the xml document is searched for an
## &lt;annotation&gt; .... &lt;/annotation&gt;. This was hopefully generated with
## writeXmlAnnoEntry() . 
## ----------------------------------------------------------------------------
body AllFiles::loadXml {filename} {

  # open the file and read in the raw data
  set fd [open "|zcat $file" r]
  set data [read $fd]
  close $fd

  # analyse the dom document
  set document [dom parse $data]
  set buffer [_dom2anno $document]
  $document delete
  set document ""

  # store the annotation into a temporary file
  set tempfile [exec mktemp /tmp/xcdg.XXXXXX]
  set fd [open $tempfile w]
  puts $fd $buffer
  close $fd

  # load it
  commandEval "load $tempfile"

  file delete $tempfile
}

## ----------------------------------------------------------------------------
## analyse a dom document and convert it to an annotation.
## 
## @param document the dom document
## @returns a string buffer containing the annotation in the cdg 
##          annotation format
##
## This method mainly is called by @ref loadXml .
## ----------------------------------------------------------------------------
body AllFiles::_dom2anno {document} {

  set result ""

  # get the root element
  set cdgpNode [$document documentElement]
  if {$cdgpNode == ""} {
    error "ERROR: no cdgp logfile\n"
  } 

  # get the annotation node
  set annoNode [$cdgpNode descendant 1 "annotation"]
  if {$annoNode == ""} {
    error "ERROR: no annotation found in this document\n"
  }

  # get annotation and latice ids
  set id "[$annoNode @id]-xml"

  # backwards compatibility
  if {[catch { set latticeId "[$annoNode @lattice]" }]} {
    set latticeId [$cdgpNode selectNodes \
      {//parse[1]/ref[@type="Lattice"]/@name} ]
  }

  # print all infos in this dom node
  append result "'$id' : '$latticeId' <->\n"
  set isFirst 1
  foreach arcNode [$annoNode selectNodes "arc"] {

    if {$isFirst} {
      set isFirst 0
    } else {
      append result ",\n"
    }
    
    # get word info
    set from [$arcNode @from]
    set to [$arcNode @to]
    set word [$arcNode @word]
    regsub -all {'} $word {\'} word

    append result [format "%3d %3d '%s' " $from $to $word]
    foreach tagNode [$arcNode selectNodes "tag"] {
      foreach attr [$tagNode attributes] {
	append result "'$attr'/'[$tagNode getAttribute $attr]' "
      }
    }
    append result "\n"

    # get dependency info
    set levels ""
    foreach depNode [$arcNode selectNodes "dep"] {
      set level [$depNode @level]
      set label [$depNode @label]
      set modifiee [$depNode @modifiee]
      set arc($level) [list $label $modifiee]
      if {[lsearch -exact $levels $level] < 0} {
	lappend  levels $level
      }
    }


    # print dependency info
    foreach level $levels {
      set label [lindex $arc($level) 0]
      set modifiee [lindex $arc($level) 1]
      append result [format "%10s -> %-10s -> %3d\n" \
	$level "'$label'" $modifiee ]
    }

    array unset arc
  }
  append result ";\n"

  return $result
}

# ----------------------------------------------------------------------------
