# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## CdgBusy - indicate work in progress
##
## This class is used to let the `busy' light blink in the status bar
## while a computationally expensive operation is scheduled in the
## background. The light keeps blinking untip the background computation has
## finished. The user can click on the light to terminate the
## background computation terminate earlier. 
##
## Actually, to let the user do that the background computation has to
## be written cooperatively, as we have no real concurrency in tcl. 
## This is accomplished by a frequent call to "update" somewhere in
## your computation loop. Also, only libcdg functions that obey the
## cdgCtrlCTrapped flag can be interrupted, not GUI commands.
##
## The busy dialog should also be closed in case of any errors. This is
## accomplished by overwriting the tcl commands "error", "catch" and
## "bgerror" in \ref error.tcl.
##
## \author Michael Daum (see also AUTHORS and THANKS for more)
## ----------------------------------------------------------------------------
class CdgBusy {
  inherit ::itk::Widget

  # public methods
  public method compute {text args}; ## \type TclString, TclList
  public method interrupt {args}; ## \type TclList
  # public variables
  public variable wait 500

  constructor {args} {}; ## \type TclList

  # private methods
  private method blink {}
  private method reset {}
  private method enterAction {w x y}; ## \type TclWidget, TclNumber, TclNumber
  private method leaveAction {}

  # private variables
  private variable computeText ""; ## \type TclString
  private variable oldCursor ""; ## \type TclString
  private variable ledState 0; ## \type Boolean
  private variable isActive 0; ## \type Boolean
  private variable blinkJob ""; ## \type AfterJob
};


## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body CdgBusy::constructor {args} {
  global env
  
  set isActive 0
  set ledState 0

  image create photo led_off -file $env(XCDG_LIB)/g_grey.gif
  image create photo led_on -file $env(XCDG_LIB)/g_red_on.gif

  itk_component add intButton {
   button $itk_interior.intButton\
       -anchor w \
       -image led_off \
       -command [code $this interrupt] \
       -state disabled \
       -relief flat
  } {
      keep -font -background -foreground
  }

  pack $itk_component(intButton) -side right
       	
  eval itk_initialize $args
}

## ----------------------------------------------------------------------------
## execute the command.
## This method sets up blinking and then spawns an \c after _compute().
## \param text the message to be displayed in the dialog
## \param args the script to be computed.
## \returns the return value of the evaluated script
## ----------------------------------------------------------------------------
body CdgBusy::compute {text args} {
  global _errorFlag cdgCtrlCAllowed

  if {$isActive > 0} {
    cmd::Puts "ERROR: cannot execute \n$text"
    return 
  } 

  set _errorFlag 0
  set cdgCtrlCAllowed 1
  .cdgmain help showstr $text
  set computeText $text
  bind $itk_interior <Enter> [code $this enterAction %W %X %Y]
  bind $itk_interior <Leave> [code $this leaveAction]

  set isActive 1
  set oldCursor [. cget -cursor]
  $itk_component(intButton) configure -state normal
  . configure -cursor watch
  update idletask

  set blinkJob [after idle [code $this blink]]
  update idletask
  set result [uplevel $args]
  reset
  return $result
}

## ----------------------------------------------------------------------------
## interrupt the current computation.
## ----------------------------------------------------------------------------
body CdgBusy::interrupt {args} {
  global _errorFlag cdgCtrlCTrapped cdgCtrlCAllowed

  if {$isActive <= 0} {
    return
  }
   
  if {$cdgCtrlCAllowed} {
    ::cmd::Puts "WARNING: Interrupt received, wait a moment ..."
    set cdgCtrlCTrapped 1
    incr isActive -1
    if {$isActive <= 0 || $_errorFlag} {
       reset
    } 
  } else {
    ::cmd::Puts "WARNING: Interrupt suppressed, try later."
  }
}

## ----------------------------------------------------------------------------
## indicate computation
## ----------------------------------------------------------------------------
body CdgBusy::blink {} {
  if {$ledState} {
    $itk_component(intButton) configure -image led_off
    set ledState 0
  } else {
    $itk_component(intButton) configure -image led_on
    set ledState 1
  }

  set blinkJob [after 1000 [code $this blink]]
  update idletask
}

## ----------------------------------------------------------------------------
## reset the blinker to a normal state.
## ----------------------------------------------------------------------------
body CdgBusy::reset {} {
  set isActive 0
  . configure -cursor $oldCursor
  .cdgmain help showstr ""
  if {$blinkJob != ""} {
    after cancel $blinkJob
  }
  $itk_component(intButton) configure -state disabled -image led_off
  set ledState 0
  set blinkJob ""
  set oldCursor ""
  set computeText ""
  bind $itk_interior <Enter> ""
  .cdgmain balloon off
}

## ----------------------------------------------------------------------------
## display the busy balloon.
## This slot is connected to the <Enter> event.
## ----------------------------------------------------------------------------
body CdgBusy::enterAction {w x y} {
  .cdgmain balloon on $x $y $computeText
}

## ----------------------------------------------------------------------------
## disable the busy balloon.
## This slot is connected to the <Leave> event.
## ----------------------------------------------------------------------------
body CdgBusy::leaveAction {} {
  .cdgmain balloon off
}
