# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## Parse - OO wrapper for the parse C structure.
##
## This is an object oriented wrapper arround the C structure with
## the same name. All access to that C structure should be encapsulated
## by this class. All methods dealing directly with a C ParseStruct
## are called via this class. 
## \todo There are rather some exceptions ;) (but why).
##
## \author Michael Daum, Kilian A. Foth, Dietmar Fnning 
## (see also AUTHORS and THANKS for more)
##
## $Id parse.tcl$
## ----------------------------------------------------------------------------
class Parse {

  # private variables

  ## pointer to the C structure
  private variable _cpointer "NULL"; ## \type Parse

  ## the height of a word in a dependency tree on a level.
  ## This hash mapps \c levelId x \c wordIndex to the word height.
  ## Note that this simply caches the values computed by parseWordHeight()
  ## and stores them at the tcl layer for faster access.
  ## \see getWordHeight(), shiftEdge()
  private variable _wordHeight; ## \type TclArray

  ## the depth of a word in a dependency tree on a level.
  ## This hash mapps \c levelId x \c wordIndex to the word depth.
  ## Note that this simply caches the values computed by parseWordDepth()
  ## and stores the mat the tcl layer for faster access.
  ## \see getWordDepth(), shiftEdge()
  private variable _wordDepth; ## \type TclArray

  ## the height of each tree on each level.
  ## This hash mapps the \c levelId to the height. Note that this
  ## simply caches the values computed by parseHeight().
  ## \see getHeight(), shiftEdge()
  private variable _treeHeight ; ## \type TclArray

  ## information about each word on each level.
  ## This has mapps the \c levelId x \c wordIndex to a boolean flag
  ## indicating whether the adressed word is being modified or not.
  private variable _isLeafNode ; ## \type TclArray

  ## cache to speed up tree access.
  ## This hash stores all bindings for each level mapping \c levelId to
  ## the list of all bindings on that level.
  ## A binding is a list consisting of the following elements
  ## - the \c wordIndex
  ## - the \c modifiee index taken from the ParseStruct::verticesStructure
  ## - the \c label taken from the ParseStruct::verticesLabels
  ## - the \c index computed for this parse (see indexOf())
  ## - the \c levelId to which this encoded binding belongs to
  ## \see getBindings(), swapLabel(), swapWord(), shiftEdge()
  private variable _bindings; ## \type TclArray

  ## flag indicating if this parse has nonspec values
  private variable _isNonSpec ""; ## \type Boolean

  # public methods
  constructor {args} {}; ## \type TclList
  public method getBindings {levelId}; ## \type TclString
  public method getBindingById {bindingId}; ## \type TclString
  public method getBindingAt {levelId wordIndex}; ## \type TclString, TclNumber
  public method getHeight {levelId}; ## \type TclString
  public method getWordHeight {levelId wordInedx}; ## \type TclString, TclNumber
  public method getWordDepth {levelId wordInedx}; ## \type TclString, TclNumber
  public method getWidth {}
  public method getModifiers {levelId wordIndex}; ## \type TclString, TclNumber
  public method getModifiee {levelId wordIndex}; ## \type TclString, TclNumber
  public method getId {}
  public method getLatticeId {}
  public method getUserName {}
  public method getDate {}
  public method getGrammarFiles {}
  public method getSearchStrategy {}
  public method getComment {}
  public method getSolutionNo {}
  public method getNoSolutions {}
  public method getLevels {}
  public method getLabels {}
  public method getLexemes {}
  public method getCurrentReading {wordIndex}; ## \type TclNumber
  public method getLexiconItem {wordIndex}; ## \type TclNumber
  public method getPredictions {wordIndex}; ## \type TclNumber
  public method getValue {wordIndex}; ## \type TclNumber
  public method getBadness {}
  public method getScore {}
  public method getViolations {}
  public method getWords {}
  public method getWord {wordIndex}; ## \type TclNumber
  public method retry {history}; ## \type TclList
  public method verify {}
  public method swapWord {wordIndex description eval}; ## \type TclNumber, TclString, TclNumber
  public method optimizeLabel {levelId wordIndex}; ## \type TclString, TclNumber
  public method optimizeWord {wordIndex}; ## \type TclNumber
  public method optimizeStructure {levelId wordIndex}; ## \type TclNumber, TclNumber
  public method swapLabel {levelId wordIndex label eval}; ## \type TclString, TclNumber, TclString, TclNumber
  public method shiftEdge {levelId wordIndex targetIndex eval}; ## \type TclString, TclNumber, TclNumber, TclNumber
  public method isAbstract {}
  public method isNonSpec {}
  public method isLeafNode {levelId wordIndex}; ## \type TclString, TclNumber
  public method register {}
  public method mirror {eval}; ## \type TclNumber
  public method toAnno {}
  public method init {cpointer}; ## \type Parse
  public method indexOf {levelId wordIndex}; ## \type TclString, TclNumber
  public method evalParse {}
  public method markOneEdgePerCycle {level}; ## \type TclString
  public method markEdge {level wordNo, mark}; ## \type TclNumber, TclBool
  public method isMarked {level wordNo}; ## \type TclNumber
  public method clearAllMarks {level};
  public method makeMarkedEdgesRootEdges {level eval};

  # private methods
  private method _invalidateCaches {levelId pattern};
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body Parse::constructor {args} {

  set _cpointer "NULL"
  eval configure $args
}

## ----------------------------------------------------------------------------
## attach a C structure to this instance.
##
## @param cpointer a pointer to a @ref ParseStruct
## Every Parse can only be initialized once.
## ----------------------------------------------------------------------------
body Parse::init {cpointer} {
  if {$_cpointer != "NULL"} {
    error "ERROR: this parse is already attached to a C structure"
  }

  set _cpointer $cpointer
}

## ----------------------------------------------------------------------------
## get all bindings on a given level.
## This method retrieves all bindings of the unerlying Parse. 
## \see getBindingAt
## \param levelId the id of the level whose bindings we are interested in.
## \returns a TclList of all bindings (that is a list of lists)
## \see Parse::_bindings for an explanation of the list format
## ----------------------------------------------------------------------------
body Parse::getBindings {levelId} {

  if {![info exists _bindings($levelId)]} {
    set _bindings($levelId) ""
    set noWords [getWidth]
    for {set i 0} {$i < $noWords} {incr i} {
      lappend _bindings($levelId) [getBindingAt $levelId $i]
    }
  }

  return $_bindings($levelId)
}

## ----------------------------------------------------------------------------
## get the binding at a position on a level.
## This is the lowest access function to the ParseStruct. All information
## retrieved from the ParseStruct is cached in Parse::_bindings for faster
## access. This cache 
## only gets invalidated by swapWord(), swapLabel() or shiftEdge().
## Thereby caching should be \em mostly transparent. 
## \note The cache gets out of sync if the C layer changes the Parse structure
## without using this Tcl layer.
## \param levelId the id of the level whose binding we are interested in.
## \param wordIndex the index of the word of the parse.
## \returns a TclList of the binding information
## \see Parse::_bindings for an explanation of the list format
## ----------------------------------------------------------------------------
body Parse::getBindingAt {levelId wordIndex} {
  
  if {$_cpointer == "NULL"} {
    return ""
  }

  set index [indexOf $levelId $wordIndex]
  set modifiee [intVectorElementAt [ParseStruct_verticesStructure_get $_cpointer] $index]
  set labelPointer [vectorElement [ParseStruct_verticesLabels_get $_cpointer] $index]
  set label [pointer2string $labelPointer]
  return [list $wordIndex $modifiee $label $index $levelId ]
}

## ----------------------------------------------------------------------------
## get the bindings of a given index.
## This comes done to the binding that is pointed to by a constraint violation.
## Its \c nodeBindingIndex1 and \c nodeBindingIndex2 correspond to the binding
## index, that is the third element in the bindings list (see Parse::_bindings).
## \todo This method is inefficient as we search thru all bindings for
## the given id. 
## Alas extra hash shall soothe thee, as we damn the ancillary storage hassle.
## \param id the node binding index in the parse (see indexOf())
## \returns the binding list or an empty list if a binding of index \c id isn't
## there. 
## \see Parse::_bindings for an explanation of the bindings list format.
## ----------------------------------------------------------------------------
body Parse::getBindingById {id} {
  foreach levelId [getLevels] {
    foreach binding [getBindings $levelId] {
      if {[lindex $binding 3] == $id} {
        return $binding
      }
    }
  }
  return ""
}

## ----------------------------------------------------------------------------
## get all modifiers of a word on a level.
## This method does not boil down to getBindingAt() as you would expect but
## sacrifies clean design for speed accessing the ParseStruct::verticesStruct
## directly retrieving what we want.
## \param levelId the id of the level whose binding we are interested in.
## \param wordIndex the index of the word of the parse.
## \returns a list of all modifiers of the given word.
## ----------------------------------------------------------------------------
body Parse::getModifiers {levelId wordIndex} {

  if {$_cpointer == "NULL"} {
    return ""
  }

  set modifiers ""
  set noWords [getWidth]
  set vertices [ParseStruct_verticesStructure_get $_cpointer]

  for {set i 0} {$i < $noWords} {incr i} {
    set index [indexOf $levelId $i]
    set modifieeIndex [intVectorElementAt $vertices $index]
    if {$wordIndex == $modifieeIndex} {
      # marked nodes are not modifiees of anyone
      if { ![isMarked $levelId $i] } {
        lappend modifiers $i
      }
    }
  }
  
  return $modifiers
}

## ----------------------------------------------------------------------------
## get the modifiee of a word on a level.
## This method lets you access the ParseStruct::verticesStruct easily.
## \param levelId the id of the level whose binding we are interested in.
## \param wordIndex the index of the word of the parse.
## \returns the modifiee index of the given word.
## ----------------------------------------------------------------------------
body Parse::getModifiee {levelId wordIndex} {
  if {$_cpointer == "NULL"} {
    return ""
  }

  set index [indexOf $levelId $wordIndex]
  return [intVectorElementAt [ParseStruct_verticesStructure_get $_cpointer] $index]
}

## ----------------------------------------------------------------------------
## get all violations of the current parse.
## Information is taken from the Parse::violations and then taken appart
## meticulously for you. We return a list of violations each 
## with the following format in the following order:
## - the \c name of the violation
## - the \c first node binding index (aka. the binding id - see getBindingById())
## - the \c second node binding index 
## - the \c score of the violation
## \returns the violations as described above, that is a list of lists.
## ----------------------------------------------------------------------------
body Parse::getViolations {} {
  if {$_cpointer == "NULL"} {
    return ""
  }

  set violations [ParseStruct_violations_get $_cpointer]
  set result ""

  for {set l $violations} {$l != "NULL" } {set l [ListStruct_next_get $l]} {
    set viola [ListStruct_item_get $l]
    set constraint [ConstraintViolationStruct_constraint_get $viola]
    set name [ConstraintStruct_id_get $constraint]
    set score [ConstraintViolationStruct_penalty_get $viola]
    set b1 [ConstraintViolationStruct_nodeBindingIndex1_get $viola]
    set b2 [ConstraintViolationStruct_nodeBindingIndex2_get $viola]

	set b3 [ConstraintViolationStruct_nodeBindingIndex3_get $viola]

    # translate LV index to vertex number
    set b1 [parseRmap $_cpointer $b1]
    set b2 [parseRmap $_cpointer $b2]

	set b3 [parseRmap $_cpointer $b3]

    lappend result [list $name $b1 $b2 $score $b3]
  }

  return $result
}

## ----------------------------------------------------------------------------
## Get all levels of the Parse.
## This method collects the ids of all levels contained in the current parse.
## So this is an easy accessor for ParseStruct::levels.
## \returns a list of \c levelIds
## ----------------------------------------------------------------------------
body Parse::getLevels {} {

  if {$_cpointer != "NULL"} {
    set levels ""
    for {set l [ParseStruct_levels_get $_cpointer]} \
	{$l != "NULL"} \
	{set l [ListStruct_next_get $l]} {
      set item [ListStruct_item_get $l]
      set levelId [pointer2string $item]
      lappend levels $levelId
    }

    # If all of the Parse's levels are also declared in the grammar,
    # make sure that the result conforms to the order in the grammar.
    set in_parse [split $levels]
    set in_grammar [getLevelSort]
    set sp [lsort $in_parse]
    set sg [lsort $in_grammar]
    if {"$sp" == "$sg"} {
      set levels $in_grammar
    }
  } else {
    set levels [getLevelSort]
  }
     
  return $levels

}

## ----------------------------------------------------------------------------
## get all labels.
## This method collects all labels from ParseStruct::labels for you. Actually
## this should be only a fallback whenever we cannot ask the grammar for its
## labels.
## \returns $_cpointer->labels as a Tcl list
## ----------------------------------------------------------------------------
body Parse::getLabels {} {

  if {$_cpointer == "NULL"} {
    return ""
  }

  set labels ""
  for {set l [ParseStruct_labels_get $_cpointer]} \
      {$l != "NULL" } \
      {set l [ListStruct_next_get $l]} {
    set item [ListStruct_item_get $l]
    lappend labels [pointer2string $item]
  }

  return $labels
}


## ----------------------------------------------------------------------------
## get all lexeme nodes of current parse.
## This method is only a wrapper for parseGetLexemNodes() converting the
## returned List into a TclList (I know this is lame)
## \returns a list of LexemNode items.
## ----------------------------------------------------------------------------
body Parse::getLexemes {} {

  if {$_cpointer == "NULL"} {
    return ""
  }

  set result ""
  set nodes [parseGetLexemNodes $_cpointer]

  for {set l $nodes} {$l != "NULL"} {set l [ListStruct_next_get $l]} {
    lappend result [ListStruct_item_get $l]
  }
  return $result
}

## ----------------------------------------------------------------------------
## get the score of the current parse.
## This is a getter method for the ParseStruct::score.
## \returns a float.
## ----------------------------------------------------------------------------
body Parse::getScore {} {

  if {$_cpointer == "NULL"} {
    return ""
  }

  return [ParseStruct_score_get $_cpointer]
}

## ----------------------------------------------------------------------------
## get the id of the current parse.
## This is a getter method for the ParseStruct::id
## \returns a TclString
## ----------------------------------------------------------------------------
body Parse::getId {} {

  if {$_cpointer == "NULL"} {
    return ""
  }

  return [ParseStruct_id_get $_cpointer]
}

## ----------------------------------------------------------------------------
## get the date of the current parse.
## This is a wrapper for parseDate() in the C layer.
## \returns the date of creation of the current parse.
## ----------------------------------------------------------------------------
body Parse::getDate {} {

  if {$_cpointer == "NULL"} {
    return ""
  }

  return [parseDate $_cpointer] 
}

## ----------------------------------------------------------------------------
## get the search strategy.
## This method returns the information about the search strategy 
## with which the current parse has been produced
## \returns a TclString
## ----------------------------------------------------------------------------
body Parse::getSearchStrategy {} {
  if {$_cpointer == "NULL"} {
    return ""
  }
  return [ParseStruct_searchStrategy_get $_cpointer] 
}

## ----------------------------------------------------------------------------
## get the comment of the current parse.
## This is a getter method for the ParseStruct::comment.
## \returns a TclString
## ----------------------------------------------------------------------------
body Parse::getComment {} {
  if {$_cpointer == "NULL"} {
    return ""
  }
  return [ParseStruct_comment_get $_cpointer] 
}

## ----------------------------------------------------------------------------
## get the number of solutions.
## This is the number of solutions this parse is one of.
## \see getSolutionNo()
## ----------------------------------------------------------------------------
body Parse::getNoSolutions {} {
  if {$_cpointer == "NULL"} {
    return ""
  }
  return [ParseStruct_noOfSolutions_get $_cpointer] 
}

## ----------------------------------------------------------------------------
## get the solution number.
## This is the solution number identifying this parse as being one of several
## solutions. 
## \see getNoSolutions()
## ----------------------------------------------------------------------------
body Parse::getSolutionNo {} {
  if {$_cpointer == "NULL"} {
    return ""
  }
  return [ParseStruct_solutionNo $_cpointer] 
}

## ----------------------------------------------------------------------------
## get all grammar files.
## This gets all grammar files related to this parse, that is return the 
## tcl-ized list of ParseStruct::grammarFiles.
## \returns a TclList of grammar file names.
## ----------------------------------------------------------------------------
body Parse::getGrammarFiles {} {
  set grammarFiles ""

  if {$_cpointer == "NULL"} {
    return ""
  }

  for {set l [ParseStruct_grammarFiles_get $_cpointer]} \
      {$l != "NULL"} {set l [ListStruct_next_get $l]} {
    set item [ListStruct_item_get $l]
    lappend grammarFiles [pointer2string $item]
  }

  return $grammarFiles
}

## ----------------------------------------------------------------------------
## get the lattice id.
## This method returns the \c latticeId this parse is annotating, that is
## return ParseStruct::latticeId.
## ----------------------------------------------------------------------------
body Parse::getLatticeId {} {
  if {$_cpointer == "NULL"} {
    return ""
  }

  return [ParseStruct_latticeId_get $_cpointer] 
}

## ----------------------------------------------------------------------------
## get the user name.
## This method returns the user who created this parse.
## ----------------------------------------------------------------------------
body Parse::getUserName {} {
  if {$_cpointer == "NULL"} {
    return ""
  }

  return [ParseStruct_userName_get $_cpointer] 
}

## ----------------------------------------------------------------------------
## Return badness of the current parse in the order
## - the number of hard violations
## - the number of soft violations
## - the remaining score not taking the hard violations into account.
##
## Note that this is NOT what the BadnessStruct counts; there, hard
## conflicts contribute both to ->no and to ->hard.
## ----------------------------------------------------------------------------
body Parse::getBadness {} {

  if {$_cpointer == "NULL"} {
    return "n.a."
  }

  if {[parseIsAbstract $_cpointer]} {
    return "n.a.";
  } 

  set b [ParseStruct_badness_get $_cpointer]
  set no [BadnessStruct_no_get $b]
  set hard [BadnessStruct_hard_get $b]
  set soft [BadnessStruct_soft_get $b]
  set nosoft [expr $no - $hard]
  return [list $hard $no $soft]
 
}

## ----------------------------------------------------------------------------
## get the height of the dependency tree on a given level.
## This method simply returns the parseHeight() of the current parse on
## the given level. Doing that it caches the result on the tcl layer for faster
## access. If the parse changes in structure (by shiftEdge()) this array
## gets invalidated again.
## \param levelId the id of the level we are interested in.
## \returns the tree height
## ----------------------------------------------------------------------------
body Parse::getHeight {levelId} {
  if {$_cpointer == "NULL"} {
    return 0
  }

# the caching functionality was disabled due to problems with the Graph marking algorithm used in
# ParseTree::drawLevel. When it is absolutly neccessary to re-enable it, a cache-invalidation
# procedure must be added.

#  if {![info exists  _treeHeight($levelId)]} {
#    set _treeHeight($levelId) [parseHeight $_cpointer $levelId]
#  }

#  return $_treeHeight($levelId)
    
  return [parseHeight $_cpointer $levelId]
}

## ----------------------------------------------------------------------------
## Wrapper for parseWordHeight().
## This method simply returns the parseWordHeight() of a word on a level in
## the current parse. Afterwards results are cached at the tcl layer. Occasionally
## shiftEdge() nullifies this again.
## \param levelId the level we are interested in
## \param wordIndex the word we are interested in
## \returns the word height
## ----------------------------------------------------------------------------
body Parse::getWordHeight {levelId wordIndex} {
  if {$_cpointer == "NULL"} {
    return 0
  }
  
  if {![info exists _wordHeight($levelId,$wordIndex)]} {
    set _wordHeight($levelId,$wordIndex) \
      [parseWordHeight $_cpointer $levelId $wordIndex]
  }

  return $_wordHeight($levelId,$wordIndex)
}

## ----------------------------------------------------------------------------
## get maximum distance of reachable leaves.
## This method works using the same caching logic like getWordHeight() and
## all the other methods that retrieve numbers from the C layer caching them
## on the tcl layer.
## \param levelId the level we are interested in
## \param wordIndex the word we are interested in
## \returns the word depth
## ----------------------------------------------------------------------------
body Parse::getWordDepth {levelId wordIndex} {
  if {$_cpointer == "NULL"} {
    return 0
  }

  if {![info exists _wordDepth($levelId,$wordIndex)]} {
    set level [findLevel $levelId]
    if {$level == "NULL"} {
      error "level $levelId unknown"
    }
    set _wordDepth($levelId,$wordIndex) \
      [parseWordDepth $_cpointer $level $wordIndex]
  }

  return $_wordDepth($levelId,$wordIndex)
}


## ----------------------------------------------------------------------------
## get the number of words.
## Actually being a bad name for what it does this simply counts the
## number of words in the ParseStruct::words vector.
## ----------------------------------------------------------------------------
body Parse::getWidth {} {
  if {$_cpointer == "NULL"} {
    return 0
  }
  return [vectorSize [ParseStruct_words_get $_cpointer]]
}

## ----------------------------------------------------------------------------
## get the words Vector of the current parse.
## This method does not attempt to extract all words from the ParseStruct::words
## vector but returns the Vector itself. 
## \returns a Vector of Word items
## ----------------------------------------------------------------------------
body Parse::getWords {} {
  if {$_cpointer == "NULL"} {
    return 0
  }
  return [ParseStruct_words_get $_cpointer]
}

## ----------------------------------------------------------------------------
## get a word from the parse.
## This method gets you the WordStruct at the given index
## \param wordIndex the word we want
## \returns a WordStruct
## ----------------------------------------------------------------------------
body Parse::getWord {wordIndex} {
  if {$_cpointer == "NULL"} {
    return ""
  }
  return [vectorElement [getWords] $wordIndex]
}

## ----------------------------------------------------------------------------
## Return the description of the word currently selected at the
## timepoint $wordIndex.
## ----------------------------------------------------------------------------
body Parse::getCurrentReading {wordIndex} {
  if {$_cpointer == "NULL"} {
    return ""
  }
  return [WordStruct_description_get [getWord $wordIndex]]
}

## ----------------------------------------------------------------------------
## Return a string representing the lexicon item currently selected at the
## timepoint $wordIndex.
## ----------------------------------------------------------------------------
body Parse::getValue {i} {
  if {$_cpointer == "NULL"} {
    return ""
  }
  return [parseFormatValue $_cpointer $i]
}

## ----------------------------------------------------------------------------
## Return a string representing the predictions made about
## timepoint $wordIndex.
## ----------------------------------------------------------------------------
body Parse::getPredictions {i} {

  if {$_cpointer == "NULL"} {
    return ""
  }

  set gn [parseGraphemeNode $_cpointer $i]
  if {$gn == "NULL"} return ""

  return [gnFormatPredictions $gn]
}

## ----------------------------------------------------------------------------
## Return the lexicon item currently selected at the timepoint $wordIndex.
## ----------------------------------------------------------------------------
body Parse::getLexiconItem {i} {
  if {$_cpointer == "NULL"} {
    return ""
  }
  return [parseLexiconItem $_cpointer $i]
}

## ----------------------------------------------------------------------------
## verify the current parse.
## This is method verifies the current parse using the first annotation
## linked to the same lattice, that is: this parse has a link to a lattice
## and there exist one or more annotations linked to this lattice; we take the
## first of them.
## \returns a ParseVerificationStruct or the empty string on an error
## ----------------------------------------------------------------------------
body Parse::verify {} {
  if {$_cpointer == "NULL"} {
    return ""
  }

  set parseId [getId]
  set latId [getLatticeId]
  set lat [findLattice $latId]
  if {"$lat" != "NULL"} {
    for {set l [findAnnotations $lat 1]} \
        {$l != "NULL"} \
        {set l [ListStruct_next_get $l]} {
	  set ae [ListStruct_item_get $l]
          set annoId [AnnoEntryStruct_id_get $ae]
          # if there is an annotation that has the same name as the parse,
          # do not use it for verification, that would be pointless.
          if {"$annoId" != "$parseId"} {
            return [parseVerify $_cpointer $ae]
          }
        }
  }
  return ""
}

## ----------------------------------------------------------------------------
## exchange a word at a given index.
## This method uses parseSwapWord() to maniplulate the parse structure.
## Note that the Parse::_bindings array is invalidated completely here.
## ----------------------------------------------------------------------------
body Parse::swapWord {wordIndex description eval} {
  array unset _bindings
  if {$_cpointer == "NULL"} {
    return ""
  }
  parseSwapWord $_cpointer $wordIndex $description $eval
}

## ----------------------------------------------------------------------------
## exchange the label of a dependency edge.
## This method uses parseSwapLabel() to do the actual work. Note that the
## Parse::_bindings array is invalidated for this level.
## \param levelId the level we want to maniplulate
## \param wordIndex the word index of the dependency edge
## \param label the new label for this dependency edge
## ----------------------------------------------------------------------------
body Parse::swapLabel {levelId wordIndex label eval} {
  catch { unset _bindings($levelId) }
  if {$_cpointer == "NULL"} {
    return ""
  }
  parseSwapLabel $_cpointer [indexOf $levelId $wordIndex] $label $eval
}

## ----------------------------------------------------------------------
## Wrapper for parseOptimizeLabel().
##
body Parse::optimizeLabel {levelId wordIndex} {
  if {$_cpointer == "NULL"} {
    return ""
  }
  return [parseOptimizeLabel $_cpointer $levelId $wordIndex]
}

## ----------------------------------------------------------------------
## Wrapper for parseOptimizeWord().
##
body Parse::optimizeWord {wordIndex} {
  if {$_cpointer == "NULL"} {
    return ""
  }
  return [parseOptimizeWord $_cpointer $wordIndex]
}

## ----------------------------------------------------------------------
## Wrapper for parseOptimizeStructure().
##
body Parse::optimizeStructure {levelId wordIndex} {
  if {$_cpointer == "NULL"} {
    return ""
  }
  return [parseOptimizeStructure $_cpointer $levelId $wordIndex]
}


## ----------------------------------------------------------------------
## invalidate all cached Data
##
body Parse::_invalidateCaches {levelId pattern} {

  array unset _wordHeight "$levelId,$pattern"
  array unset _wordDepth "$levelId,$pattern"
  array unset _isLeafNode "$levelId,$pattern"
  catch {
    unset _bindings($levelId)
    unset _treeHeight($levelId)
  }
  
}

## ----------------------------------------------------------------------------
## redirect a dependency edge to a new modifiee.
## This method uses parseShiftEdge() to do the actual work. Note that the
## arrays Parse::_wordHeight, Parse::_wordDepth, Parse::_bindings and 
## Parse::_treeHeight are invalidated for this level.
## \param levelId the level we want to maniplulate
## \param wordIndex the word index of the dependency edge
## \param targetIndex the index of the new modifiee
## ----------------------------------------------------------------------------
body Parse::shiftEdge {levelId wordIndex targetIndex eval} {
  
  _invalidateCaches $levelId "*"
  
  if {$_cpointer == "NULL"} {
    return ""
  }
  parseShiftEdge $_cpointer [indexOf $levelId $wordIndex] $targetIndex $eval
}

## ----------------------------------------------------------------------------
## Wrapper for parseIndex().
## ----------------------------------------------------------------------------
body Parse::indexOf {levelId wordIndex} {
  if {$_cpointer == "NULL"} {
    return 0
  }
  return [parseIndex $_cpointer $levelId $wordIndex]
}

## ----------------------------------------------------------------------------
## get the abstraction state of the current parse.
## This is a straight delegation to parseIsAbstract().
## \returns a bool.
## ----------------------------------------------------------------------------
body Parse::isAbstract {} {
  if {$_cpointer == "NULL"} {
    return "FALSE"
  }
  return [parseIsAbstract $_cpointer]
}

## ----------------------------------------------------------------------------
## returns true if this parse has nonspec values
## ----------------------------------------------------------------------------
body Parse::isNonSpec {} {
  if {$_isNonSpec != ""} {
    return $_isNonSpec
  }

  foreach levelId [getLevels] {
    foreach binding [getBindings $levelId] {
      if {[lindex $binding 1] == -2} {
	set _isNonSpec "TRUE"
	return $_isNonSpec
      }
    }
  }

  set _isNonSpec "FALSE"
  return $_isNonSpec
}

## ----------------------------------------------------------------------------
## test whether a word on a given level is a leaf.
## This is done by examining the ParseStruct::verticesStruct storing the
## result into Parse::_isLeafNode at the index \c levelId x \c wordIndex.
## This caches information of the C layer in the tcl layer. Invalidating this
## cache is done when needed, i.e. in shiftEdge().
## \param levelId the level we are interested in
## \param wordIndex the word index 
## \returns 0 or 1 
## ----------------------------------------------------------------------------
body Parse::isLeafNode {levelId wordIndex} {
  if {$_cpointer == "NULL"} {
    return 0
  }

# the caching functionality was disabled due to problems with the Graph marking algorithm used in
# ParseTree::drawLevel. When it is absolutly neccessary to re-enable it, a cache-invalidation
# procedure must be added.
  
#  if {![info exists _isLeafNode($levelId,$wordIndex)]} {
#    set _isLeafNode($levelId,$wordIndex) 1
    set isLeafNode 1
    set noWords [getWidth]
    set vertices [ParseStruct_verticesStructure_get $_cpointer]
    for {set i 0} {$i < $noWords} {incr i} {
      set index [indexOf $levelId $i]
      set modifieeIndex [intVectorElementAt $vertices $index]
      
      if { [isMarked $levelId $i] } {
        set modifieeIndex -1
      } else {
      }
      
      if {$wordIndex == $modifieeIndex} {
#	set _isLeafNode($levelId,$wordIndex) 0
	set isLeafNode 0
	break
      }
    }
#  }

#  return $_isLeafNode($levelId,$wordIndex)
  return $isLeafNode
}

## ----------------------------------------------------------------------------
## register the current parse to th C layer.
## This puts the C Parse into inputCurrentGrammar->parses. Actually this
## operation is blindly delegated to parseRegister().
## ----------------------------------------------------------------------------
body Parse::register {} {
  if {$_cpointer == "NULL"} {
    return 0
  }
  parseRegister $_cpointer
}

## ----------------------------------------------------------------------------
## compute the mirror edges.
## This is a straight delegation to parseMirror(). Note that all caches on
## the tcl layer are possibly invalidated now.
## ----------------------------------------------------------------------------
body Parse::mirror {eval} {
  array unset _wordHeight 
  array unset _wordDepth
  array unset _isLeafNode
  array unset _bindings
  array unset _treeHeight
  if {$_cpointer == "NULL"} {
    return 
  }
  parseMirror $_cpointer $eval
}

## ----------------------------------------------------------------------------
## convert the current parse to an annotation.
## This is a straight delegation to parse2annotation().
## ----------------------------------------------------------------------------
body Parse::toAnno {} {
  if {$_cpointer == "NULL"} {
    return ""
  }
  return [parse2annotation $_cpointer]
}



## ----------------------------------------------------------------------------
## Try to parse the sentence again, but respecting the user's previous
## edits.
## ----------------------------------------------------------------------------
body Parse::retry {history} {
  
  # find fitting constraint net
  set lat [getLatticeId]
  set net [cnFindNetForLattice $lat]
  if {"NULL" == $net} {
    commandEval "newnet $lat"
    set net [cnFindNetForLattice $lat]

    # the user might have pressed C-c...
    if {"NULL" == $net} {
      ::cmd::Puts "ERROR: failed to build constraint net"
      return
    }
  }

  # put the net in the desired state
  set netname [ConstraintNetStruct_id_get $net]
  commandEval "renewnet $netname"

  # Laboriously reverse-engineer the command history list. We have to
  # collect the items first and lock them only after we have seen them
  # all; otherwise, if the user changes a label twice, we would make
  # two incompatible demands and kill the problem.
  foreach x $history {
    set command [lindex $x 3]
    set opcode [lindex $command 1]
    if {"undoLabel" eq $opcode} {
      set index [lindex $command 2]
      set label [lindex $command 3]
      set level [lindex $command 4]
      set labels("$index:$level") $label
    } elseif {"undoLexeme" eq $opcode} {
      set index [lindex $command 2]
      set desc [lindex $command 3]
      set words($index) $desc
    } elseif {"undoEdge" eq $opcode} {
      set index [lindex $command 2]
      set regent [lindex $command 3]
      set level [lindex $command 4]
      set edges("$index:$level") $regent
    }
  }

  # Lock the current values of all changed things
  foreach {key label} [array get labels] {
    regexp {^([0-9]+):(.+)$} $key dummy index level
    cnLockLabel $net $label $index $level
  }

  foreach {key desc} [array get words] {
    cnLockWord $net $desc $index
  }

  foreach {key regent} [array get edges] {
    regexp {^([0-9]+):(.+)$} $key dummy index level
    cnLockEdge $net $regent $index $level
  }

  # start another computation whose results go into the same viewer
  set method [ParseStruct_searchStrategy_get $_cpointer]
  if {"frobbing" == $method || "anno2parse" == $method} {
    set parsename [getId]
    commandEval "frobbing net=$netname parse=$parsename method=combined execute=q"
  } elseif {"gls" == $method} {
    commandEval "gls $netname"
  } elseif {"netsearch" == $method} {
    commandEval "netsearch $netname"
  } else {
    ::cmd::Puts "ERROR: can't repeat search strategy `$method'"
  }
}

## ----------------------------------------------------------------------------
## Call parseEval().
## ----------------------------------------------------------------------------
body Parse::evalParse {} {
  parseEval $_cpointer
}

## ----------------------------------------------------------------------------
## Call parseMarkOneEdgePerCycle
## ----------------------------------------------------------------------------
body Parse::markOneEdgePerCycle {level} {
  
  _invalidateCaches $level "*"
  
  if {$_cpointer == "NULL"} {
    return ""
  }

  return [parseMarkOneEdgePerCycle $_cpointer $level];
}

## ----------------------------------------------------------------------------
## Call parseMarkEdge
## ----------------------------------------------------------------------------
body Parse::markEdge {level wordNo, mark} {
  
  _invalidateCaches "$level" "*"
  
  if {$_cpointer == "NULL"} {
    return ""
  }

  return [parseMarkEdge $_cpointer $level $wordNo $mark];
}

## ----------------------------------------------------------------------------
## Call parseIsMarked
## ----------------------------------------------------------------------------
body Parse::isMarked {level wordNo} {
  if {$_cpointer == "NULL"} {
    return ""
  }

  return [parseIsMarked $_cpointer $level $wordNo];
}

## ----------------------------------------------------------------------------
## Call parseClearAllMarks
## ----------------------------------------------------------------------------
body Parse::clearAllMarks {level} {
  
  _invalidateCaches "$level" "*"
  
  if {$_cpointer == "NULL"} {
    return ""
  }

  return [parseClearAllMarks $_cpointer $level];
}

## ----------------------------------------------------------------------------
## Call parseMakeMarkedEdgesRootEdges
## ----------------------------------------------------------------------------
body Parse::makeMarkedEdgesRootEdges {level eval} {
  
  _invalidateCaches "$level" "*"
  
  if {$_cpointer == "NULL"} {
    return ""
  }

  return [parseMakeMarkedEdgesRootEdges $_cpointer $level $eval];
}

