# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaConfigItem - interface for configurable objects.
## This class has no implementation and only serves as a marker of derived
## configurable items in YadaConfigDocument.
## \ingroup YadaConfiguration
##
## \author Michael Daum
##
## ----------------------------------------------------------------------------
class YadaConfigItem {
  public variable name ""
  public variable isModified 1; ## \type TclBoolean

  protected variable _domDocument ""

  public method load {{fileName ""}}; ## \type TclString
  public method save {}
  public method delete {}
  public method toDom {}
  public method fromDom {document}; ## \type domDoc
  public method fromDomNode {node}; ## \type domNode
  public method hasSavedState {}
  public method getWindowTitle {}
  public method getFileName {}; ## \purevirtual
  public method getZippedFileName {}
  public method evalXpath {xpath}; ## \type TclString
  public method clone {args}; ## \purevirtual \type TclList

  constructor {args} {}; ## \type TclList
  public method destructor {}
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaConfigItem::constructor {args} {
  eval configure $args
  if {$name == "<none>"} {
    set isModified 0
  }
}

## ----------------------------------------------------------------------------
## destructor
body YadaConfigItem::destructor {} {
  if {$_domDocument != ""} {
    $_domDocument delete
    set _domDocument ""
  }
}

## ----------------------------------------------------------------------------
## ----------------------------------------------------------------------------
## Serializes object state to unique file (depends on object name)
## ----------------------------------------------------------------------------
body YadaConfigItem::save {} {

  if {$_domDocument != ""} {
    $_domDocument delete
  }
  set _domDocument [toDom]	
  set fileName [getFileName]

  if { [::file exists ${fileName}.gz] } {
    ::file delete ${fileName}.gz
  }

  set fileID [open $fileName {CREAT WRONLY}]
  set rootNode [$_domDocument documentElement]
  puts $fileID [$rootNode asXML]
  close $fileID

  if {[catch {exec gzip -f $fileName} errMsg]} {
      error "Saving failed: $errMsg"
  }
  set isModified 0
}

## ----------------------------------------------------------------------------
## Loads object state from unique file (depends on object name)
## @param fileName optionally specify the file to load from
## ----------------------------------------------------------------------------
body YadaConfigItem::load {{fileName ""}} {

  if {$fileName == ""} {
    set fileName [getZippedFileName]
    if { ![::file exists $fileName]} {   
      error "File $fileName not found."
    }
  }

  set fileID [open "|zcat $fileName" {RDONLY}]
  set xmlData [read  $fileID]
  close $fileID

  if {$_domDocument != ""} {
    $_domDocument delete
  }
  set _domDocument [dom parse $xmlData]
  fromDom $_domDocument

  set isModified 0
}

## ----------------------------------------------------------------------------
## Predicate returning 1, if file containing persistent state of YadaGrammar
## exists, otherwise 0
## ----------------------------------------------------------------------------
body YadaConfigItem::hasSavedState {} {
  set fileName [getZippedFileName] 
  if { [::file exists $fileName] } {
      return 1
  } else {
      return 0
  }
}

## ----------------------------------------------------------------------------
## get the item title reflecting modification state of document
## ----------------------------------------------------------------------------
body YadaConfigItem::getWindowTitle {} {
  if {$isModified && $name != "<none>"} {
    return "$name *** changed *** " 
  } else {
    return "$name"
  }
}

## ----------------------------------------------------------------------------
## get filename of compressed data
## ----------------------------------------------------------------------------
body YadaConfigItem::getZippedFileName {} {
  return "[getFileName].gz"
}

## ----------------------------------------------------------------------------
## evaluate an xpath expression in the related xml document
## @param xpath a vaild xpath expression for a yada document
## @returns the result value on success and the empty string on failure
## ----------------------------------------------------------------------------
body YadaConfigItem::evalXpath {xpath} {

  if {$_domDocument == ""} {
    set fileName [getZippedFileName]
    if {![file exists $fileName]} {  
      return ""
    }

    set fileID [open "|zcat $fileName" {RDONLY}]
    set xmlData [read  $fileID]
    close $fileID
    set _domDocument [dom parse $xmlData]
  }

  set rootNode [$_domDocument documentElement]
  return [$rootNode selectNodes $xpath]
}

## ----------------------------------------------------------------------------
## delete the persistence file
## ----------------------------------------------------------------------------
body YadaConfigItem::delete {} {
  set fileName [getZippedFileName]
  if {![file exists $fileName]} {
    return
  }
  file delete -- $fileName
}
