# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaLabelStats - load sets of results.
## \ingroup YadaDocuments
##
## \author Michael Daum
##
## ----------------------------------------------------------------------------
class YadaLabelStats {
  inherit YadaDocument

  # variables ----------------------------------------------------------------
  private variable _isRunning 0
  private variable _gotBreak 0
  private variable _followDataFlag 1
  private variable maxWg 10000

  # indicates whether the statistics for the annotation are computed
  private variable _grammarIsInitialized 0

  # the experiments of the selected grammar
  private variable allWeights ""

  # the labels for which is searched in the experiments
  private variable allLabels {}

  # the labels displayed in the table
  private variable displayLabels {}

  # count of labels
  private variable totalNrLabels

  # count of annos
  private variable totalTries 0

  # lists of labels in an annotation
  private variable labelsInAnno

  # counts of labelling errors per label and weight
  private variable labelErrorCount

  # count errors per weight
  private variable errorCount

  # count labelling hits per weight
  private variable totalLabelledHits

  # count rejected root attachments per label
  private variable rejectedLabelCount

  # count wrong roots per weight
  private variable rejectionCount

  # recall per label
  private variable recallLabels

  # total recall per weight
  private variable totalRecall

  # rejections per label
  private variable rejectionLabels


  # methods ------------------------------------------------------------------
  public method init {}
  public method reload {}
  public method clear {}
  public method run {{experiment ""}}
  public method activationHandle {}
  public method toDom {};
  public method fromDom {document}; ## \type domDoc
  public method fromDomNode {labelstatsNode}; ## \type domNode
  public method getWindowTitle {}
  public method getFileName {}
  public method isModified {}
  public method toggleFollowData {}
  public method isFollowData {}

  constructor {args} {}; ## \type TclList
  destructor {}

  private method _breakButtonAction {}
  private method _leaveTableCommand {}
  private method _escapeCommand {}
  private method _returnCommand {}

  private method _selectGrammar {}
  private method _initializeGrammar {}

  private method _showLabels {{labels ""}}
  private method _insertData {experiment}
  private method _clearRow {}
  private method _rowTagCommand {row}; ## \type TclNumber

  private method _getCurrentGrammar {}
  private method _getSelectedExperiments {}
  private method _getExperimentsToCompute {}
  private method _getExperimentsOfGrammar {}
  private method _getUndoneExperiments {}
  private method _getAnnotationDir {}
  private method _getRowOfExperimentName {experimentName}; ## \type TclString

  private method _initAnnos {{anno ""}}
  private method _initData {{weight ""}}
  private method _readExperiment {weight}
  private method _readFile {fileName weight}
  private method _readAnno {fileName} 
  private method _parseAnno {document}
  private method _evaluateExperiment {weight}

};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaLabelStats::constructor {args} {
  
  # the labels for which is searched in the experiments
  set allLabels [lsort {
    ADV APP ATTR AUX AVZ CJ DET ETH EXPL GMOD GRAD KOM KON KONJ
    MOD NEB NP2 OBJA OBJA2 OBJD OBJG OBJC OBJI OBJP PAR 
    PART PN PP PRED REL S SUBJ SUBJC VOK ZEIT ""}]



  # toolbar
  $itk_component(toolbar) add FlatButton reloadButton \
    -image [YadaImages::get reload] \
    -balloonstr "reload" \
    -command [code $this reload]

  $itk_component(toolbar) add FlatButton clearButton \
	-image [YadaImages::get erase] \
	-command [code $this _clearRow] \
	-balloonstr "clear"

  $itk_component(toolbar) add FlatButton runButton \
	-image [YadaImages::get run] \
	-command [code $this run] \
	-balloonstr "run"

  # heading space
  itk_component add heading {
    frame $itk_component(childsite).heading
  }

  # the grammar
  itk_component add grammarBoxLabel {
    label $itk_component(heading).grammarBoxLabel \
	-text "Grammar:" \
	-anchor w
  }  {}

  itk_component add grammarBox {
    iwidgets::combobox $itk_component(heading).grammarBox \
	-completion false \
	-borderwidth 2 \
	-margin 4 \
	-selectioncommand [code $this _selectGrammar]
  }  {}

  # the table
  itk_component add pane {
      iwidgets::panedwindow $itk_component(childsite).pane \
          -orient horizontal \
	  -sashindent -8 \
	  -thickness 2 \
	  -sashheight 8 \
	  -sashwidth 18
  } {
      keep -background -cursor
  }
  $itk_component(pane) add top -margin 0
  set top [$itk_component(pane) childsite top]

  itk_component add table {
     MyTable $top.table \
       	-exportselection 1 \
	-selecttitles 1 \
	-rowseparator "\n" \
	-colseparator "\t" \
	-rows 30 \
       	-cols 10 \
       	-state normal \
       	-cellanchor w \
       	-colstretchmode last \
       	-rowstretchmode none \
       	-selectmode extended \
       	-selecttype row \
       	-titlerows 1 \
       	-titlecols 1 \
	-borderwidth 3 \
	-outerborderwidth 2 \
	-outerrelief sunk \
	-selforeground "" \
	-selbackground "" \
	-font {-adobe-helvetica-medium-r-normal--12-*-*-*-*-*-*-*} \
	-hscrollmode dynamic \
	-rowtagcommand [code $this _rowTagCommand]
  } { 
     keep -background -cursor 
  }
  
  # bindings
  set hull $itk_component(hull)
  set tableInTable [$itk_component(table) component table]
  bind $hull <Leave> [code $this _leaveTableCommand]
  bind $hull <FocusOut> [code $this _leaveTableCommand]
  bind $tableInTable <KeyPress-Escape> [code "$this _escapeCommand; break"]
  bind $tableInTable <KeyPress-Return> [code "$this _returnCommand; break"]
  
  # packing + gridding
  pack $itk_component(heading) -fill x -side top -anchor n
  pack $itk_component(pane) -fill both -expand 1 -padx 2 -pady 5 -side top
  pack $itk_component(table) -fill both -expand 1 -anchor n

  grid columnconfigure $itk_component(heading) 1 -weight 1
  grid columnconfigure $itk_component(heading) 3 -weight 1
  grid columnconfigure $itk_component(heading) 5 -weight 1
  grid columnconfigure $itk_component(heading) 6 -weight 1
  
  grid $itk_component(grammarBoxLabel) -row 0 -column 0 -sticky ew -pady 2
  grid $itk_component(grammarBox)      -row 0 -column 1 -sticky ew -pady 2 -columnspan 3

  eval itk_initialize $args
  .main registerDocument $this
}
		

## ----------------------------------------------------------------------------
## Destructor
## ----------------------------------------------------------------------------
body YadaLabelStats::destructor {} {
}

## ----------------------------------------------------------------------------
## init
## ----------------------------------------------------------------------------
body YadaLabelStats::init {} {
  if {$_isInitialized} {
    return
  }
  chain
  clear
}


## ----------------------------------------------------------------------------
## clear the table from all values and set to default dimensions
## ----------------------------------------------------------------------------
body YadaLabelStats::clear {} {
  set table $itk_component(table)
  set cmax [$table cget -cols]
  set rmax [$table cget -rows]

  $table delete rows -holddimensions 0 [$table cget -rows]
  $table configure -rows 30
  $table configure -cols 10
  
  $table tag config default -anchor e
  $table tag config title -font {-adobe-helvetica-medium-r-normal--12-*-*-*-*-*-*-*} -bg #a3a3a3 -fg white
  $table tag config sel -fg "" -bg "" -relief raised -bd 1
  $table tag config active -anchor w -fg black -bg gray -relief sunk -bd 1
  $table tag config colored -bg gray80 -anchor e
  $table tag config runner -fg red3 -bg #a3a3a3
  $table tag raise runner title
  $table width 0 30
  .main printStatus ""
  setModified
}

## ----------------------------------------------------------------------------
## run
## ----------------------------------------------------------------------------
body YadaLabelStats::run {{experiment ""}} {

  set experiments [_getExperimentsToCompute]
  foreach experiment $experiments {
    set args ""
    
    # compute experiment 
    _initData $experiment
    _readExperiment $experiment
    _evaluateExperiment $experiment

    # insert computed data into the table (gui)
    _insertData $experiment
  }

  setModified
  .main printStatus ""
}

## ----------------------------------------------------------------------------
## activationHandle
## ----------------------------------------------------------------------------
body YadaLabelStats::activationHandle {} {
  chain

  # get the previous selection
  set box $itk_component(grammarBox)
  set grammarName [$box get]

  # default grammar to choose
  if {$grammarName == ""} {
    set grammarName "<none>"
  }

  $box clear list 
  eval $box insert list end [.main getGrammarNames]
  $box clear entry
  $box insert entry 0 $grammarName
}

## ----------------------------------------------------------------------------
## _selectGrammar is called when the user selects a grammar.
## It triggers the initialization of the grammar (_initializeGrammar)
## and fills the title row and column of the table
## ----------------------------------------------------------------------------
body YadaLabelStats::_selectGrammar {} {

  clear
  set grammarName [_getCurrentGrammar]
  if {[regexp "<none>" $grammarName]} {
    return
  }

  set table $itk_component(table)
  set box $itk_component(grammarBox)
  if {$grammarName == ""} {
    set grammarName [$box get]
    if {$grammarName == ""} {
      set grammarName "<none>"
    }
  } 
  set experiments [_getExperimentsOfGrammar]
  set rmax [llength $experiments]
  if {$rmax > 30} {
    $table configure -rows [llength $experiments]
  }
  set allWeights $experiments

  $box clear entry
  $box insert entry 0 $grammarName
  set listOfEntries [[$box component list] get 0 end]
  set index [lsearch $listOfEntries $grammarName]
  $box selection set $index $index

  # initialize grammar
  if {!$_grammarIsInitialized} {
    _initializeGrammar
  }

  # if there is an xml file for the selected grammar, load it
  # otherwise prepare the table
  if {![load]} {

    # fill title column: show experiments of the given grammar
    set row 1
    foreach experiment $experiments {
      $table setCell $row,0 $experiment
      incr row
    }
    # fill title row: show the labels
    _showLabels
  }

  updateWindowTitle
  updateMenus
}

## ----------------------------------------------------------------------------
## _initializeGrammar triggers the comuptation of the annotation
## ----------------------------------------------------------------------------
body YadaLabelStats::_initializeGrammar {} {
  _initAnnos
  _initData
  set _grammarIsInitialized 1
  .main printStatus ""
}

## ----------------------------------------------------------------------------
## show the labels occuring in the selected grammar
## in the title-row of the table
## ----------------------------------------------------------------------------
body YadaLabelStats::_showLabels {{labels ""}} {

  set table $itk_component(table)
  if {$labels == ""} {
    set displayLabels $allLabels
    # remove the label "" -- only used for computing
    set index [lsearch -exact $displayLabels ""]
    if {$index >= 0} {
      set displayLabels [lreplace $displayLabels $index $index]
    }
    lappend displayLabels total

  } else {
    set displayLabels $labels
  }

  set noLabels [llength $displayLabels]
  if {$noLabels > 0} {
    $itk_component(table) configure -cols [expr $noLabels + 1]
  }

  # fill it
  $table setCell 0,0 "Experiment"
  $table setCell row 0,1 $displayLabels

}

## ----------------------------------------------------------------------------
## inserts the rounded data for an experiment in the given row
## ----------------------------------------------------------------------------
body YadaLabelStats::_insertData {experiment} {
  set table $itk_component(table)
  set cmax [$table cget -cols]

  foreach label $displayLabels {
    lappend args [format "%.2f" $recallLabels($experiment,$label)]
  }
  set row [_getRowOfExperimentName $experiment]

  eval $table setCell row $row,1 [list $args]
}

## ----------------------------------------------------------------------------
## clear the selected rows
## ----------------------------------------------------------------------------
body YadaLabelStats::_clearRow {} {

  set table $itk_component(table)
  set cmax [$table cget -cols]

  foreach experiment [_getSelectedExperiments] {
    set row [_getRowOfExperimentName $experiment]
    $table clear tags $row,0 $row,$cmax
    $table erase $row,1 $row,$cmax
  }
}

## ----------------------------------------------------------------------------
## loads the label statistic xml-file
## TODO user is not warned that all unsaved values will be discarded
## ----------------------------------------------------------------------------
body YadaLabelStats::reload {} {
  clear
  setModified
  load
}

## ----------------------------------------------------------------------------
## _leaveTableCommand
## ----------------------------------------------------------------------------
body YadaLabelStats::_leaveTableCommand {} {
  set table $itk_component(table)
  set lastActiveCell [$table tag cell active]
  if {$lastActiveCell == ""} {
    return
  }

  scan $lastActiveCell "%d,%d" row col
  $table activate -1,-1
}

## ----------------------------------------------------------------------------
## _escapeCommand
## ----------------------------------------------------------------------------
body YadaLabelStats::_escapeCommand {} {
  set table $itk_component(table)
  set lastActiveCell [$table tag cell active]
  scan $lastActiveCell "%d,%d" row col
  
  if {$col == 0} {
    $table selection clear all
  } else {
    $table activate -1,-1
  }
}

## ----------------------------------------------------------------------------
## @returns the row of an experiment
## ----------------------------------------------------------------------------
body YadaLabelStats::_getRowOfExperimentName {experimentName} {

  set table $itk_component(table)
  set rmax [$itk_component(table) cget -rows]
  set experimentNames [$itk_component(table) getCell 0,0 $rmax,0]
  set row [lsearch -exact $experimentNames $experimentName]

  return $row
}

## ----------------------------------------------------------------------------
## _returnCommand
## ----------------------------------------------------------------------------
body YadaLabelStats::_returnCommand {} {
  set table $itk_component(table)
  set lastActiveCell [$table tag cell active]
  scan $lastActiveCell "%d,%d" row col
  set colMax [expr [$table cget -cols] -1]
  if {$col == $colMax} {
    incr row
    set col 1
  } else {
    incr col
  }

  $table activate $row,$col
  $table see $row,$col
}

## ----------------------------------------------------------------------------
## isFollowData
## ----------------------------------------------------------------------------
body YadaLabelStats::isFollowData {} {
  return $_followDataFlag
}

## ----------------------------------------------------------------------------
## toggleFollowData
## ----------------------------------------------------------------------------
body YadaLabelStats::toggleFollowData {} {
  if {$_followDataFlag == 1} {
    set _followDataFlag 0
  } else {
    set _followDataFlag 1
  }

  return $_followDataFlag
}

## ----------------------------------------------------------------------------
## _rowTagCommand
## ----------------------------------------------------------------------------
body YadaLabelStats::_rowTagCommand {row} {
  if {[expr $row % 2]} {
    return "colored"
  } else {
    return "default"
  }
}

## ----------------------------------------------------------------------------
## serializes this label statistics into a dom document
## @returns a new dom document
## ----------------------------------------------------------------------------
body YadaLabelStats::toDom {} {

  set document [dom createDocument "yada"]
  set root [$document documentElement]

  set labelstatsNode [$document createElement "labelstats"]
  $root appendChild $labelstatsNode

  # grammar
  set grammarNode [$document createElement "grammar"]
  set grammarName [_getCurrentGrammar]
  $labelstatsNode appendChild $grammarNode
  $grammarNode setAttribute "name" $grammarName

  set table $itk_component(table)
  set rmax [$table cget -rows]
  set cmax [$table cget -cols] ; # dummy
  set labels [$table getCell 0,1 0,$cmax]

  # Add the occuring labels, thus the title row
  if {$labels != ""} {
    set labelsNode [$document createElement "labels"]
    $labelstatsNode appendChild $labelsNode
    set labelsNodeText [$document createTextNode $labels]
    $labelsNode appendChild $labelsNodeText
  }

  # statistics read from table
  for {set row 1} {$row < $rmax} {incr row} {

    set dataRow [$table getCell $row,0 $row,$cmax]
    set experimentName [lindex $dataRow 0]
    if {$experimentName == ""} {
        continue
    }

    set experimentNode [$document createElement "experiment"]
    $labelstatsNode appendChild $experimentNode
    $experimentNode setAttribute "name" $experimentName
    for {set col 1} {$col < $cmax} {incr col} {

      set label [lindex $labels [expr $col - 1]]
      set value [lindex $dataRow $col]
      if {$value != ""} {
	set labelNode [$document createElement "label"]
	$experimentNode appendChild $labelNode
	$labelNode setAttribute name $label
	$labelNode setAttribute value $recallLabels($experimentName,$label)
      }
    }
  }

  return $document
}

## ----------------------------------------------------------------------------
## initialize this label statistics from a dom document
## @param document a dom document containing a /yada/labelstats
## ----------------------------------------------------------------------------
body YadaLabelStats::fromDom {document} {
  set rootNode [$document documentElement]
  set labelstatsNode [$rootNode selectNodes "/yada/labelstats"]
  fromDomNode $labelstatsNode
}

## ----------------------------------------------------------------------------
## @param labelstatsNode containg a label statistics
## ----------------------------------------------------------------------------
body YadaLabelStats::fromDomNode {labelstatsNode} {

  set labelsNode [$labelstatsNode selectNodes "./labels"]
  if {$labelsNode == "" } {
    return
  } else {
    _showLabels [$labelsNode text]
    set labelsOfExperiment [$labelsNode text]
  }

  set table $itk_component(table)
  set rmax [$table cget -rows]
  set cmax [$table cget -cols]
  $table erase 1,1 $rmax,$cmax
  
  # read in the experiments saved in the xml-file
  set experimentNodes [$labelstatsNode selectNodes "./experiment"]
  set row 0
  foreach experimentNode $experimentNodes {
    incr row
    set args ""
    set experimentName [$experimentNode getAttribute "name"]
    $table setCell $row,0 $experimentName
    foreach label $labelsOfExperiment {

    # read in the data for this experiment, if available
      set labelNode [$experimentNode selectNode "./label\[@name=\"$label\"\]"]
      set isComputed 0
      if {$labelNode != ""} {
	set isComputed 1
        set value [$labelNode getAttribute value]
	set recallLabels($experimentName,$label) $value
      }
    }

    # ... and display the data in the table thus in the gui
    if {$isComputed} {
      _insertData $experimentName
    }
  }
}

## ----------------------------------------------------------------------------
## get the title string representing the state of this document
## ----------------------------------------------------------------------------
body YadaLabelStats::getWindowTitle {} {
  if {[info exists itk_component(grammarBox)]} {
    set box $itk_component(grammarBox)
    set grammarName [$box get]
  } else {
    set grammarName "<none>"
  }

  if {$grammarName == "" || $grammarName == "<none>" } {
    return "[chain]"
  } else {
    return "[chain]: $grammarName"
  }
}


## ----------------------------------------------------------------------------
## get the filename where the persistent state is saved in
## ----------------------------------------------------------------------------
body YadaLabelStats::getFileName {} {
  global env
  
  if { [info exists itk_component(grammarBox)] } {
    set box $itk_component(grammarBox)
    set grammarName [$box get]
  } else {
    set grammarName "none"
  }
  
  set fileName [file join $env(YADA_DATA) ${grammarName}_labelstats.xml]
  return [getSecureFileName $fileName]
}

## ----------------------------------------------------------------------------
## Returns name of currently selected grammar or the empty string,
## if selection box is not initialized yet 
## ----------------------------------------------------------------------------
body YadaLabelStats::_getCurrentGrammar {} {
  if { [info exists itk_component(grammarBox)] } {
    return [$itk_component(grammarBox) get]
  } else {
    return ""
  }
}

## ----------------------------------------------------------------------------
## @returns true if the selected experiment is not <none> and the document
## is modified
## ----------------------------------------------------------------------------
body YadaLabelStats::isModified {} {
  set grammarName [_getCurrentGrammar]
  if {$grammarName == "" || $grammarName == "<none>"} {
    return 0
  }
  
  return [chain]
}

## ----------------------------------------------------------------------------
## computes the statistics for the annotations of an experiment
## ----------------------------------------------------------------------------
body YadaLabelStats::_initAnnos {{anno ""}} {
  
  foreach label $allLabels {
    set totalNrLabels($label) 0
  }
  set i 0
  set datadir [_getAnnotationDir]
  foreach file [lsort -dictionary [glob $datadir/*.xml.gz]] {
    if {$i >= $maxWg}  {
      break
    }
    _readAnno $file
    incr i
  }

  set totalTries 0
  foreach label $allLabels {
    incr totalTries $totalNrLabels($label)
  }
}

## ----------------------------------------------------------------------------
## Sets all variables for the statistics of an experiment to 0
## If no experiment is specified, does that for all experiments.
## @param weight experiment
## ----------------------------------------------------------------------------
body YadaLabelStats::_initData {{weight ""}} {
  if {$weight == ""} {
    set weights $allWeights
  } else {
    set weights [list $weight]
  }

  foreach arr {labelErrorCount rejectedLabelCount recallLabels} {
    foreach weight $weights {
      foreach label $allLabels {
	set [subst $arr]($weight,$label) 0
      }
      set recallLabels($weight,total) 0
    }
  }

  foreach arr {errorCount rejectionCount totalLabelledHits totalRecall\
    }  {
    foreach weight $weights {
      set [subst $arr]($weight) 0
    }
  }

}

## ----------------------------------------------------------------------------
## _readExperiment
## @param weight experiment
## ----------------------------------------------------------------------------
body YadaLabelStats::_readExperiment {weight} {

  set directory ""
  foreach exp [.main getExperiments] {
    if {[regexp $weight [$exp cget -name]]} {
      set directory [$exp cget -dataDir]
      break
    }
  }

  set i 0
  foreach file [lsort -dictionary [glob $directory/*.xml.gz]] {
    if {$i >= $maxWg}  {
      break
    }
    _readFile $file $weight
    incr i
  }
}

## ----------------------------------------------------------------------------
## _readFile
## @param weight experiment
## ----------------------------------------------------------------------------
body YadaLabelStats::_readFile {fileName weight} {

  set fileName [glob $fileName]
  .main printStatus "... computing statistics"

  set fileID [open "|zcat $fileName" {RDONLY}]
  set xmlData [read  $fileID]
  close $fileID

  set document [dom parse $xmlData]
  set rootNode [$document documentElement]
  set annoNode [$rootNode selectNode {*/annotation}]
  if {$annoNode != ""} {
    set errorNodes [$rootNode selectNode {*/parseverification/error[@level="SYN"]}]
    foreach errorNode $errorNodes {
      set type [$errorNode getAttribute "type"]
      if {$type == "label"} {
	set annoLabel [$errorNode getAttribute "annoLabel"]
	incr labelErrorCount($weight,$annoLabel); # count label errors
	continue
      }
      if {$type == "struct"} {
	set modifiee [$errorNode getAttribute "modifiee"]
	set annoModifiee [$errorNode getAttribute "annoModifiee"]
	set annoLabel [$errorNode getAttribute "annoLabel"]
	if {$modifiee == 0 && $annoModifiee != 0} {
	  incr rejectedLabelCount($weight,$annoLabel); # count rejections
	  continue
	}
	incr labelErrorCount($weight,$annoLabel); # count label errors in struct errors
      }
    }
    set labelledHitsNumberNode [$rootNode selectNode {*/parseverification/labelledhits/number[@name="SYN"]}]
    incr totalLabelledHits($weight) [$labelledHitsNumberNode getAttribute "value"]
  } else {
    # no network found
    set fileName [[$rootNode selectNode "*/file"] @name]
    regexp ".*\/(.*?)\.xml$" $fileName dummy wg
    puts "\nWARNING: no parse for $fileName"
    foreach label $labelsInAnno($wg)  {
      incr labelErrorCount($weight,$label); # count label errors
    }
  }

  $document delete
}

## ----------------------------------------------------------------------------
## _readAnno
## ----------------------------------------------------------------------------
body YadaLabelStats::_readAnno {fileName}  {

  .main printStatus "... computing statistics for annotation"
  set fileName [glob $fileName]

  set fileID [open "|zcat $fileName" {RDONLY}]
  set xmlData [read  $fileID]
  close $fileID

  set document [dom parse $xmlData]
  _parseAnno $document
  $document delete
}

## ----------------------------------------------------------------------------
## compute statistics for an annotation
## ----------------------------------------------------------------------------
body YadaLabelStats::_parseAnno {document} {

  set rootNode [$document documentElement]
  set fileName [[$rootNode selectNode "*/file"] @name]
  regexp ".*\/(.*?)\.xml$" $fileName dummy wg

  set annoNode [$rootNode selectNode "*/annotation"]
  foreach synNode [$annoNode selectNodes {arc/dep[@level="SYN"]}] {
    set label [$synNode getAttribute "label"]
    lappend labelsInAnno($wg) $label
    incr totalNrLabels($label)
  }
}

## ----------------------------------------------------------------------------
## compute statistics for an experiment
## ----------------------------------------------------------------------------
body YadaLabelStats::_evaluateExperiment {weight} {
  
  foreach label $allLabels {
    if {$totalNrLabels($label) == 0} {
      set recallLabels($weight,$label) 0.0
      set rejectionLabels($weight,$label) 100.0
    } else {
      set recallLabels($weight,$label) \
	[expr (1 - (0.0 + $labelErrorCount($weight,$label) + $rejectedLabelCount($weight,$label)) / $totalNrLabels($label)) * 100.0]
    }
    lappend labelRecalls($recallLabels($weight,$label)) $label
    incr errorCount($weight) $labelErrorCount($weight,$label)
    incr rejectionCount($weight) $rejectedLabelCount($weight,$label)
  }
  set totalRecall($weight) \
    [expr (1 - (0.0 + $errorCount($weight) + $rejectionCount($weight)) / $totalTries) * 100.0]
  set recallLabels($weight,total) $totalRecall($weight)
}

## ----------------------------------------------------------------------------
## _getSelectedExperiments
## ----------------------------------------------------------------------------
body YadaLabelStats::_getSelectedExperiments {} {
  set table $itk_component(table)

  set cells [$table curselection]

  foreach index $cells {
    scan $index "%d,%d" row col
    set experimentName [$table getCell $row,0]
    if {$experimentName != ""} {
      set id($row,0) $experimentName
    }
  }

  set experimentNames ""
  foreach index [array names id] {
    lappend experimentNames $id($index)
  }


  return [lsort -dictionary $experimentNames]
}

## ----------------------------------------------------------------------------
## _getUndoneExperiments
## ----------------------------------------------------------------------------
body YadaLabelStats::_getUndoneExperiments {} {
  set table $itk_component(table)
  set rmax [$table cget -rows]
  set experimentNames ""
  foreach {experimentName firstColumn} [$table getCell 1,0 $rmax,1] {
    if {$experimentName != "" && $firstColumn == ""} {
      lappend experimentNames $experimentName
    }
  }

  return $experimentNames
}

## ----------------------------------------------------------------------------
## get all Experiemnts we should compute:
## these are the selected ones
## or, if there are no experiments selected, all experiments
## ----------------------------------------------------------------------------
body YadaLabelStats::_getExperimentsToCompute {} {
  set experiments [lsort -dictionary [_getSelectedExperiments]]

  # nothing selected ?
  if {$experiments == ""} {
    set experiments [lsort -dictionary [_getUndoneExperiments]]
  }
  return $experiments
}

## ----------------------------------------------------------------------------
## _getExperimentsOfGrammar
## ----------------------------------------------------------------------------
body YadaLabelStats::_getExperimentsOfGrammar {} {

  set grammarName [_getCurrentGrammar]

  # get all experiments using this grammar
  set experimentNames ""
  foreach experiment [.main getExperiments] {
    set experimentName [$experiment cget -name]
    if {$experimentName == "<none>" || 
        [regexp annotation $experimentName] ||
	![regexp $grammarName [$experiment cget -grammarName]]} {
      continue
    }

    if {![$experiment cget -isComplete]} {
      .main printStatus "... checking experiment $experimentName"
      $experiment check
    }

    if {![$experiment cget -isComplete]} {
      continue
    }

    lappend experimentNames $experimentName
  }
  .main printStatus ""
  set experimentNames [lsort $experimentNames]

  return $experimentNames
}

## ----------------------------------------------------------------------------
## @returns the directory path of the annotation experiment for 
## the selected grammar 
## ----------------------------------------------------------------------------
body YadaLabelStats::_getAnnotationDir {} {

  set grammarName [_getCurrentGrammar]

  # get the annotation experiment of this grammar
  foreach experiment [.main getExperiments] {
    set experimentName [$experiment cget -name]
    if {$experimentName == "<none>" || 
        ![regexp $grammarName [$experiment cget -grammarName]] ||
        ![regexp annotation $experimentName]} {
      continue
    }

      if {![$experiment cget -isComplete]} {
	.main printStatus "... checking experiment $experimentName"
	$experiment check
      }

      .main printStatus ""

      if {![$experiment cget -isComplete]} {
	continue
      }

      return [$experiment cget -dataDir]
  }

  return ""
}
