import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URI;
import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;

import org.semanticweb.owl.apibinding.OWLManager;
import org.semanticweb.owl.io.OWLXMLOntologyFormat;
import org.semanticweb.owl.model.OWLAxiom;
import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLDataFactory;
import org.semanticweb.owl.model.OWLIndividual;
import org.semanticweb.owl.model.OWLObjectPropertyExpression;
import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.model.OWLOntologyChangeException;
import org.semanticweb.owl.model.OWLOntologyCreationException;
import org.semanticweb.owl.model.OWLOntologyManager;
import org.semanticweb.owl.model.OWLOntologyStorageException;
import org.semanticweb.owl.util.SimpleURIMapper;

public class VanKampenContextGenerator {

    private OWLOntologyManager _manager;

    private OWLOntology _tBox;

    public VanKampenContextGenerator(OWLOntologyManager manager,
	    OWLOntology tBox) {
	_manager = manager;
	_tBox = tBox;
    }

    public static OWLOntology loadOntologyFromFile(OWLOntologyManager manager,
	    String filename) {
	URI uri = URI.create("file:" + System.getProperty("user.dir")
		+ File.separator + filename);
	OWLOntology ontology = null;
	try {
	    ontology = manager.loadOntologyFromPhysicalURI(uri);
	} catch (OWLOntologyCreationException e) {
	    System.err.println("Could not load file: " + filename);
	    e.printStackTrace();
	    System.exit(1);
	}
	return ontology;
    }

    private OWLOntology createABox(String owlFile, TBoxParser tBoxParser) {
	URI aBoxURI = URI
		.create("http://cinacs.informatik.uni-hamburg.de/ontologies/"
			+ owlFile);
	Logger.getLogger("ContextFileGenerator").fine("A-Box URI: " + aBoxURI);
	OWLOntology aBox = null;

	try {
	    URI aBoxPhysicalURI = URI.create("file:"
		    + new File(".").getCanonicalPath() + "/" + owlFile);
	    _manager
		    .addURIMapper(new SimpleURIMapper(aBoxURI, aBoxPhysicalURI));
	    Logger.getLogger("ContextFileGenerator").fine(
		    "A-Box physical URI: " + aBoxPhysicalURI);

	    aBox = _manager.createOntology(aBoxURI);
	    _manager.setOntologyFormat(aBox, new OWLXMLOntologyFormat());

	    _manager.addAxiom(aBox, _manager.getOWLDataFactory()
		    .getOWLImportsDeclarationAxiom(aBox, tBoxParser.getURI()));
	} catch (Exception e) {
	    System.err.println("Error creating A-Box: " + owlFile);
	    e.printStackTrace();
	}

	return aBox;
    }

    private Map<Integer, OWLIndividual> createIndividuals(OWLOntology aBox,
	    Map<Integer, OWLClass> concepts) {
	Map<Integer, OWLIndividual> individuals = new HashMap<Integer, OWLIndividual>();

	Map<String, Integer> wordCount = new HashMap<String, Integer>();
	OWLDataFactory factory = _manager.getOWLDataFactory();
	for (int id : concepts.keySet()) {
	    OWLClass concept = concepts.get(id);

	    String word = concept.toString();

	    int counter;
	    if (wordCount.containsKey(word)) {
		counter = wordCount.get(word) + 1;
	    } else {
		counter = 1;
	    }
	    wordCount.put(word, counter);

	    URI individualURI = URI.create(String.format("%s#%s_%02d", aBox
		    .getURI(), word, counter));
	    Logger.getLogger("ContextFileGenerator").fine(
		    "Individual URI for " + word + ": " + individualURI);

	    OWLIndividual individual = factory.getOWLIndividual(individualURI);
	    individuals.put(id, individual);

	    try {
		_manager.addAxiom(aBox, factory.getOWLClassAssertionAxiom(
			individual, concept));
		_manager.addAxiom(aBox, factory
			.getOWLDeclarationAxiom(individual));
	    } catch (OWLOntologyChangeException e) {
		System.err.println("Error creating individual " + individual
			+ " in A-Box: " + aBox);
	    }
	}

	return individuals;
    }

    private void createRoles(OWLOntology aBox, ThematicRelations relations,
	    Map<Integer, OWLIndividual> individuals, TBoxParser tBoxParser) {
	OWLDataFactory factory = _manager.getOWLDataFactory();
	for (int id : individuals.keySet()) {
	    if (relations.hasRole(id)
		    && individuals.containsKey(relations.getRegent(id))) {
		OWLIndividual dependant = individuals.get(id);
		OWLIndividual regent = individuals.get(relations.getRegent(id));
		OWLObjectPropertyExpression property = tBoxParser
			.getPropertyForRole(relations.getRole(id));

		OWLAxiom axiom = factory.getOWLObjectPropertyAssertionAxiom(
			dependant, property, regent);
		try {
		    _manager.addAxiom(aBox, axiom);
		} catch (OWLOntologyChangeException e) {
		    System.err.println("Error creating role " + axiom
			    + " in A-Box: " + aBox);
		}
	    }
	}
    }

    /**
     * Generate context models in OWL-XML format.
     * 
     * @param filename
     *            The filename of the van-Kampen-Corpus file for which thematic
     *            relations shall be generated and saved in context models.
     * @throws IOException
     */
    public void generateFromLines(String filename, SentencesParser parser)
	    throws IOException {
	BufferedReader in = new BufferedReader(new InputStreamReader(
		new FileInputStream(filename), Charset.forName("ISO-8859-1")));
	Pair<ThematicRelations, ThematicRelations> relations;

	while ((relations = parser.parseLine(in)) != null) {
	    ThematicRelations genitiveReading = relations.x;
	    ThematicRelations dativeReading = relations.y;

	    TBoxParser genitiveTBoxParser = new TBoxParser(_manager, _tBox);
	    Map<Integer, OWLClass> genitiveConcepts = genitiveTBoxParser
		    .mapIndividualsToConcepts(genitiveReading);
	    OWLOntology genitiveABox = createABox(
		    "van.Kampen_337_Genitive.Context_s"
			    + genitiveReading.getID() + ".owl",
		    genitiveTBoxParser);
	    Map<Integer, OWLIndividual> genitiveIndividuals = createIndividuals(
		    genitiveABox, genitiveConcepts);
	    createRoles(genitiveABox, genitiveReading, genitiveIndividuals,
		    genitiveTBoxParser);
	    try {
		_manager.saveOntology(genitiveABox);
	    } catch (OWLOntologyStorageException e) {
		System.err.println("Error saving genitive reading A-Box: "
			+ genitiveABox);
		e.printStackTrace();
	    }

	    TBoxParser dativeTBoxParser = new TBoxParser(_manager, _tBox);
	    Map<Integer, OWLClass> dativeConcepts = dativeTBoxParser
		    .mapIndividualsToConcepts(dativeReading);
	    OWLOntology dativeABox = createABox(
		    "van.Kampen_337_Dative.Context_s" + dativeReading.getID()
			    + ".owl", dativeTBoxParser);
	    Map<Integer, OWLIndividual> dativeIndividuals = createIndividuals(
		    dativeABox, dativeConcepts);
	    createRoles(dativeABox, dativeReading, dativeIndividuals,
		    dativeTBoxParser);
	    try {
		_manager.saveOntology(dativeABox);
	    } catch (OWLOntologyStorageException e) {
		System.err.println("Error saving dative reading A-Box: "
			+ dativeABox);
		e.printStackTrace();
	    }
	}
	in.close();
    }

    /**
     * @param args
     *            The first arg should be the path to the T-Box, the second arg
     *            should be the path to file to get the base forms from, and the
     *            third arg should be the path to the van-Kampen-Corpus file.
     */
    public static void main(final String[] args) {
	final OWLOntologyManager manager = OWLManager
		.createOWLOntologyManager();

	final String tBoxFilename = args[0];
	final String grammarFilename = args[1];
	final String corpusFilename = args[2];

	final SentencesParser parser = new SentencesParser(grammarFilename);

	final VanKampenContextGenerator generator = new VanKampenContextGenerator(
		manager, loadOntologyFromFile(manager, tBoxFilename));
	try {
	    generator.generateFromLines(corpusFilename, parser);
	} catch (FileNotFoundException e) {
	    System.err.println("Corpus file not found: " + corpusFilename);
	    e.printStackTrace();
	} catch (IOException e) {
	    System.err.println("Error reading corpus file: " + corpusFilename);
	}
    }
}
